/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * <OWNER> = Apple Inc.
 * <ORGANIZATION> = Apple Inc.
 * <YEAR> = 2015
 *
 * Copyright (c) 2015, Apple Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the <ORGANIZATION> nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *************************************************************************************
 * \file ColorSpaceMatrix.H
 *
 * \brief
 *    ColorSpaceMatrix class Header
 *
 * \author
 *     - Alexis Michael Tourapis         <atourapis@apple.com>
 *
 *************************************************************************************
 */


#ifndef __ColorSpaceMatrix_H__
#define __ColorSpaceMatrix_H__

#include "Global.H"
#include "Frame.H"
#include "ColorTransform.H"

enum {
  CTF_IDENTITY            = 0,
  CTF_RGB709_2_YUV709     = 1,
  CTF_RGBP3_2_YUVP3       = 2,
  CTF_RGB2020_2_YUV2020   = 3,
  CTF_RGBSRGB0_2_XYZ      = 4,
  CTF_RGBSRGB1_2_XYZ      = 5,
  CTF_RGBCIE_2_XYZ        = 6,
  CTF_RGBWG_2_XYZ         = 7,
  CTF_RGBSMPTEC_2_XYZ     = 8,
  CTF_RGBSGAMUT_2_XYZ     = 9,
  CTF_RGBACES_2_XYZ       = 10,
  CTF_RGBP3D60_2_XYZ      = 11,
  CTF_RGB2020_2_XYZ       = 12,
  CTF_RGBP3DCI_2_XYZ      = 13,
  CTF_RGBP3D65_2_XYZ      = 14,
  CTF_RGB709_2_XYZ        = 15,
  CTF_RGB709_2_RGB2020    = 16,
  CTF_RGBP3D65_2_RGB2020  = 17,
  CTF_RGB709_2_RGBP3D65   = 18,
  CTF_XYZ_2_DZDX          = 19, // SMPTE ST 2085  Y' D'z' D'x
  CTF_RGB2020_2_YUV2020CL = 20,
  CTF_RGBP3D65_2_YUVP3D65 = 21,
  CTF_RGBP3D60_2_RGB2020  = 22,
  CTF_RGB709_2_RGBP3D60   = 23,
  CTF_RGBP3D60_2_YUVP3D60 = 24,
  CTF_RGB_2_YCOCG         = 26, // Currently uses YCoCg
  CTF_RGB_2_AMT           = 30,
  CTF_RGBEXT_2_YUVEXT     = 31,
  CTF_RGBEXT_2_XYZ        = 32,
  CTF_RGB709_2_RGBEXT     = 33,
  CTF_RGB2020_2_RGBEXT    = 34,
  CTF_RGBP3D65_2_RGBEXT   = 35,
  CTF_TOTAL
};


static const double FWD_TRANSFORM[36][3][3] = {
  { // Identity transform (0)
    {1.0, 0.0, 0.0},
    {0.0, 1.0, 0.0},
    {0.0, 0.0, 1.0}
  },
  { // RGB to YUV BT.709 (1)
    {  0.212600,  0.715200,   0.072200 },   // Y'
    { -0.114572, -0.385428,   0.500000 },   // Cb'
    {  0.500000, -0.454153,  -0.045847 }    // Cr'
  },
  { // RGB to YUV P3 DCI (2)
    {  0.2096,     0.7215,   0.0690 },
    { -0.1125,    -0.3875,   0.5000 },
    {  0.5000,    -0.4564,  -0.0436 }
  },
  { // RGB to YUV BT.2020 (3)
    {  0.262700,   0.678000,   0.059300 },
    { -0.139630,  -0.360370,   0.500000 },
    {  0.500000,  -0.459786,  -0.040214 }
  },
  { // RGB to XYZ srgb (high precision) (4)
    { 0.4124564,  0.3575761,  0.1804375 },
    { 0.2126729,  0.7151522,  0.0721750 },
    { 0.0193339,  0.1191920,  0.9503041 }
  },
  { // RGB to XYZ srgb (low precision) (5)
    { 0.4124,    0.3576,    0.1805},
    { 0.2126,    0.7152,    0.0722},
    { 0.0193,    0.1192,    0.9504}
  },
  { // RGB to XYZ CIE (6)
    { 0.4887180,  0.3106803,  0.2006017 },
    { 0.1762044,  0.8129847,  0.0108109 },
    { 0.0000000,  0.0102048,  0.9897952 }
  },
  { // RGB to XYZ  Wide Gamut RGB (7)
    { 0.7161046,  0.1009296,  0.1471858 },
    { 0.2581874,  0.7249378,  0.0168748 },
    { 0.0000000,  0.0517813,  0.7734287 }
  },
  { // RGB to XYZ  SMPTE-C RGB (8)
    { 0.3935891,  0.3652497,  0.1916313 },
    { 0.2124132,  0.7010437,  0.0865432 },
    { 0.0187423,  0.1119313,  0.9581563 }
  },
  { // RGB to XYZ S-Log (9)
    {  0.7064827132,  0.1288010498,  0.1151721641 },
    {  0.2709796708,  0.7866064112, -0.0575860820 },
    { -0.0096778454,  0.0046000375,  1.0941355587 }
  },
  { // RGB to XYZ ACES RGB (10)
    { 0.9525523959, 0.0000000000,  0.0000936786 },
    { 0.3439664498, 0.7281660966, -0.0721325464 },
    { 0.0000000000, 0.0000000000,  1.0088251844 }
  },
  { // RGB to XYZ P3D60 (11)
    { 0.5049495342, 0.2646814889, 0.1830150515 },
    { 0.2376233102, 0.6891706692, 0.0732060206 },
    { 0.0000000000, 0.0449459132, 0.9638792711 }
  },
  { // RGB to XYZ BT.2020 (12)
    { 0.636958, 0.144617, 0.168881 },
    { 0.262700, 0.677998, 0.059302 },
    { 0.000000, 0.028073, 1.060985 }
    /*
     { 0.636958048301291,   0.144616903586208,   0.168880975164172 },
     { 0.262700212011267,   0.677998071518871,   0.059301716469862 },
     { 0.000000000000000,   0.028072693049087,   1.060985057710791 }
     */
  },
  { // RGB to XYZ P3 DCI (13)
    { 0.445169815564552,   0.277134409206778,   0.172282669815565 },
    { 0.209491677912731,   0.721595254161044,   0.068913067926226 },
    { 0.000000000000000,   0.047060560053981,   0.907355394361973 }
  },
  { // RGB to XYZ P3 D65 (14)
    { 0.486571, 0.265668, 0.198217 },
    { 0.228975, 0.691739, 0.079287 },
    { 0.000000, 0.045113, 1.043944 }
  },
  { // RGB to XYZ BT.709 (15) // values are slightly different from 4
    { 0.412391,  0.357584,  0.180481 },
    { 0.212639,  0.715169,  0.072192 },
    { 0.019331,  0.119195,  0.950532 }
  },
  {  // RGB BT.709 to RGB BT.2020 (16)
    { 0.627404078626000,   0.329282097415000,   0.043313797587000 },
    { 0.069097233123000,   0.919541035593000,   0.011361189924000 },
    { 0.016391587664000,   0.088013255546000,   0.895595009604000 }
  },
  { // RGB P3 D65 to RGB BT.2020 (17)
    {  0.753832826496000,   0.198597635641000,   0.047569409186000 },
    {  0.045744636411000,   0.941777687331000,   0.012478735611000 },
    { -0.001210377285000,   0.017601107390000,   0.983608137835000 }
  },
  {  // RGB BT.709 to RGB P3 D65 (18)
    { 0.822462153657398,   0.177537532355145,   0.000000768358644 },
    { 0.033193617909869,   0.966804868483336,  -0.000000257149122 },
    { 0.017082707432523,   0.072397832913504,   0.910519870270803 }
  },
  { // X'Y'Z' to Y'Dz'Dx'  (SMPTE ST 2085 section 4) (19)
    //        X                    Y                     Z
    { 0.000000000000000,   1.000000000000000,   0.000000000000000 },  // Y' = Y'
    { 0.000000000000000,  -0.500000000000000,   0.493283000000000 },  // D'z = ( c1*Z' - Y' ) / 2.0  where c1 = 0.986566
    { 0.500000000000000,  -0.495951000000000,   0.000000000000000 }   // D'x = ( X' - c2*Y' ) / 2.0  where c2 = 0.991902
  },
  // Place holder for constant luminance (does not work as a single matrix ...
  { // ITU-R BT.2020 constant luminance --- similar to case (3) above. (20)
    //        R                   G                    B
    { 0.000000000000000,   1.000000000000000,   0.000000000000000 },  // Y'c = 0.2627R + 0.6780G + 0.0593B
    { 0.000000000000000,   0.000000000000000,   0.000000000000000 },  //
    { 0.000000000000000,   0.000000000000000,   0.000000000000000 },  //
  },
  { // RGB to YUV P3 D65 (21)
    //{  0.228975, 0.691739,  0.079287 },
    //{ -0.124346,-0.375654,  0.500000 },
   // {  0.500000,-0.448583, -0.051417 }
    
    {0.228974564069749,   0.691738521836506,   0.079286914093745 },
    {-0.124346317856648,  -0.375653682143352,   0.500000000000000 },
    {0.500000000000000,  -0.448583464047406,  -0.051416535952594 }
  },
  
  { // RGB P3 D60 to RGB BT.2020 (22)
    {  0.689691223849297,   0.207169095236480,   0.041345659123841},
    {  0.041851838607333,   0.982426319652515,   0.010845834163914},
    { -0.001107373497014,   0.018361151170161,   0.854913989638303},
  },
  { // RGB BT.709 to P3 D60 (23)
    { 0.898952033670153,   0.194047855009400,   0.000000964735234},
    { 0.031820589914418,   0.926804052445161,  -0.000000704588889},
    { 0.019654377250385,   0.083296008054360,   1.047585145827092},
  },
  { // RGB to YUV P3 D60 (24)
    //{  0.2376233102,  0.6891706692,  0.0732060206 },
    //{ -0.1125000000, -0.3875000000,  0.5000000000 },
    //{  0.5000000000, -0.4564000000, -0.0436000000 }
    {  0.237524539590539,   0.689333457978873,   0.073142002430588 },
    { -0.128134266637081,  -0.371865733362919,   0.500000000000000 },
    {  0.500000000000000,  -0.452036487580000,  -0.047963512420000 }
  },
  {
  /*
    {   0.2000,    0.7500,     0.0500 },
    {  -0.1053,    -0.3947,   0.5000 },
    {   0.5000,    -0.4688,     -0.0312 },
   
    { 0.1750,      0.8000,    0.0250 },
    { -0.08974,   -0.41026,    0.5000 },
    { 0.5000   -0.48485,   -0.01515 }
   
    { 0.200000000000000,   0.700000000000000,   0.100000000000000},
    {   -0.111111111111111,  -0.388888888888889,   0.500000000000000},
  {0.500000000000000,  -0.437500000000000,  -0.062500000000000}
   
    {0.250000000000000,   0.650000000000000,   0.100000000000000},
    {-0.138888888888889,  -0.361111111111111,   0.500000000000000},
    {0.500000000000000,  -0.433333333333333,  -0.066666666666667},
    { 0.250000000000000,   0.700000000000000,   0.050000000000000},
    {  -0.131578947368421,  -0.368421052631579,   0.500000000000000},
    {  0.500000000000000,  -0.466666666666667,  -0.033333333333333}, */
  /*  {  0.300000000000000,   0.600000000000000,   0.100000000000000},
    {  -0.166666666666667,  -0.333333333333333,   0.500000000000000},
    { 0.500000000000000,  -0.428571428571429,  -0.071428571428571}, */
    { 0.250000000000000,   0.500000000000000,   0.250000000000000 },
    {-0.166666666666667,  -0.333333333333333,   0.500000000000000 },
    { 0.500000000000000,  -0.333333333333333,  -0.166666666666667 }
  },
  { // YCoCg
    { 0.25, 0.5,  0.25},
    {-0.25, 0.5, -0.25},
    { 0.50, 0.0, -0.50}
  },
  {
    {0.150000000000000,   0.700000000000000,   0.150000000000000},
    {-0.088235294117647,  -0.411764705882353,   0.500000000000000},
    {0.500000000000000,  -0.411764705882353,  -0.088235294117647}
  },
  {
    { 0.125000000000000,   0.750000000000000,   0.125000000000000},
    {-0.071428571428571,  -0.428571428571429,   0.500000000000000},
    { 0.500000000000000,  -0.428571428571429,  -0.071428571428571}
  },
  {
    {0.250000000000000,   0.500000000000000,   0.250000000000000},
    {-0.250000000000000,   0.500000000000000,  -0.250000000000000},
    {0.500000000000000,  -0.333333333333333,  -0.166666666666667}
  },
  {
    {0.161000000000000,   0.678000000000000,   0.161000000000000},
    {-0.095947556615018,  -0.404052443384982,   0.500000000000000},
    {0.500000000000000,  -0.404052443384982,  -0.095947556615018}
  },
  { // RGB to YUV EXT
    {0.175794546217068,   0.746416651985284,   0.077788801797648},
    {-0.095311435471474,  -0.404688564528526,   0.500000000000000},
    {0.500000000000000,  -0.452809823424115,  -0.047190176575885}
  },
  {  // RGB to XYZ EXT
    {0.602036117181740,   0.126890830837498,   0.221528979032433},
    {0.175794546217068,   0.746416651985284,   0.077788801797648},
    {-0.175794546217068,  -0.126890830837498,  1.391743127814445}
  },
  { // RGB709 to RGB EXT
    {0.577729863358965,   0.383620846616010,   0.016278641336825},
    {0.074034985428650,   0.9434012956405108,  -0.012642384061594},
    {0.149353242485433,   0.244045782506475,   0.645565896260605}
  },
  { // RGB2020 to RGB EXT
    {1.042250326793069,   0.008906997646749,  -0.045051787368051},
    {0.171509962343516,   0.873483503706327,  -0.033511688373427},
    {0.173919286133257,   0.131684893520626,   0.727298887127671}
  },
  {
    {0.743892327224535,   0.231347240210574,   0.011084709036028},
    {0.112507906240620,   0.905655168687551,  -0.011558592742523},
    {0.163869496838724,   0.154198908143321,   0.715395815125253}
  }
};



static const double INV_TRANSFORM[36][3][3] = {
  { // Identity transform
    {1.0, 0.0, 0.0},
    {0.0, 1.0, 0.0},
    {0.0, 0.0, 1.0}
  },
  { // YUV to RGB BT.709
    {  1.0000,    0.00000,    1.57480 },
    {  1.0000,   -0.18733,   -0.46813 },
    {  1.0000,    1.85563,    0.00000 }
  },
  { // YUV to RGB P3
    {  1.0000,    0.0000,    1.5808 },
    {  1.0000,   -0.1780,   -0.4592 },
    {  1.0000,    1.8620,    0.0000 }
  },
  { // YUV to RGB BT.2020
    {  1.0000,    0.00000,    1.47460 },
    {  1.0000,   -0.16455,   -0.57135 },
    {  1.0000,    1.88140,    0.00000 }
  },
  { // XYZ to RGB srgb (high precision)
    {  3.2404542, -1.5371385, -0.4985314 },
    { -0.9692660,  1.8760108,  0.0415560 },
    {  0.0556434, -0.2040259,  1.0572252 }
  },
  { // XYZ to RGB srgb (low precision)
    {  3.2406,   -1.5372,   -0.4987 },
    { -0.9689,    1.8757,    0.0415 },
    {  0.0557,   -0.2040,    1.0571 }
  },
  { // XYZ to RGB CIE
    {  2.3706743, -0.9000405, -0.4706338 },
    { -0.5138850,  1.4253036,  0.0885814 },
    {  0.0052982, -0.0146949,  1.0093968 }
  },
  { // XYZ to RGB Wide Gamut RGB
    {  1.4628067, -0.1840623, -0.2743606 },
    { -0.5217933,  1.4472381,  0.0677227 },
    {  0.0349342, -0.0968930,  1.2884099 }
  },
  { // XYZ to RGB SMPTE-C RGB
    {  3.5053960, -1.7394894, -0.5439640 },
    { -1.0690722,  1.9778245,  0.0351722 },
    {  0.0563200, -0.1970226,  1.0502026 }
  },
  { // XYZ to RGB S-Log
    {  1.5073998991, -0.2458221374, -0.1716116808 },
    { -0.5181517271,  1.3553912409,  0.1258786682 },
    {  0.0155116982, -0.0078727714,  0.9119163656 }
  },
  { // XYZ to RGB ACES RGB
    {  1.0498110175,  0.0000000000, -0.0000974845 },
    { -0.4959030232,  1.3733130458,  0.0982400361 },
    {  0.0000000000,  0.0000000000,  0.9912520182 }
  },
  { // XYZ to RGB P3D60
    {  2.4027414142, -0.8974841639, -0.3880533700 },
    { -0.8325796487,  1.7692317536,  0.0237127115 },
    {  0.0388233815, -0.0824996856,  1.0363685998 }
  },
  { // XYZ to RGB BT.2020
    {  1.716651, -0.355671, -0.253366},
    { -0.666684,  1.616481,  0.015768},
    {  0.017640, -0.042771,  0.942103}
    /*
    {  1.7166511880, -0.3556707838, -0.2533662814 },
    { -0.6666843518,  1.6164812366,  0.0157685458 },
    {  0.0176398574, -0.0427706133,  0.9421031212 } */
  },
  { // XYZ to RGB P3 DCI
    {  2.7253940305, -1.0180030062, -0.4401631952 },
    { -0.7951680258,  1.6897320548,  0.0226471906 },
    {  0.0412418914, -0.0876390192,  1.1009293786 }
  },
  { // XYZ to RGB P3D65
    {  2.493497,     -0.931384,     -0.402711     },
    { -0.829489,      1.762664,      0.023625     },
    {  0.035846,      0.076172,      0.956885     }
  },
  { // XYZ to RGB BT.709 (this has a difference in precision from the 5th entry above
    {  3.240970, -1.537383,  -0.498611 },
    { -0.969244,  1.875968,   0.041555 },
    {  0.055630, -0.203977,   1.056972 }
  },
  { // RGB BT.2020 to RGB BT.709 (16)
    {  1.660490254890140,  -0.587638564717282,  -0.072851975229213 },
    { -0.124550248621850,   1.132898753013895,  -0.008347895599309 },
    { -0.018151059958635,  -0.100578696221493,   1.118729865913540 }
  },
  { // RGB BT.2020 to RGB P3 D65 (17)
    {  1.343578941012770,  -0.282180098355451,  -0.061398440347996 },
    { -0.065298577056771,   1.075787625819622,  -0.010490208679671 },
    {  0.002821819575533,  -0.019597843054290,   1.016777195655686 }
  },
  { // RGB P3 D65 to RGB BT.709 (18)
    {   1.224939741445000,  -0.224939599120000,  -0.000001097215000 },
    {  -0.042056249524000,   1.042057784075000,   0.000000329788000 },
    {  -0.019637688845000,  -0.078636557327000,   1.098273664879000 }
  },
  {  // Y'Dz'Dx' to X'Y'Z' (SMPTE ST 2085 section 5) (19)
     //       Y'                D'z                  D'x
    { 0.991902000000000,   0.000000000000000,   2.000000000000000 },  // X' = 2* D'x - c2* Y'  where c2 = 0.991902
    { 1.000000000000000,   0.000000000000000,   0.000000000000000 },  // Y' = Y'
    { 1.013616929835409,   2.027233859670817,   0.000000000000000 }   // Z' = (2*D'z + Y')/c1  where c1 = 0.986566
  },
  // Place holder for constant luminance (does not work as a single matrix ...
  { // ITU-R BT.2020 constant luminance --- similar to case (3) above. (20)
    //        R                   G                    B
    { 0.000000000000000,   1.000000000000000,   0.000000000000000 },  //
    { 0.000000000000000,   0.000000000000000,   0.000000000000000 },  //
    { 0.000000000000000,   0.000000000000000,   0.000000000000000 },  //
  },
  { // YUV P3 D65 to RGB P3 D65 (approximate) - (21)
    //{ 1.0000,  -0.024589,  1.536518 },
    //{ 1.0000,  -0.202480, -0.503507 },
    //{ 1.0000,   1.837546, -0.044502 },
    {1.000000000000000,                   0,   1.542050871860503 },
    {1.000000000000000,  -0.211063854455947,  -0.510439154407954 },
    {1.000000000000000,   1.841426171812510,   0.000000000000000 }
  },
  { // RGB BT.2020 to RGB P3 D60 (22)
    { 1.468532141150479,  -0.308422395067619,  -0.067108892668074},
    {-0.062596024581556,   1.031275913884886,  -0.010055951532639},
    { 0.003246582318521,  -0.022548410685530,   1.169837359498846},
  },
  { // RGB P3 D60 to BT.709 (23)
    {  1.120712328737458,  -0.234646928643142,  -0.000001189898782},
    { -0.038478187476584,   1.087032952904620,   0.000000766555928},
    { -0.017966867475328,  -0.082030245165852,   0.954576306774709},
  },
  { // YUV P3 D60 to RGB P3 D60 (approximate) - 24
    {1.000000000000000,                   0,   1.524950920818921 },
    {1.000000000000000,  -0.196689277522663,  -0.525454351842566 },
    {1.000000000000000,   1.853715995138824,                   0 }
  },
  { // AMT - 25
    /*
    { 1.0000,    0,         1.6000},
    { 1.0000,   -0.1267,   -0.4267},
    { 1.0000,    1.9000,    0},
  
    {1.0000,    0,       1.6500},
    {  1.0000,   -0.06094,   -0.36094},
    {  1.0000,    1.9500,         0},
    
    { 1.000000000000000,   0.000000000000000,   1.600000000000000},
    {  1.000000000000000,  -0.257142857142857,  -0.457142857142857},
    {  1.000000000000000,   1.800000000000000,   0.000000000000000},
     
    {1.000000000000000,                   0,   1.500000000000000},
    {  1.000000000000000,  -0.276923076923077,  -0.576923076923077},
    {  1.000000000000000,   1.800000000000000,                   0}
     
    {1.000000000000000,                   0,   1.500000000000000},
    { 1.000000000000000,  -0.135714285714286,  -0.535714285714286},
    {  1.000000000000000,   1.900000000000000,  -0.000000000000000},
     */
    /*{   1.000000000000000,                   0,   1.400000000000000},
    {1.000000000000000,  -0.300000000000000,  -0.700000000000000},
    {1.000000000000000,   1.800000000000000,                   0},*/
    {1.000000000000000,                   0,   1.500000000000000},
    {1.000000000000000,  -0.750000000000000,  -0.750000000000000},
    {1.000000000000000,   1.500000000000000,                   0}
  },
  { // YCoCg - 26
    { 1.0, -1.0,  1.0},
    { 1.0,  1.0,  0.0},
    { 1.0, -1.0, -1.0}
  },
  { // 27
    {1.000000000000000,   0.000000000000000,   1.700000000000000},
    {1.000000000000000,  -0.364285714285714,  -0.364285714285714},
    {1.000000000000000,   1.700000000000000,                   0}
  },
  {
    {1.000000000000000,                   0,   1.750000000000000},
    {1.000000000000000,  -0.291666666666667,  -0.291666666666667},
    {1.000000000000000,   1.750000000000000,                   0}
  },
  {
    {1.000000000000000,                  0,   1.500000000000000},
    {1.000000000000000,   1.000000000000000,                   0},
    {1.000000000000000,  -2.000000000000000,  -1.500000000000000}
  },
  {
    {1.000000000000000,   0.000000000000000,   1.678000000000000},
    {1.000000000000000,  -0.398463126843658,  -0.398463126843658},
    {1.000000000000000,   1.678000000000000,   0.000000000000000}
  },
  { // EXT YUV to RGB
    {1.000000000000000,                   0,   1.648410907565864},
    {1.000000000000000,  -0.192218927382528,  -0.388230416221368},
    {1.000000000000000,   1.844422396404704,                   0}
  },
  { // EXT XYZ to RGB
    { 1.684113963357831,  -0.328747024117586,  -0.249692060863265},
    {-0.414866520100502,   1.408108123688060,  -0.012667630080784},
    { 0.174899582889521,   0.086858036790653,   0.685829211069835}
  },
  {
    { 1.840514447001081,  -0.732701606826046,  -0.060759358419057},
    {-0.149386826105148,   1.114121775408967,   0.025585257886749},
    {-0.369334218255630,  -0.251663479552585,   1.553413274645461}
  },
  {
    { 0.952797770099054,  -0.018485134621291,   0.058168347595533},
    {-0.194473866989056,   1.140716526786005,   0.040514212224643},
    {-0.192631447699981,  -0.202118022660768,   1.353705210076763},
  },
  { // RGB EXT to RGB D65
    { 1.403861990395281,  -0.353935649511439,  -0.027470666314414},
    {-0.178013642694086,   1.146023899361818,   0.021274450643752},
    {-0.283200774138676,  -0.165945026066769,   1.399534475933237}
  }
};

#endif
