/* The copyright in this software is being made available under the BSD
* License, included below. This software may be subject to other third party
* and contributor rights, including patent rights, and no such rights are
* granted under this license.
*
* <OWNER> = Apple Inc.
* <ORGANIZATION> = Apple Inc.
* <YEAR> = 2014
*
* Copyright (c) 2014, Apple Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  * Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*  * Neither the name of the <ORGANIZATION> nor the names of its contributors may
*    be used to endorse or promote products derived from this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
*************************************************************************************
* \file Reshaper.cpp
*
* \brief
*    Class for reshaper application
*
* \author
*     - Alexis Michael Tourapis         <atourapis@apple.com>
*
*************************************************************************************
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------
#include <list>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include "Global.H"
#include "Reshaper.H"
#if TCH_CROSSPLANE_CRCB
#include "CrossPlaneFunctions.H"
#if ARRS_MODEL_OPTIMIZATION
#include "CrossPlaneReshaper.H"
#endif
#endif
#if DI_RESHAPER_m37269_m37267
#if 1
#define round(x)        (int) floor((float)x+0.5)
#endif

#if MODE0_UPDATE
typedef int Int;
#define INVERSE_FRAC_PRECISION 9

void identifyComponentRangeIdx(int sampleVal, Int shiftVal, std::vector<fixedPrec> &rangeValues,  Int &idxS)
{
  Int shiftedSampleVal = sampleVal << shiftVal;
  if( shiftedSampleVal < rangeValues[0].val ) {
    idxS = 0;
  } else if( shiftedSampleVal > rangeValues.back().val ) {
    idxS = rangeValues.size();
  } else {
    for(idxS = 1; idxS < rangeValues.size(); idxS++) {
      if( (shiftedSampleVal > rangeValues[idxS-1].val) && (shiftedSampleVal <= rangeValues[idxS].val) )
        break;
    }
  }
}
void getMinMaxSampleVal(const int reshaperRangeId, int const bitDepth, int *minSampleVal, int *maxSampleVal, int *deltaSampleVal) {
  switch(reshaperRangeId) {
  case 0:
    for(int c = 0; c < 3; c++) {
      minSampleVal[c] = 0;
      maxSampleVal[c] = (1 << bitDepth) - 1;
    }
    break;
  case 1:
    for(int c = 0; c < 3; c++) {
      minSampleVal[c] = 16 << (bitDepth - 8);
      if(c == 0)
        maxSampleVal[c] = minSampleVal[c] + (219 << (bitDepth - 8));
      else
        maxSampleVal[c] = minSampleVal[c] + (224 << (bitDepth - 8));
    }
    break;
  default:
    printf("Check range; not supported yet.\n");
    break;
  } // End of switch
  
  for(int c = 0; c < 3; c++)
    deltaSampleVal[c] = maxSampleVal[c] - minSampleVal[c];

}
fixedPrec getSmallestVals(int const shiftVal, int const bitDepth, int const *minSampleVal, int const c)
{
  int shiftOffset, smallestVal;
  if(shiftVal > 0)
  {
    shiftOffset = (1 << (shiftVal - 1));
    smallestVal = (minSampleVal[c] + shiftOffset) >> shiftVal;
  }
  else
  {
    shiftOffset = 0;
    smallestVal = minSampleVal[c] << abs(shiftVal);
  }
  fixedPrec smallestValue = fixedPrec( smallestVal, bitDepth - shiftVal);
  return smallestValue;
}
#endif
//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
TempMatrices::TempMatrices(int nSegments, int nParams, int nSamples)
{

  /* set members */
  m_nSegments = nSegments;
  m_nParams   = nParams;
  m_nSamples  = nSamples;

  /* memory allocation for 2-D matrices*/
  Param2A = new double*[nSegments];
  Param2B = new double*[nSegments];
  Param2C = new double*[nSegments];
  for (int i = 0; i < nSegments; i++)
  {
    Param2A[i] = new double[nParams];
    Param2B[i] = new double[nParams];
    Param2C[i] = new double[nParams];
  }
  for (int i = 0; i < nSegments; i++){
    for (int j = 0; j < nParams; j++){
      Param2A[i][j] = 0;
      Param2B[i][j] = 0;
      Param2C[i][j] = 0;
    }
  }

  Param2IntermediateB = new double*[nSegments + 1];
  for (int i = 0; i < nSegments + 1; i++)
    Param2IntermediateB[i] = new double[nParams];
  for (int i = 0; i < nSegments + 1; i++){
    for (int j = 0; j < nParams; j++){
      Param2IntermediateB[i][j] = 0;
    }
  }

  IntermediateB2C = new double*[nSegments];
  for (int i = 0; i < nSegments; i++)
    IntermediateB2C[i] = new double[nSegments + 1];
  for (int i = 0; i < nSegments; i++){
    for (int j = 0; j < nSegments + 1; j++){
      IntermediateB2C[i][j] = 0;
    }
  }

  Param2StackedABC = new double*[3 * nSegments];
  for (int i = 0; i < 3 * nSegments; i++)
    Param2StackedABC[i] = new double[nParams];
  for (int i = 0; i < 3 * nSegments; i++){
    for (int j = 0; j < nParams; j++){
      Param2StackedABC[i][j] = 0;
    }
  }

  StackedABC2InterleavedABC = new double*[3 * nSegments];
  for (int i = 0; i < 3 * nSegments; i++)
    StackedABC2InterleavedABC[i] = new double[3 * nSegments];
  for (int i = 0; i < 3 * nSegments; i++){
    for (int j = 0; j < 3 * nSegments; j++){
      StackedABC2InterleavedABC[i][j] = 0;
    }
  }

  InterleavedABC2Y = new double*[nSamples];
  for (int i = 0; i < nSamples; i++)
    InterleavedABC2Y[i] = new double[3 * nSegments];
  for (int i = 0; i < nSamples; i++){
    for (int j = 0; j < 3 * nSegments; j++){
      InterleavedABC2Y[i][j] = 0;
    }
  }

  Param2Y = new double*[nSamples];
  for (int i = 0; i < nSamples; i++)
    Param2Y[i] = new double[nParams];
  for (int i = 0; i < nSamples; i++){
    for (int j = 0; j < nParams; j++){
      Param2Y[i][j] = 0;
    }
  }

  Param2Y_t = new double*[nParams];
  for (int i = 0; i < nParams; i++)
    Param2Y_t[i] = new double[nSamples];
  for (int i = 0; i < nParams; i++){
    for (int j = 0; j < nSamples; j++){
      Param2Y_t[i][j] = 0;
    }
  }

  M = new double*[nSamples];
  for (int i = 0; i < nSamples; i++)
    M[i] = new double[nParams - 1];
  for (int i = 0; i < nSamples; i++){
    for (int j = 0; j < nParams - 1; j++){
      M[i][j] = 0;
    }
  }

  M_t = new double*[nParams - 1];
  for (int i = 0; i < nParams - 1; i++)
    M_t[i] = new double[nSamples];
  for (int i = 0; i < nParams - 1; i++){
    for (int j = 0; j < nSamples; j++){
      M_t[i][j] = 0;
    }
  }

  M_txM = new double*[nParams - 1];
  for (int i = 0; i < nParams - 1; i++)
    M_txM[i] = new double[nParams - 1];
  for (int i = 0; i < nParams - 1; i++){
    for (int j = 0; j < nParams - 1; j++){
      M_txM[i][j] = 0;
    }
  }

  Param2Y_txParam2Y = new double*[nParams];
  for (int i = 0; i < nParams; i++)
    Param2Y_txParam2Y[i] = new double[nParams];
  for (int i = 0; i < nParams; i++){
    for (int j = 0; j < nParams; j++){
      Param2Y_txParam2Y[i][j] = 0;
    }
  }

  temp1 = new double*[nSamples];
  for (int i = 0; i < nSamples; i++)
    temp1[i] = new double[3 * nSegments];
  for (int i = 0; i < nSamples; i++){
    for (int j = 0; j < 3 * nSegments; j++){
      temp1[i][j] = 0;
    }
  }

  /* memory allocation for 1-D vectors */
  SegmentLowerEndPoints = new double[nSegments];
  for (int i = 0; i < nSegments; i++)
    SegmentLowerEndPoints[i] = 0;

  Y1 = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    Y1[i] = 0;

  Y2 = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    Y2[i] = 0;

  M_txY2 = new double[nParams - 1];
  for (int i = 0; i < nParams - 1; i++)
    M_txY2[i] = 0;

  ParamOpt_temp = new double[nParams - 1];
  for (int i = 0; i < nParams - 1; i++)
    ParamOpt_temp[i] = 0;

  ParamOpt = new double[nParams];
  for (int i = 0; i < nParams; i++)
    ParamOpt[i] = 0;

  Param2Y_txlut = new double[nParams];
  for (int i = 0; i < nParams; i++)
    Param2Y_txlut[i] = 0;

  V = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    V[i] = 0;

  V_t = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    V_t[i] = 0;

  Y = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    Y[i] = 0;

  pred = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    pred[i] = 0;

  __x = new double[nSegments];
  for (int i = 0; i < nSegments; i++)
    __x[i] = 0;

  x = new double[nSegments + 1];
 
  for (int i = 0; i < nSegments; i++)
    x[i] = 0;

  error = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    error[i] = 0;

  Param2YxParamOpt = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    Param2YxParamOpt[i] = 0;

  final_pred = new double[nSamples];
  for (int i = 0; i < nSamples; i++)
    final_pred[i] = 0;

}

TempMatrices::~TempMatrices()
{

  /* delete 2-D matrices */
  for (int i = 0; i < m_nSegments; i++)
  {
    delete[] Param2A[i];
    delete[] Param2B[i];
    delete[] Param2C[i];
  }
  delete[] Param2A;
  delete[] Param2B;
  delete[] Param2C;

  for (int i = 0; i < m_nSegments + 1; i++)
    delete[] Param2IntermediateB[i];
  delete[] Param2IntermediateB;

  for (int i = 0; i < m_nSegments; i++)
    delete[] IntermediateB2C[i];
  delete[] IntermediateB2C;

  for (int i = 0; i < 3 * m_nSegments; i++)
    delete[] Param2StackedABC[i];
  delete[] Param2StackedABC;

  for (int i = 0; i < 3 * m_nSegments; i++)
    delete[] StackedABC2InterleavedABC[i];
  delete[] StackedABC2InterleavedABC;

  for (int i = 0; i < m_nSamples; i++)
    delete[] InterleavedABC2Y[i];
  delete[] InterleavedABC2Y;

  for (int i = 0; i < m_nSamples; i++)
    delete[] Param2Y[i];
  delete[] Param2Y;

  for (int i = 0; i < m_nParams; i++)
    delete[] Param2Y_t[i];
  delete[] Param2Y_t;

  for (int i = 0; i < m_nSamples; i++)
    delete[] M[i];
  delete[] M;

  for (int i = 0; i < m_nParams - 1; i++)
    delete[] M_t[i];
  delete[] M_t;

  for (int i = 0; i < m_nParams - 1; i++)
    delete[] M_txM[i];
  delete[] M_txM;

  for (int i = 0; i < m_nParams; i++)
    delete[] Param2Y_txParam2Y[i];
  delete[] Param2Y_txParam2Y;

  for (int i = 0; i < m_nSamples; i++)
    delete[] temp1[i];
  delete[] temp1;

  /* delete 1-D vectors */
  delete[]   SegmentLowerEndPoints;
  delete[]   Y1;
  delete[]   Y2;
  delete[]   M_txY2;
  delete[]   ParamOpt_temp;
  delete[]   ParamOpt;
  delete[]   Param2Y_txlut;
  delete[]   V;
  delete[]   V_t;
  delete[]   Y;
  delete[]   pred;
  delete[]   __x;
  delete[]   x;
  delete[]   error;
  delete[]   Param2YxParamOpt;
  delete[]   final_pred;

}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------
Reshaper::Reshaper() {
#if QC_CBCR_m37064
  m_chromaHistScaleCb	= 1.0;
  m_chromaHistScaleCr	= 1.0;
  m_chromaHistScaleCbInv= 1.0;
  m_chromaHistScaleCrInv= 1.0;
  m_chromaHistOffsetCb	= 0.0;
  m_chromaHistOffsetCr	= 0.0;
  m_flagScaleChromaHist = FALSE;
#endif
  m_numOfScenes = 1;
  m_sceneNum = 0;
#if !MODE0_UPDATE
  m_sceneData = NULL; 
#endif
  m_reshapeData = NULL;
  memset(m_inverseLumaReshapingLUT, 0, MAX_LUMA_RESHAPING_LUT_SIZE);
#if ARRS_MODEL_OPTIMIZATION
  memset(m_forwardLumaReshapingLUT, 0, MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE);
#else
  memset(m_forwardLumaReshapingLUT, 0, MAX_LUMA_RESHAPING_LUT_SIZE);
#endif
  memset(m_inverseChromaReshapingLUT[0], 0, MAX_LUMA_RESHAPING_LUT_SIZE);
  memset(m_inverseChromaReshapingLUT[1], 0, MAX_LUMA_RESHAPING_LUT_SIZE);
#if ARRS_MODEL_OPTIMIZATION
  memset(m_forwardChromaReshapingLUT[0], 0, MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE);
  memset(m_forwardChromaReshapingLUT[1], 0, MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE);
#else
  memset(m_forwardChromaReshapingLUT[0], 0, MAX_LUMA_RESHAPING_LUT_SIZE);
  memset(m_forwardChromaReshapingLUT[1], 0, MAX_LUMA_RESHAPING_LUT_SIZE);
#endif
  m_peakLum = 0.f;
}

Reshaper::~Reshaper() {
#if !MODE0_UPDATE
  if (m_sceneData != NULL)
  {
    delete [] m_sceneData;
    m_sceneData = NULL;
  }
#endif
  if (m_reshapeData != NULL)
  {
    delete [] m_reshapeData;
    m_reshapeData = NULL;
  }
}


//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------
/* sum over the elements of vector vec */
double Reshaper::sumVec(vector<double> vec, int beg_idx, int end_idx)
{
  double sum = 0;
  for (int i = beg_idx; i <= end_idx; i++)
    sum += vec[i];
  return sum;
};

/* multiply 2 matrixes a and b and put the result in c. */
void Reshaper::mulMatMat(double** a, double** b, double** c, int a_rows, int b_cols, int a_cols)
{
  for (int row = 0; row < a_rows; row++){
    for (int col = 0; col < b_cols; col++){
      c[row][col] = 0;
      for (int i = 0; i < a_cols; i++)
        c[row][col] += a[row][i] * b[i][col];
    }
  }
}

/* multiply 2-D matrix a with vector b and put the result in c */
void Reshaper::mulMatVec(double** a, double* b, double* c, int a_rows, int a_cols)
{
  for (int row = 0; row < a_rows; row++){
    c[row] = 0;
    for (int i = 0; i < a_cols; i++)
      c[row] += a[row][i] * b[i];
  }
}

/* compute dot product of two vectors a and b. */
double Reshaper::dotProd(double* a, double* b, int size)
{
  double dPrd = 0;
  for (int i = 0; i < size; i++)
    dPrd += a[i] * b[i];

  return dPrd;
}

/* compute transpose of matrix x and put the result in trans*/
void Reshaper::transpose(double** x, double** trans, int rows, int cols)
{
  for (int i = 0; i < cols; i++)
  {
    for (int j = 0; j < rows; j++)
      trans[i][j] = x[j][i];
  }
}

/*
LUFactorization() borrowed from HM
solves A * X = Y
A_txA = A_t * A
A_txY = A_t * Y
*/
void Reshaper::LUFactorization(double** A_txA, double* A_txY, double* X, int N)
{
  int i, j, a;
  double*  z     = new double [N];
  double** alpha = new double*[N];
  double** beta  = new double*[N];

  for (int i = 0; i < N; i++)
  {
    alpha[i] = new double[N];
    beta[i]  = new double[N];
  }
  for (int i = 0; i < N; i++)
  {
    for (int j = 0; j < N; j++)
    {
      alpha[i][j] = 0;
      beta [i][j] = 0;
    }
    z[i] = 0;
  }


  //1.Step : LU Decomposition
  for (j = 0; j < N; j++)
  {
    for (i = 0; i < N; i++)
    {
      beta[i][j] = 0;
      alpha[i][j] = i == j ? 1 : 0;
    }
    z[j] = 0;
    X[j] = 0;
  }
  for (j = 0; j < N; j++)
  {
    for (i = 0; i <= j; i++)
    {
      beta[i][j] = A_txA[i][j];
      for (a = 0; a < i; a++)
      {
        beta[i][j] -= alpha[i][a] * beta[a][j];
      }
    }
    for (i = j + 1; i < N; i++)
    {
      alpha[i][j] = A_txA[i][j];
      for (a = 0; a < j; a++)
      {
        alpha[i][j] -= alpha[i][a] * beta[a][j];
      }
      alpha[i][j] = alpha[i][j] / beta[j][j];
    }
  }
  //2. Step: Solving L*z = b, forward substitution L*U*x=b, U*x=z
  z[0] = A_txY[0];
  for (i = 1; i < N; i++)
  {
    z[i] = A_txY[i];
    for (j = 0; j < i; j++)
    {
      z[i] -= alpha[i][j] * z[j];
    }
  }
  //3. Step: Solving U*x = z, backsubstitution
  X[N - 1] = z[N - 1] / beta[N - 1][N - 1];
  for (i = N - 2; i >= 0; i--)
  {
    X[i] = z[i];
    for (j = i + 1; j < N; j++)
    {
      X[i] -= beta[i][j] * X[j];
    }
    X[i] = X[i] / beta[i][i];
  }

  for (int i = 0; i < N; i++)
  {
    delete[] alpha[i];
    delete[] beta [i];
  }
  delete[] alpha;
  delete[] beta;
  delete[] z;

}

/**
- Comute MSE optimal piecewise model parameters given target LUT and locations of pivot point segments

\param   lut                 lut to fit
\param   lut_len             length of LUT in number of samples
\param   SegmentLengths      vector of segment lengths
\param   smoothness_flag     flag indicating piecewise linear or smooth model fit
\param   force_initial_value_flag      flag to force fit to achieve first value specified in LUT
\param   nFractionBitsA      number of fraction bits used for constant model coefficients
\param   nFractionBitsB      number of fraction bits used for linear model coefficient
\retval  A0S                 optimal values for constant model coefficients on each segment and the final endpoint
\retval  B0                  optimal values for linear model coefficient on first segment
\retval  mse                 mean square error of optimal fit
\retval  pUnifiedLumaReshapeInfo  syntax describing model and full set of model parameters
*/
void Reshaper::MinimalMSE_Fit
  (
  double*          lut,
  int              lut_len,
  vector<double>   SegmentLengths,
  bool             smoothness_flag,
  bool             force_initial_value_flag,
  int              nFractionBitsA,
  int              nFractionBitsB,
  TempMatrices*    cMatrices,
  vector<double>&  A0S,
  double&          B0,
  double&          mse
  )
{

  int nSamples   = cMatrices->m_nSamples;
  int nSegments  = cMatrices->m_nSegments;
  int nParams    = cMatrices->m_nParams;
  int MinLUT     = 0;
  int MaxLUT     = nSamples - 1;

  // guard against zero length segments
  bool hasZeroLengthSegment = false;
  for (int i = 0; i < nSegments; i++)
    if(SegmentLengths[i] <=0)
      hasZeroLengthSegment = true;
  if(hasZeroLengthSegment)
  {
    for(int i=0;i<nSegments+1;i++)
      A0S[i] = 0;
    B0 = 0;
    mse = 10000000;
    return;
  }

  vector<double>	ModifiedSegmentLengths = SegmentLengths;
  ModifiedSegmentLengths[nSegments - 1] = nSamples - sumVec(SegmentLengths, 0, nSegments - 2);

  for (int s = 1; s <= nSegments - 1; s++)
    cMatrices->SegmentLowerEndPoints[s] = cMatrices->SegmentLowerEndPoints[s - 1] + SegmentLengths[s - 1];


  for (int i = 0; i < nSegments; i++)
    cMatrices->__x[i] = cMatrices->SegmentLowerEndPoints[i];
  for (int i = 0; i < nSegments + 1; i++)
    cMatrices->x[i] = cMatrices->SegmentLowerEndPoints[i];


  cMatrices->x[nSegments] = cMatrices->__x[nSegments - 1] + SegmentLengths[nSegments - 1];
  for (int i = 0; i < nSegments; i++)
    cMatrices->x[i] = cMatrices->__x[i];

  for (int i = 0; i <= nSegments - 1; i++)
    cMatrices->Param2A[i][i] = 1;

  if (smoothness_flag)
  {
    int i, j;
    cMatrices->Param2IntermediateB[0][nParams - 1] = 1;

    for (i = 0; i <= nSegments - 1; i++)
    {
      cMatrices->Param2IntermediateB[i + 1][0] = 2 * pow((double)-1, i + 1) / (cMatrices->x[1] - cMatrices->x[0]);

      for (j = 0; j <= i - 1; j++)
        cMatrices->Param2IntermediateB[i + 1][j + 1] = 2 * (pow((double)-1, i - j))*(1 / (cMatrices->x[j + 1] - cMatrices->x[j]) +
        1 / (cMatrices->x[j + 2] - cMatrices->x[j + 1]));

      cMatrices->Param2IntermediateB[i + 1][i + 1] = 2 / (cMatrices->x[i + 1] - cMatrices->x[i]);
      j = nSegments;
      cMatrices->Param2IntermediateB[i + 1][j + 1] = pow((double)-1, i + 1);
    }

    for (i = 0; i <= nSegments - 1; i++)
    {
      for (j = 0; j < nParams; j++)
        cMatrices->Param2B[i][j] = cMatrices->Param2IntermediateB[i][j];
    }

    for (i = -1; i <= nSegments - 2; i++)
    {
      j = i;
      cMatrices->IntermediateB2C[i + 1][j + 1] = -1 / (2 * ((cMatrices->x[i + 2] - cMatrices->x[i + 1])));
      j = i + 1;
      cMatrices->IntermediateB2C[i + 1][j + 1] = 1 / (2 * ((cMatrices->x[i + 2] - cMatrices->x[i + 1])));
    }

    mulMatMat(cMatrices->IntermediateB2C, cMatrices->Param2IntermediateB, cMatrices->Param2C, nSegments, nParams, nSegments + 1);

  }
  else
  {
    for (int i = 0; i <= nSegments - 1; i++)
    {
      cMatrices->Param2B[i][i] = -1 / SegmentLengths[i];
      cMatrices->Param2B[i][i + 1] = 1 / SegmentLengths[i];
    }
  }

  for (int j = 0; j <= nSegments - 1; j++)
  {
    for (int col = 0; col < nParams; col++)
    {
      cMatrices->Param2StackedABC[0 * nSegments + j][col] = cMatrices->Param2A[j][col];
      cMatrices->Param2StackedABC[1 * nSegments + j][col] = cMatrices->Param2B[j][col];
      cMatrices->Param2StackedABC[2 * nSegments + j][col] = cMatrices->Param2C[j][col];
    }
  }

  for (int s = 0; s <= nSegments - 1; s++)
  {
    cMatrices->StackedABC2InterleavedABC[3 * s + 0][0 * nSegments + s] = 1;
    cMatrices->StackedABC2InterleavedABC[3 * s + 1][1 * nSegments + s] = 1;
    cMatrices->StackedABC2InterleavedABC[3 * s + 2][2 * nSegments + s] = 1;
  }

  int idx = -1;
  for (int i = 0; i <= nSegments - 1; i++)
  {
    for (int j = 0; j <= ModifiedSegmentLengths[i] - 1; j++)
    {
      idx = idx + 1;

      cMatrices->InterleavedABC2Y[idx][3 * (i - 1) + 1 + 3 - 1] = 1;
      cMatrices->InterleavedABC2Y[idx][3 * (i - 1) + 2 + 3 - 1] = j;
      cMatrices->InterleavedABC2Y[idx][3 * (i - 1) + 3 + 3 - 1] = j * j;
    }
  }

  mulMatMat(cMatrices->InterleavedABC2Y, cMatrices->StackedABC2InterleavedABC, cMatrices->temp1, nSamples, 3 * nSegments, 3 * nSegments);
  mulMatMat(cMatrices->temp1, cMatrices->Param2StackedABC, cMatrices->Param2Y, nSamples, nParams, 3 * nSegments);

  if (force_initial_value_flag)
  {
    double A0 = lut[0];

    for (int i = 0; i < nSamples; i++)
      cMatrices->Y1[i] = cMatrices->Param2Y[i][0] * A0;

    for (int i = 0; i < nSamples; i++)
      cMatrices->Y2[i] = lut[i] - cMatrices->Y1[i];

    for (int i = 0; i < nSamples; i++) {
      for (int j = 0; j < nParams - 1; j++) {
        cMatrices->M[i][j] = cMatrices->Param2Y[i][j + 1];
      }
    }

    transpose(cMatrices->M, cMatrices->M_t, nSamples, nParams - 1);

    mulMatMat(cMatrices->M_t, cMatrices->M, cMatrices->M_txM, nParams - 1, nParams - 1, nSamples);    // M_txM = M_t * M

    mulMatVec(cMatrices->M_t, cMatrices->Y2, cMatrices->M_txY2, nParams - 1, nSamples);

    /* solving M * ParamOpt_temp = Y2 uisng HM function  LUFactorization() */
    LUFactorization(cMatrices->M_txM, cMatrices->M_txY2, cMatrices->ParamOpt_temp, nParams - 1);

    cMatrices->ParamOpt[0] = A0;
    for (int i = 1; i < nParams; i++)
      cMatrices->ParamOpt[i] = cMatrices->ParamOpt_temp[i - 1];
  }
  else
  {
    transpose(cMatrices->Param2Y, cMatrices->Param2Y_t, nSamples, nParams);

    mulMatMat(cMatrices->Param2Y_t, cMatrices->Param2Y, cMatrices->Param2Y_txParam2Y, nParams, nParams, nSamples);

    mulMatVec(cMatrices->Param2Y_t, lut, cMatrices->Param2Y_txlut, nParams, nSamples);

    LUFactorization(cMatrices->Param2Y_txParam2Y, cMatrices->Param2Y_txlut, cMatrices->ParamOpt, nParams);
  }

  for (int i = 0; i <= nParams - 2; i++)
    cMatrices->ParamOpt[i] = round(cMatrices->ParamOpt[i] * (1 << nFractionBitsA)) / ((double)(1<<nFractionBitsA));
#if MODE0_UPDATE
  double cint, frac,cfloat;
  for (int i = 0; i <= nParams - 2; i++)
  {
    cfloat = cMatrices->ParamOpt[i];
    cint = (1 << nFractionBitsA) * floor(cfloat/(1 << nFractionBitsA));
    frac = cfloat - cint;
    cint = dClip(cint, MIN_INT_CONSTANT_COEFFICIENT, MAX_INT_CONSTANT_COEFFICIENT);
    cMatrices->ParamOpt[i] = cint + frac;
  }
#endif
  if (smoothness_flag)
  {
    for (int i = 0; i < nSamples; i++)
      cMatrices->V[i] = cMatrices->Param2Y[i][nParams - 1];

    cMatrices->ParamOpt[nParams - 1] = 0;

    mulMatVec(cMatrices->Param2Y, cMatrices->ParamOpt, cMatrices->Param2YxParamOpt, nSamples, nParams);

    for (int i = 0; i < nSamples; i++)
      cMatrices->Y[i] = lut[i] - cMatrices->Param2YxParamOpt[i];


    for (int i = 0; i < nSamples; i++)
      cMatrices->V_t[i] = cMatrices->V[i];


    double b0 = (1 / dotProd(cMatrices->V, cMatrices->V_t, nSamples)) *  dotProd(cMatrices->V_t, cMatrices->Y, nSamples);

    cMatrices->ParamOpt[nParams - 1] = round(b0 * (1<< nFractionBitsB)) / ((double)(1<<nFractionBitsB));
#if MODE0_UPDATE
    cMatrices->ParamOpt[nParams - 1] = dClip(cMatrices->ParamOpt[nParams - 1], MIN_LINEAR_COEFFICIENT, MAX_LINEAR_COEFFICIENT);
#endif
    for (int i = 0; i <= nParams - 2; i++)
      A0S[i] = cMatrices->ParamOpt[i];
    B0 = cMatrices->ParamOpt[nParams - 1];
  }
  else
  {
    for (int i = 0; i <= nParams - 1; i++)
      A0S[i] = cMatrices->ParamOpt[i];
    B0 = 0;
  }

  mulMatVec(cMatrices->Param2Y, cMatrices->ParamOpt, cMatrices->pred, nSamples, nParams);

  vector<double>min_MaxLUT_lut(nSamples);
  for (int t = 0; t < (int)min_MaxLUT_lut.size(); t++)
    min_MaxLUT_lut[t] = lut[t];
  setNewUpperBound(min_MaxLUT_lut, MaxLUT);

  vector<double> final_lut(nSamples);
  for (int t = 0; t < (int) final_lut.size(); t++)
    final_lut[t] = min_MaxLUT_lut[t];
  setNewLowerBound(final_lut, MinLUT);

  vector<double> min_MaxLUT_pred(nSamples);
  for (int t = 0; t < (int) min_MaxLUT_pred.size(); t++)
    min_MaxLUT_pred[t] = cMatrices->pred[t];
  setNewUpperBound(min_MaxLUT_pred, MaxLUT);

  vector<double> max_MinLUT_min_MaxLUT_pred(nSamples);
  for (int t = 0; t < (int) max_MinLUT_min_MaxLUT_pred.size(); t++)
    max_MinLUT_min_MaxLUT_pred[t] = min_MaxLUT_pred[t];
  setNewLowerBound(max_MinLUT_min_MaxLUT_pred, MinLUT);

  for (int i = 0; i < nSamples; i++)
    cMatrices->final_pred[i] = round(max_MinLUT_min_MaxLUT_pred[i]);

  for (int i = 0; i < nSamples; i++)
    cMatrices->error[i] = round(final_lut[i] - cMatrices->final_pred[i]);

  mse = 0;
  for (int i = 0; i < nSamples; i++)
    mse += cMatrices->error[i] * cMatrices->error[i];

  mse = mse / nSamples;
}

double  Reshaper::minVec(vector<double>& vec)
{
  double min_val = vec[0];
  for (int i = 0; i < (int) vec.size(); i++)
  {
    if (vec[i] < min_val)
      min_val = vec[i];
  }
  return min_val;
}

double  Reshaper::maxVec(vector<double>& vec)
{
  double max_val = vec[0];
  for (int i = 0; i < (int) vec.size(); i++)
  {
    if (vec[i] > max_val)
      max_val = vec[i];
  }
  return max_val;
}

void Reshaper::setNewUpperBound(vector<double>& vec, double upBnd)
{
  for (int i = 0; i < (int) vec.size(); i++){
    if (vec[i] > upBnd)
      vec[i] = upBnd;
  }
}

void Reshaper::setNewLowerBound(vector<double>& vec, double loBnd)
{
  for (int i = 0; i < (int) vec.size(); i++){
    if (vec[i] < loBnd)
      vec[i] = loBnd;
  }
}

void Reshaper::blockproc(const double *input, int inp_wid, int inp_hei, SceneBlkData sceneBLK) 
{
  int size = sceneBLK.blockSize;
  int wid = sceneBLK.blockWidth;
  int hei = sceneBLK.blockHeight;
  double *block;
  int block_size, block_wid, block_hei;
  int wid_left = size * wid - inp_wid;    // 0
  int hei_left = size * hei - inp_hei;	// 8
  int i, j, i_pos, j_pos;
  double diff = 0.0, std = 0.0, mean = 0.0, min, max;

  ///////////// anaylze scene block data //////////////
  for (i_pos=0; i_pos<hei; i_pos++)
  {
    for (j_pos=0; j_pos<wid; j_pos++)
    {
      if(i_pos==hei-1)
        block_hei = size-hei_left;
      else
        block_hei = size;
      if(j_pos==wid-1)
        block_wid = size-wid_left;
      else
        block_wid = size;
      block_size = block_wid * block_hei;

      block = new double[block_size];

      // read one 16x16 block
      for (i=0; i<block_hei; i++)
      {
        for (j=0; j<block_wid; j++)
        {
          block[i*block_wid+j] = input[(i_pos*size+i) * inp_wid + (j_pos*size+j)];
        }
      }

      // calculating block statitics
      min = block[0];
      max = block[0];
      mean= 0.0;
      for (i=0; i<block_hei; i++)
      {
        for (j=0; j<block_wid; j++)
        {
          if (block[i*block_wid+j]<min)
            min = block[i*block_wid+j];
          if (block[i*block_wid+j]>max)
            max = block[i*block_wid+j];
          mean += block[i*block_wid+j];
        }
      }
      diff = max - min;
      mean /= (double)(block_hei*block_wid);

      std = 0.0;
      for (i=0; i<block_hei; i++)
      {
        for (j=0; j<block_wid; j++)
        {
          std += (block[i*block_wid+j]-mean) * (block[i*block_wid+j]-mean);
        }
      }
      std = sqrt( std/(double)(block_hei*block_wid-1) );

      sceneBLK.blockDiff[i_pos*wid+j_pos] = diff;
      sceneBLK.blockMean[i_pos*wid+j_pos] = mean;
      sceneBLK.blockStd[i_pos*wid+j_pos] = std;

      delete[] block;
    }
  }
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------
/**
- Deterine piecewise model paramters fitting specified LUT, return sytnax values used to describe model and full set of model coefficients

\param   full_lut_len        length of LUT to fit
\param   target_lut          data values for LUT
\param   input_bit_depth     bit data values in model calculation
\param   nFractionBits       number of fraction bits for model constant coefficients
\param   smoothness_flag     flag indicating use smooth fit or search both smooth and piecewise linear
\retval  pUnifiedLumaReshapeInfo  syntax describing model and full set of model parameters
*/
void  Reshaper::AutomaticMinimalMSE_Fitting
  (
  int                      full_lut_len,
  double*                  full_target_lut,
  int                      input_bit_depth,
  int                      nFractionBits,
  bool                     smoothness_flag,
  UnifiedLumaReshapeInfo*  pUnifiedLumaReshapeInfo
  )
{
  double         target_lut[MAX_LUMA_RESHAPING_LUT_SIZE];
  bool           first_segment_flat_flag;
  int            flatFirstSegmentLength;
  bool           last_segment_flat_flag;  // TBD handling of last flat segments
  int            flatLastSegmentLength;
  int            low = 1;
  int            high = 1 << input_bit_depth;
  bool           force_initial_value_flag;    // forces first value of model to agree with target LUT needed for flat segments, lower MSE if allowed to change in for nonflat
  bool           force_final_value_flag;              // forces last value of model to agree with target LUT useful for last flat segments TBD use flag
  int            lut_len = full_lut_len;
  double         NumberOfSegmentDistortionFactor = 0.1; // accounts for number of segments in MSE fit decision default to zero to ignore overhead
  int            TargetNumberPivots;
  int            strategey;
  int            NumberRemainingPivotPoints;
  double         dx;
  int            left;
  double         Ystep;
  int            nSegments;
  int            s_opt, j_opt;
  double         B0;
  double         mse;
  int            xx = 0;
  int            MaxNumberPivots = MAX_N_SEGMENTS + 1;
  int            MinNumberPivots = MIN_N_SEGMENTS + 1;
  int            nPivotStrategies = NUMBER_PIVOT_STRATEGIES;
  int            ListNumPivotPoints_len = MaxNumberPivots - MinNumberPivots + 1;
  vector<double> A0S;
  vector<double> Asignalled;
  int            nFractionBitsA = nFractionBits;
  int            nFractionBitsB = input_bit_depth;
  int            j;

  vector<int> ListNumPivotPoints(ListNumPivotPoints_len, 0);

  for (int t = 0; t < ListNumPivotPoints_len; t++)
    ListNumPivotPoints[t] = t + MinNumberPivots;

  // test for nonmonotonic LUT used to avoid uniform Y axis partitioning in candidate construction
  vector<double> slope(full_lut_len - 1, 0);

  for (int t = 0; t < (int) slope.size(); t++)
    slope[t] = full_target_lut[t + 1] - full_target_lut[t];

  if (minVec(slope) < 0)
    nPivotStrategies = 1;

  // detect first flat segment of target lut
  first_segment_flat_flag = false;
  for(j=0;j<full_lut_len;j++)
  {
    if(full_target_lut[j]>full_target_lut[0])
      break;
  }
  if(j>1)
  {
    first_segment_flat_flag = true;
    force_initial_value_flag = true;
    flatFirstSegmentLength = j-1;
  }
  else
  {
    first_segment_flat_flag = false;
    force_initial_value_flag = true;
    flatFirstSegmentLength = 0;
  }
  low = flatFirstSegmentLength+1;
  last_segment_flat_flag = false;
  for(j=full_lut_len-1;j>=0;j--)
  {
    if(full_target_lut[j]<full_target_lut[full_lut_len-1])
      break;
  }
  if(j<(full_lut_len-1))
  {
    last_segment_flat_flag = true;
    force_final_value_flag = true;
    flatLastSegmentLength = (full_lut_len-1)-j - 1;
  }
  else
  {
    last_segment_flat_flag = false;
    force_final_value_flag = false;
    flatLastSegmentLength = 0;
  }
  high = (1 << input_bit_depth) - flatLastSegmentLength;

  // modify target LUT for MSE design by removing initial flat segment
  lut_len = full_lut_len-flatFirstSegmentLength-flatLastSegmentLength;
  for(int i = 0;i<lut_len;i++)
    target_lut[i] = full_target_lut[i+flatFirstSegmentLength];

  // remove one segment from the list of candidate segement lenghts due to used by first flat segment
  if (first_segment_flat_flag)
  {
    for (int t = 0; t < ListNumPivotPoints_len; t++)
      ListNumPivotPoints[t] = ListNumPivotPoints[t] - 1;
  }
  if (last_segment_flat_flag)
  {
    for (int t = 0; t < ListNumPivotPoints_len; t++)
      ListNumPivotPoints[t] = ListNumPivotPoints[t] - 1;
  }

  double*** CandidatePivots = new double**[MaxNumberPivots];
  for (int d3 = 0; d3 < MaxNumberPivots; d3++)
  {
    CandidatePivots[d3] = new double*[ListNumPivotPoints_len];
    for (int d2 = 0; d2 < ListNumPivotPoints_len; d2++)
    {
      CandidatePivots[d3][d2] = new double[nPivotStrategies];
      for (int d1 = 0; d1 < nPivotStrategies; d1++)
        CandidatePivots[d3][d2][d1] = 0;
    }
  }

  vector<double> temp(lut_len, 0);
  for (int j = 0; j <= ListNumPivotPoints_len - 1; j++)
  {
    TargetNumberPivots = ListNumPivotPoints[j];

    strategey = 1; // uniform x axis partitioning for pivot points
    vector<double> PivotPoints(TargetNumberPivots);
    PivotPoints[0] = low - 1;
    PivotPoints[TargetNumberPivots - 1] = high - 1;
    NumberRemainingPivotPoints = TargetNumberPivots - 2;
    if (NumberRemainingPivotPoints > 0)
    {
      dx = (PivotPoints[TargetNumberPivots - 1] - PivotPoints[0]) / (NumberRemainingPivotPoints + 1);
      for (int i = 0; i <= NumberRemainingPivotPoints - 1; i++)
        PivotPoints[i + 1] = round(PivotPoints[0] + (i + 1)*dx);
    }
    for (int i = 0; i <= TargetNumberPivots - 1; i++)
      CandidatePivots[i][j][strategey - 1] = PivotPoints[i];

    strategey = 2; // uniform Y axis partitioning for pivot points
    if (nPivotStrategies > 1)
    {
      for (int i = 0; i <= TargetNumberPivots - 1; i++)
        PivotPoints[i] = 0;
      PivotPoints[0] = low - 1;
      PivotPoints[TargetNumberPivots - 1] = high - 1;
      NumberRemainingPivotPoints = TargetNumberPivots - 2;
      left = 1;
      if (NumberRemainingPivotPoints > 0)
      {
        Ystep = (target_lut[lut_len - 1] - target_lut[0]) / (NumberRemainingPivotPoints + 1);
        for (int i = 1; i <= NumberRemainingPivotPoints; i++)
        {
          double y = target_lut[0] + (i)*Ystep;
          for (int q = 0; q < lut_len; q++)
          {
            if (target_lut[q] <= y)
              temp[q] = 1;
            else
              temp[q] = 0;
          }
          for (int q = lut_len-1; q >= 0; q--)
          {
            if (temp[q] == 1)
            {
              xx = q;
              break;
            }
          }
          double dXX; // TBD don't need double here or for PivotPoints
          dXX = xx + PivotPoints[0];
          dXX = dXX > (PivotPoints[left + i - 1 - 1] + MIN_PIVOT_SPACING) ? dXX : (PivotPoints[left + i - 1 - 1] + MIN_PIVOT_SPACING);
          dXX = dXX < PivotPoints[TargetNumberPivots - 1]  ? dXX : PivotPoints[TargetNumberPivots - 1];
          PivotPoints[left - 1 + i] = dXX;
        }
      }

      for (int t = 0; t <= TargetNumberPivots - 1; t++)
        CandidatePivots[t][j][strategey - 1] = PivotPoints[t];
    }
  }
#if MODE0_UPDATE
  const int nSmoothness = 2;
  double** AllCandidateMSE[nSmoothness];
  int sm;
  int sm_opt;
  for(sm=0;sm<nSmoothness;sm++)
  {
    AllCandidateMSE[sm] = new double*[nPivotStrategies];
    for (int t = 0; t < nPivotStrategies; t++)
      AllCandidateMSE[sm][t] = new double[ListNumPivotPoints_len];
  }
#endif
  double** CandidateMSE = new double*[nPivotStrategies];
  for (int t = 0; t < nPivotStrategies; t++)
    CandidateMSE[t] = new double[ListNumPivotPoints_len];

#if !MODE0_UPDATE
  double*** CandidateAopB0 = new double**[MaxNumberPivots + 1];
  for (int d3 = 0; d3 < MaxNumberPivots + 1; d3++)
  {
    CandidateAopB0[d3] = new double*[ListNumPivotPoints_len];
    for (int d2 = 0; d2 < ListNumPivotPoints_len; d2++)
    {
      CandidateAopB0[d3][d2] = new double[nPivotStrategies];
      for (int d1 = 0; d1 < nPivotStrategies; d1++)
        CandidateAopB0[d3][d2][d1] = 0;
    }
  }
#endif
  int A0S_len, nPivots;
#if MODE0_UPDATE
  for(sm=0;sm<nSmoothness;sm++)
  {
#endif
  for (int s = 0; s <= nPivotStrategies - 1; s++)
  {
    for (int j = 0; j <= ListNumPivotPoints_len - 1; j++)
    {
      nPivots = ListNumPivotPoints[j];
      vector<double> PivotPoints(nPivots);
      for (int t = 0; t < nPivots; t++)
        PivotPoints[t] = 0;
      for (int t = 0; t < nPivots; t++)
        PivotPoints[t] = CandidatePivots[t][j][s];
      nSegments = nPivots - 1;
      vector<double> SegmentLengths(nSegments);
      for (int t = 0; t < nSegments; t++)
        SegmentLengths[t] = 0;
      for (int i = 0; i < nSegments; i++)
        SegmentLengths[i] = PivotPoints[i + 1] - PivotPoints[i];

      if (smoothness_flag)
      {
        A0S_len = (nSegments + 2) - 1;
        A0S.clear();
        A0S.resize(A0S_len, 0);
      }
      else
      {
        A0S_len = nSegments + 1;
        A0S.clear();
        A0S.resize(A0S_len, 0);
      }

      /* ------------------------------------------------------------------------------- */
      /* create some temporary matrices to be used by function MinimalMSE_Fit(). */
      int nSamples   = lut_len;
      int nParams    = smoothness_flag ? (nSegments + 2) : (nSegments + 1);
      TempMatrices* cTempMatrices_1 = new TempMatrices(nSegments, nParams, nSamples);

      MinimalMSE_Fit
        (
        target_lut,
        lut_len,
        SegmentLengths,
        smoothness_flag,
        force_initial_value_flag,
        nFractionBitsA,
        nFractionBitsB,
        cTempMatrices_1,
        A0S,
        B0,
        mse
        );

      delete cTempMatrices_1;
      /* ------------------------------------------------------------------------------- */

#if !MODE0_UPDATE
      int AopB0_len	= A0S_len + 1;
      vector<double> AopB0(AopB0_len, 0);

      for (int t = 0; t < AopB0_len - 1; t++)
        AopB0[t] = A0S[t];

      AopB0[AopB0_len - 1] = B0;
#endif
      CandidateMSE[s][j] = NumberOfSegmentDistortionFactor * (ListNumPivotPoints[j] - 2) + mse; // NumberOfSegmentDistortionFactor * (ListNumPivotPoints[j_opt] - 2) + CandidateMSE[s_opt][j_opt];
#if MODE0_UPDATE
      AllCandidateMSE[sm][s][j] = CandidateMSE[s][j];
#endif
#if !MODE0_UPDATE
      for (int t = 0; t <= nPivots + 1 - 1; t++)
        CandidateAopB0[t][j][s] = AopB0[t];
#endif
    }
  }
#if MODE0_UPDATE
  }
#endif
  s_opt = 0;
  j_opt = 0;
#if MODE0_UPDATE
  int sm_first;
  sm_first = smoothness_flag ? 1 : 0;
  sm_opt = sm_first;
  CandidateMSE[s_opt][j_opt] = AllCandidateMSE[sm_opt][s_opt][j_opt]; 
#endif
  double err = NumberOfSegmentDistortionFactor * (ListNumPivotPoints[j_opt] - 2) + CandidateMSE[s_opt][j_opt];
#if MODE0_UPDATE
  double OldNumberOfSegmentDistortionFactor = NumberOfSegmentDistortionFactor;
  for(sm=sm_first;sm<nSmoothness;sm++)
  {
    for (int s = 0; s <= nPivotStrategies - 1; s++)
    for (int j = 0; j < ListNumPivotPoints_len; j++)
      CandidateMSE[s][j] = AllCandidateMSE[sm][s][j]; 
    // lower distortion factor when searching PL model note need to fix slightly but may cause MD5 match
    if(sm==0)
      NumberOfSegmentDistortionFactor = 0.1* OldNumberOfSegmentDistortionFactor;
    else
      NumberOfSegmentDistortionFactor = OldNumberOfSegmentDistortionFactor;
#endif
  for (int s = 0; s <= nPivotStrategies - 1; s++)
  {
    for (int j = 0; j < ListNumPivotPoints_len; j++)
    {
      if (NumberOfSegmentDistortionFactor*(ListNumPivotPoints[j] - 2) + CandidateMSE[s][j] < err)
      {
        err = NumberOfSegmentDistortionFactor*(ListNumPivotPoints[j] - 2) + CandidateMSE[s][j];
        s_opt = s;
        j_opt = j;
#if MODE0_UPDATE
        sm_opt = sm;
#endif
      }
    }
  }
#if MODE0_UPDATE
  }
  smoothness_flag = (sm_opt==1);
#endif
  vector<double> PivotPoints(ListNumPivotPoints[j_opt], 0);

  nPivots = PivotPoints.size();

  for (int t = 0; t < nPivots; t++)
    PivotPoints[t] = CandidatePivots[t][j_opt][s_opt];

#if !MODE0_UPDATE
  vector<double> AopB0_mse(nPivots + 1, 0);

  for (int t = 0; t < nPivots + 1; t++)
    AopB0_mse[t] = CandidateAopB0[t][j_opt][s_opt];
#endif
  nSegments = nPivots - 1;

  vector<double> SegmentLengths(nSegments, 0);

  for (int i = 0; i < nSegments; i++)
    SegmentLengths[i] = PivotPoints[i + 1] - PivotPoints[i];

  int			Asignalled_len;
  double		Bsignalled;

  if (smoothness_flag)
  {
    Asignalled_len = (nSegments + 2) - 1;
    Asignalled.clear();
    Asignalled.resize(Asignalled_len, 0);
  }
  else
  {
    Asignalled_len = nSegments + 1;
    Asignalled.clear();
    Asignalled.resize(Asignalled_len, 0);
  }

  /* ------------------------------------------------------------------------------- */
  /* create some temporary matrices to be used by function MinimalMSE_Fit(). */
  int nSamples = lut_len;
  int nParams  = smoothness_flag ? (nSegments + 2) : (nSegments + 1);
  TempMatrices* cTempMatrices_2 = new TempMatrices(nSegments, nParams, nSamples);
  //            MinimalMSE_Fit
  MinimalMSE_Fit
    (
    target_lut,
    lut_len,
    SegmentLengths,
    smoothness_flag,
    force_initial_value_flag,
    nFractionBitsA,
    nFractionBitsB,                
    cTempMatrices_2,
    Asignalled,
    Bsignalled,
    mse
    );

  delete cTempMatrices_2;
  /* ------------------------------------------------------------------------------- */

  pUnifiedLumaReshapeInfo->luma_bit_depth = input_bit_depth;
  pUnifiedLumaReshapeInfo->reshape_equal_ranges = 0;
  pUnifiedLumaReshapeInfo->reshape_num_ranges_minus1 = nSegments - 1;
  pUnifiedLumaReshapeInfo->reshape_global_offset_val = flatFirstSegmentLength; 

  for (int s = 0; s < pUnifiedLumaReshapeInfo->reshape_num_ranges_minus1 + 1; s++)
  {
    pUnifiedLumaReshapeInfo->reshape_range_val[s] = (int)SegmentLengths[s];
  }
  pUnifiedLumaReshapeInfo->reshape_continuity_flag = smoothness_flag;  

  pUnifiedLumaReshapeInfo->coeff_log2_offset_minus2 = nFractionBitsA - 2;  
  pUnifiedLumaReshapeInfo->fractionBits[0] = pUnifiedLumaReshapeInfo->coeff_log2_offset_minus2 + 2;
  pUnifiedLumaReshapeInfo->fractionBits[1] = pUnifiedLumaReshapeInfo->fractionBits[0] + input_bit_depth;
  pUnifiedLumaReshapeInfo->fractionBits[2] = pUnifiedLumaReshapeInfo->fractionBits[0] + 2 * input_bit_depth;

  for (int i = 0; i < Asignalled_len; i++)
  {
    pUnifiedLumaReshapeInfo->constantCoeffSignalled[i] = (int)(Asignalled[i] * (1 << pUnifiedLumaReshapeInfo->fractionBits[0]));
    pUnifiedLumaReshapeInfo->constantCoeffSignalledInt[i] = pUnifiedLumaReshapeInfo->constantCoeffSignalled[i] >> pUnifiedLumaReshapeInfo->fractionBits[0];
    pUnifiedLumaReshapeInfo->constantCoeffSignalledFrac[i] = pUnifiedLumaReshapeInfo->constantCoeffSignalled[i] & ((1<<pUnifiedLumaReshapeInfo->fractionBits[0])-1);
  }
  if (pUnifiedLumaReshapeInfo->reshape_continuity_flag)
  {
    pUnifiedLumaReshapeInfo->linearCoeffSignalled = (int)(Bsignalled *(1 << pUnifiedLumaReshapeInfo->fractionBits[1]) + 0.5);
    pUnifiedLumaReshapeInfo->linearCoeffSignalledInt = pUnifiedLumaReshapeInfo->linearCoeffSignalled >> pUnifiedLumaReshapeInfo->fractionBits[1];
    pUnifiedLumaReshapeInfo->linearCoeffSignalledFrac = pUnifiedLumaReshapeInfo->linearCoeffSignalled & ((1<<pUnifiedLumaReshapeInfo->fractionBits[1])-1);
  }
  else
  {
    pUnifiedLumaReshapeInfo->linearCoeffSignalled = 0;
    pUnifiedLumaReshapeInfo->linearCoeffSignalledInt = 0;
    pUnifiedLumaReshapeInfo->linearCoeffSignalledFrac = 0;
  }
  // derive full set of coefficients
  DeriveAllSegments(
    pUnifiedLumaReshapeInfo->reshape_num_ranges_minus1 + 1,
    pUnifiedLumaReshapeInfo->reshape_range_val,
    pUnifiedLumaReshapeInfo->reshape_global_offset_val,
    full_lut_len,
    &pUnifiedLumaReshapeInfo->nAllSegmentLengths,
    pUnifiedLumaReshapeInfo->allSegmentLengths,
    &pUnifiedLumaReshapeInfo->firstSegmentFlat,&pUnifiedLumaReshapeInfo->lastSegmentFlat);

  if (smoothness_flag)
    DeriveSmoothPolynomialCoefficients(
    pUnifiedLumaReshapeInfo->reshape_num_ranges_minus1 + 1,
    pUnifiedLumaReshapeInfo->reshape_range_val,
    pUnifiedLumaReshapeInfo->constantCoeffSignalled,
    pUnifiedLumaReshapeInfo->linearCoeffSignalled,
    pUnifiedLumaReshapeInfo->firstSegmentFlat,
    pUnifiedLumaReshapeInfo->lastSegmentFlat,
    pUnifiedLumaReshapeInfo->fractionBits,
    pUnifiedLumaReshapeInfo->polyCoeff);
  else
    DerivePiecewiseLinearPolynomialCoefficients(
    pUnifiedLumaReshapeInfo->reshape_num_ranges_minus1 + 1,
    pUnifiedLumaReshapeInfo->reshape_range_val,
    pUnifiedLumaReshapeInfo->constantCoeffSignalled,
    pUnifiedLumaReshapeInfo->firstSegmentFlat,
    pUnifiedLumaReshapeInfo->lastSegmentFlat,
    pUnifiedLumaReshapeInfo->fractionBits,
    pUnifiedLumaReshapeInfo->polyCoeff);
}

/**
- write piecewise polynomial luma reshaping  paramters to text file
\param   filename                 output text file name
\param   POC                      frame POC to indicate scene start
\param   pUnifiedLumaReshape      structure containing luma reshaping model
*/
void Reshaper::WriteMetaData(char *filename, int POC, UnifiedLumaReshapeInfo *pUnifiedLumaReshape)
{
  FILE *fpMetaData = fopen (filename, "a+");
  fprintf(fpMetaData,"%d ",POC-m_startFrame); //output metadata: POC always starts from zero to be aligned with POC at encoder.
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->luma_bit_depth);
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->reshape_num_ranges_minus1 + 1);
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->reshape_equal_ranges);
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->reshape_global_offset_val);
  for (int s = 0; s < pUnifiedLumaReshape->reshape_num_ranges_minus1 + 1; s++)
    fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->reshape_range_val[s]);
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->coeff_log2_offset_minus2 + 2);
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->reshape_continuity_flag);
  for (int s = 0; s < pUnifiedLumaReshape->reshape_num_ranges_minus1 + 2; s++)
    fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->constantCoeffSignalled[s]);
  if(pUnifiedLumaReshape->reshape_continuity_flag)
    fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->linearCoeffSignalled);
#if RESHAPER_JOINT_OPT
  fprintf(fpMetaData,"%d ",pUnifiedLumaReshape->opt);
  if (pUnifiedLumaReshape->opt)
  {
    for (int i=0; i<15; i++)
    {
      fprintf(fpMetaData, "%d ", pUnifiedLumaReshape->lumaDQP[i]);
    }
  }
#endif
  fprintf(fpMetaData,"\n");
  fclose(fpMetaData);
  return;
}

/**
- set fraction bits for all coefficients of differnt order from order 0 coefficients

\param   coeff_log2_offset_minus2  syntax element defining number of fraction bits
\param   bitDepth                  bitdepth of source
\retval  fractionBits              array of number of coefficient fraction bits per  order
*/
void Reshaper::DetermineAllFractionBits(int coeff_log2_offset_minus2,int bitDepth,int fractionBits[MAX_PRED_ORDER+1])
{
  int i;
  for(i=0;i<MAX_PRED_ORDER+1;i++)
    fractionBits[i] = coeff_log2_offset_minus2 + 2 + i*bitDepth;
  return;
}

/**
- Determine number and length of all segments from signalled information, presence of last flat segment determined if sum of lengths less than full model domain i.e. 1024
\param   nSegmentSignaled      number of segments explicitly signaled
\param   SegmentLengths        signalled segment lengths
\param   global_offset_val     length of initial flat segment
\param   total_length          total length of model i.e. 1<< bit_depth
\retval  pNumberAllSegments    number of all segments
\retval  pAllSegmentLengths    length of all segments not just those signalled
\retval  pFirstSegmentFlat     flag indicating first segment is flat
\retval  pLastSegmentFlat      flag indicating last segment is flat
*/
void Reshaper::DeriveAllSegments(int nSegmentSignaled, int SegmentLengths[MAX_N_SEGMENTS],int global_offset_val,int total_length,int *pNumberAllSegments,int pAllSegmentLengths[MAX_N_SEGMENTS],bool *pFirstSegmentFlat,bool *pLastSegmentFlat)
{
  int nAllSegments;
  int sum;
  int RemainingLength;
  int first;
  int i;

  nAllSegments = nSegmentSignaled;

  *pFirstSegmentFlat = (global_offset_val>0);
  if(*pFirstSegmentFlat)
  {
    nAllSegments++;
    first = 1;
    pAllSegmentLengths[0] = global_offset_val;
  }  
  else
  {
    first = 0;
  }
  for(i=0;i<nSegmentSignaled;i++)
    pAllSegmentLengths[i+first] = SegmentLengths[i];

  sum = 0;
  for(i=0;i<nAllSegments;i++)
    sum = sum + pAllSegmentLengths[i];
  RemainingLength = total_length - sum;
  assert(RemainingLength>=0);
  // TBD fix for limit of current syntax can't signal pivot point of 1024 so sum of lengths is 1023 at most
  if(RemainingLength==1)
  {
    pAllSegmentLengths[nAllSegments-1] = pAllSegmentLengths[nAllSegments-1] + 1;
    RemainingLength = 0;
  }
  if(RemainingLength>0)
  {
    *pLastSegmentFlat = true;
    nAllSegments++;
    pAllSegmentLengths[nAllSegments-1] = RemainingLength;
  }
  else
  {
    *pLastSegmentFlat = false;
  }

  *pNumberAllSegments = nAllSegments;
  return;
}
/**
- Determine piecewise model polynomial coefficients from reduced signalled set.  Note coefficients on flat intervals are also set , return full set of model coefficients

\param   nSegmentSignaled    number of segments explicitly signaled
\param   SegmentLengths      signalled segment lengths
\param   constantCoeff       values of constant coefficients signalled
\param   linearCoeff         value of signalled linear coefficient
\param   FirstSegmentFlatFlag flag indicating last segment is flat
\param   LastSegmentFlatFlag flag indicating last segment is flat
\param   nFractionBits       number of fraction bits for model constant coefficients
\retval  A                   values of constant model coefficients at all segments
\retval  B                   values of linear model coefficients at all segments
\retval  C                   values of quadratic model coefficients at all segments
*/
void Reshaper::DeriveSmoothPolynomialCoefficients(int nSegmentSignaled, int SegmentLengths[MAX_N_SEGMENTS], int constantCoeff[1+MAX_N_SEGMENTS], int linearCoeff, bool FirstSegmentFlatFlag, bool LastSegmentFlatFlag,int nFractionBits[3], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1])
{
  long Atemp[1+MAX_N_SEGMENTS];
  long Btemp[1+MAX_N_SEGMENTS];
  long s;
  int FirstSegment; // first signalled segments
  int nSegments;

  nSegments = nSegmentSignaled;
  if(FirstSegmentFlatFlag)
  {
    FirstSegment = 1;
    nSegments = nSegments +1;
  }
  else
  {
    FirstSegment = 0;
  }
  if(LastSegmentFlatFlag)
  {
    nSegments = nSegments +1;
  }
  // set temporary A values for derivation
  for(s=0;s<nSegmentSignaled+1;s++)
  {
    Atemp[s] = constantCoeff[s];
  }
  Btemp[0] = (int64) linearCoeff;
  // solve for remaining B values for derivation
  for(s=1;s<nSegmentSignaled+1;s++)
  {
    long Numerator;
    Numerator = (Atemp[s] - Atemp[s-1])<<(nFractionBits[1]-nFractionBits[0] + 1);
    Numerator = Numerator + (SegmentLengths[s-1]>>1);

    Btemp[s] = (Numerator/SegmentLengths[s-1]) - Btemp[s-1];
  }
  // solve for C values intermediate calculations may exceed 32 bits and set output A an B values in signalled range
  for(s=0;s<nSegmentSignaled;s++)
  {
    int64 Numerator;
    polyCoeff[s+FirstSegment][0] = Atemp[s];
    polyCoeff[s+FirstSegment][1]  = Btemp[s];
    Numerator = ((int64)(Btemp[s+1] - Btemp[s])) << (nFractionBits[2]-nFractionBits[1] - 1);
    Numerator = Numerator + (SegmentLengths[s]>>1);
    polyCoeff[s+FirstSegment][2] = (long) Numerator/SegmentLengths[s];
  }  
  if(FirstSegmentFlatFlag)
  {
    polyCoeff[0][0] = constantCoeff[0];
    polyCoeff[0][1] = 0;
    polyCoeff[0][2] = 0;
  }  
  if(LastSegmentFlatFlag)
  {
    polyCoeff[nSegments-1][0] = constantCoeff[nSegmentSignaled];
    polyCoeff[nSegments-1][1] = 0;
    polyCoeff[nSegments-1][2] = 0;
  }
}

/**
- Determine piecewise model polynomial coefficients from reduced signalled set.  Note coefficients on flat intervals are also set , return full set of model coefficients

\param   nSegmentSignaled    number of segments explicitly signaled
\param   SegmentLengths      signalled segment lengths
\param   constantCoeff       values of constant coefficients signalled
\param   FirstSegmentFlatFlag flag indicating last segment is flat
\param   LastSegmentFlatFlag flag indicating last segment is flat
\param   nFractionBits       number of fraction bits for model constant coefficients
\retval  A                   values of constant model coefficients at all segments
\retval  B                   values of linear model coefficients at all segments
\retval  C                   values of quadratic model coefficients at all segments
*/
void Reshaper::DerivePiecewiseLinearPolynomialCoefficients(int nSegmentSignaled, int SegmentLengths[MAX_N_SEGMENTS], int constantCoeff[1+MAX_N_SEGMENTS],bool FirstSegmentFlatFlag,bool LastSegmentFlatFlag,int nFractionBits[3], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1])
{
  long s;
  int FirstSegment; // first signalled segment
  int nSegments;

  nSegments = nSegmentSignaled;
  if(FirstSegmentFlatFlag)
  {
    FirstSegment = 1;
    nSegments = nSegments + 1;
  }
  else
  {
    FirstSegment = 0;
  }
  if(LastSegmentFlatFlag)
  {
    nSegments = nSegments + 1;
  }
  // set A, B, C values within segments 
  for(s=0;s<nSegmentSignaled;s++)
  {
    polyCoeff[s+FirstSegment][0] = constantCoeff[s];
    polyCoeff[s+FirstSegment][1] = (((constantCoeff[s+1]-constantCoeff[s])<<(nFractionBits[1]-nFractionBits[0])) + (SegmentLengths[s]>>1))/SegmentLengths[s];
    polyCoeff[s+FirstSegment][2] = 0;
  }
  if(FirstSegmentFlatFlag)
  {
    polyCoeff[0][0] = constantCoeff[0];
    polyCoeff[0][1] = 0;
    polyCoeff[0][2] = 0;
  }  
  if(LastSegmentFlatFlag)
  {
    polyCoeff[nSegments-1][0] = constantCoeff[nSegmentSignaled];
    polyCoeff[nSegments-1][1] = 0;
    polyCoeff[nSegments-1][2] = 0;
  }
}

/**
- Build inverse reshaping LUT from full set of model segments and coefficients

\param   nSegmentAll         number of segments explicitly signaled and inferred
\param   allSegmentLengths   lengths of all segments thoes explicitly signalled and inferred
\param   nFractionBits       number of fraction bits for model constant coefficients
\param   A                   values of constant model coefficients at all segments
\param   B                   values of linear model coefficients at all segments
\param   C                   values of quadratic model coefficients at all segments
\param   lut_size            size of LUT in number of samples
\retval  InverseReshapingLUT LUT describing inverse reshaping
*/
void Reshaper::BuildInverseReshapingLUT(int nSegmentAll, int allSegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS],int nFractionBits[MAX_PRED_ORDER+1], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1],int lut_size,int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE])
{
  int s;
  int dx;
  int idx;
  int64 result;
  int xi[MAX_LUMA_RESHAPE_SEGMENTS];
  int order;
  int64 ModPolyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER + 1];
  // modifiy coefficients
  // determine left endpoints of each segment
  xi[0]=0;
  for(s=1;s<nSegmentAll;s++)
    xi[s] = xi[s-1] + allSegmentLengths[s-1];
  // constant coefficients 
  order = 0;
  for(s=0;s<nSegmentAll;s++)
  {
    ModPolyCoeff[s][order] = ( 
      (((int64)polyCoeff[s][0]) << (nFractionBits[2]-nFractionBits[0])) - 
      (((int64)polyCoeff[s][1]) << (nFractionBits[2]-nFractionBits[1]))*xi[s] + 
      (((int64)polyCoeff[s][2])*xi[s]*xi[s]) + 
      (((int64)1)<<(nFractionBits[2]-nFractionBits[0]-1)) ) >> 
      (nFractionBits[2]-nFractionBits[0]);
  }
  // linear coefficients 
  order = 1;
  for(s=0;s<nSegmentAll;s++)
  {
    ModPolyCoeff[s][order] = ( 
      (((int64)polyCoeff[s][1]) <<(nFractionBits[2]-nFractionBits[1])) - 
      (((int64)polyCoeff[s][2])*xi[s]*2) + 
      (((int64)1)<<(nFractionBits[2]-nFractionBits[1]-1)) ) >> 
      (nFractionBits[2]-nFractionBits[1]);
  }
  // quadratic coefficients unmodified
  order = 2;
  for(s=0;s<nSegmentAll;s++)
    ModPolyCoeff[s][order] = polyCoeff[s][order];
  // calculate using modified coefficients and description relative to absolute zero A' + B'*x + C'*x*x
  idx = 0;
  for(s=0;s<nSegmentAll;s++)
  {
    for(dx=0;dx<allSegmentLengths[s];dx++)
    {
      result = ModPolyCoeff[s][0] << (nFractionBits[2]-nFractionBits[0]);
      result = result + (ModPolyCoeff[s][1] << (nFractionBits[2]-nFractionBits[1]))*((int64)idx);
      result = result + ModPolyCoeff[s][2]*((int64)idx)*idx;                             // TBD note on rounding offset in the spec text right now
#if MODE0_UPDATE
      if(idx<MAX_LUMA_RESHAPING_LUT_SIZE)
      {
        InverseReshapingLUT[idx] = iClip((int)(result >> nFractionBits[2]), 0, MAX_LUMA_RESHAPING_LUT_SIZE-1 ); // TBD determine upper clipping limit from bitdepth
        idx = idx+1;
      }
#else
      InverseReshapingLUT[idx] = iClip((int)(result >> nFractionBits[2]), 0, MAX_LUMA_RESHAPING_LUT_SIZE-1 ); // TBD determine upper clipping limit from bitdepth
      idx = idx+1;
#endif
    }
  }
}

/**
- Build forward reshaping LUT from full set of model segments and coefficients
\param   InverseReshapingLUT LUT describing inverse reshaping
\param   lut_size            size of LUT in number of samples
\retval  ForwardReshapingLUT LUT describing forward reshaping
*/
void Reshaper::BuildForwardReshapingLUT(int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int ForwardReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size)
{
  int low = InverseReshapingLUT[0];
  int high = InverseReshapingLUT[lut_size-1];
  int j;
  int first = 0;;
  for(int i=0;i<lut_size;i++)
    if (InverseReshapingLUT[0] < InverseReshapingLUT[i])
    {
      first = i-1;
      break;
    }
    int last=lut_size-1;
    for(int i=lut_size-1;i>=0;i--)
      if (InverseReshapingLUT[lut_size-1] > InverseReshapingLUT[i])
      {
        last = i+1;
        break;
      }
      for(int i=0;i<lut_size;i++)
      {
        if(i<=low)
        {
          ForwardReshapingLUT[i] = first;
        }
        else if(i>=high)
        {
          ForwardReshapingLUT[i] = last;
        }
        else
        {
          for(j=0;j<lut_size-1;j++)
            if(InverseReshapingLUT[j]>=i)
            {
              ForwardReshapingLUT[i] = j;
              break;
            }
        }
      }
}

#if TCH_CROSSPLANE_CRCB
/**
- Build forward reshaping LUT from full set of model segments and coefficients
\param   ForwardYlogLUT LUT describing forward logY  reshaping
\param   lut_size            size of LUT in number of samples
\retval  InverseYlogLUT LUT describing invesse logY  reshaping
*/
void Reshaper::BuildInverseYlogLUT(double ForwardYlogLUT[MAX_LUMA_RESHAPING_LUT_SIZE],double InverseYlogLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size)
{
  double low = ForwardYlogLUT[0];
  double high = ForwardYlogLUT[lut_size-1];
  int j;
  int first = 0;;
  for(int i=0;i<lut_size;i++)
    if (ForwardYlogLUT[0] < ForwardYlogLUT[i])
    {
      first = i-1;
      break;
    }
    int last=lut_size-1;
    for(int i=lut_size-1;i>=0;i--)
      if (ForwardYlogLUT[lut_size-1] > ForwardYlogLUT[i])
      {
        last = i+1;
        break;
      }
      for(int i=0;i<lut_size;i++)
      {
        if(i<=low)
        {
          InverseYlogLUT[i] = first;
        }
        else if(i>=high)
        {
          InverseYlogLUT[i] = last;
        }
        else
        {
          for(j=0;j<lut_size-1;j++)
            if(ForwardYlogLUT[j]>=i)
            {
              InverseYlogLUT[i] = j;
              break;
            }
        }
      }
}
#endif
/**
- Build an identity matrix 
\param   lut_size            size of LUT in number of samples
\retval  LUT                 LUT describing an identiy matrix
*/
void Reshaper::BuildIdentityLUT(int LUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size)
{
  int i;
  for(i=0;i<lut_size;i++)
    LUT[i] = i;
}

//---------------------------------------------------------------------------------
#if QC_CBCR_m37064
float Reshaper::evaluateFloatRepn(int x, int const numBitsVal, int const numBitsFrac)
{
  assert(x < (1 << numBitsVal) && x >= 0);

  int signPart = x >> numBitsFrac;
  int numSignBits = numBitsVal - numBitsFrac;
  assert(signPart >= 0 && signPart < (1 << numSignBits));

  int fracPart = x - (signPart << numBitsFrac);
  assert(fracPart >= 0 && fracPart < (1 << numBitsFrac));

  float retVal = signPart + (float)fracPart / (1 << numBitsFrac);

  assert(retVal >= 0 && retVal < (1 << numBitsVal));
  return retVal;
}


int Reshaper::evaluateIntRepn(float const x, int const numBitsVal, int const numBitsFrac)
{
  assert(x >= 0);
  float fracVal = x - (int)x;
  assert(fracVal >= 0 && fracVal < 1);

  int numBitsSignif = numBitsVal - numBitsFrac;
  if (numBitsSignif == 0)
  {
    assert(x < 1 && (x - fracVal) == 0);
  }
  else
  {
    assert(x - fracVal < (1 << numBitsSignif));
  }


  int intRepnOfFrac = (int) (float)floor((fracVal * (1 << numBitsFrac)) + 0.5); // / (1 << numBitsFrac);


  int retVal = (int)x * (1 << numBitsFrac) + intRepnOfFrac;

  assert(retVal >= 0 && retVal < (1 << numBitsVal));
  return retVal;
}
void Reshaper::setChromaScaling(float *trg_scale, float *trg_offset, float src_scale, float src_offset)
{
  //float temp;

  *trg_scale = evaluateFloatRepn(evaluateIntRepn(src_scale, QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH), QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
  *trg_offset = evaluateFloatRepn(evaluateIntRepn(src_offset, QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH), QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH);
}


void Reshaper::writeCbCrATF(char *filename, int POC)
{
	FILE *atfParamFile = fopen(filename, "a+");
	float cbScale = m_chromaHistScaleCbInv;
	float crScale = m_chromaHistScaleCrInv;

	int m_atfDomainId = 1;
	int		l_atfNumberRanges[3] = { 1, 1, 1 };
	int		l_atfFlagEqualRanges[3] = { 1, 1, 1 };
	float	l_atfOffset[3] = { 0.0, 0.0, 0.0 };
	float	l_atfRanges[3][32];
	float	l_atfScale[3][32];
	
	int	l_atfRangesInt[3][32];
	int	l_atfScaleInt[3][32];
	int	l_atfOffsetInt[3];

	//output metadata: POC always starts from zero to be aligned with POC at encoder.
	fprintf(atfParamFile, "%d ", POC - m_startFrame);
	// Output reshaper_range_id :  0 - full range, 1 - standard range. 

	fprintf(atfParamFile, "%d ", getRangeId());

	l_atfScale[0][0] = 1.0;
	l_atfScale[1][0] = cbScale;
	l_atfScale[2][0] = crScale;
	

	for (int i = 1; i < 3; i++)
	{
		double l_offset = 0.0;
#if !MODE0_UPDATE
    if (l_atfScale[i][0] < 0.5)
    {
      l_offset = 1.0 / 1023;
    }
#endif
		l_atfOffset[i] = (1.0 - l_atfScale[i][0] - l_offset) / 2;
	}

	for (int i = 0; i < 3; i++)
	{
		l_atfRanges[i][0] = 1.0;
	}

	for (int i = 0; i < 3; i++)
	{
		l_atfScaleInt[i][0] = evaluateIntRepn(l_atfScale[i][0], QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
		if (l_atfFlagEqualRanges[i] == 0)
		{
			l_atfRangesInt[i][0] = evaluateIntRepn(l_atfRanges[i][0], QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH);
		}
		l_atfOffsetInt[i] = evaluateIntRepn(l_atfOffset[i], QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH);
	}

	fprintf(atfParamFile, "%d ", m_atfDomainId);
	fprintf(atfParamFile, "%d ", 0); //chroma Mode = 0 means auto-plane
    fprintf(atfParamFile, "%d ", 2); //number of chroma reshaping parameters. 

	for (int j = 1; j < 3; j++)
	{
		// Write number of bins
		fprintf(atfParamFile, "%d ", l_atfNumberRanges[j]);

		// Write flag Equal Ranges
		fprintf(atfParamFile, "%d ", l_atfFlagEqualRanges[j]);

		// Write offset value
		fprintf(atfParamFile, "%d ", l_atfOffsetInt[j]);
		// Write scale value for each bin
		for (int i = 0; i < l_atfNumberRanges[j]; i++)
		{
			fprintf(atfParamFile, " %d ", l_atfScaleInt[j][i]);
		}
		if (l_atfFlagEqualRanges[j] == 0)
		{
			for (int i = 0; i < l_atfNumberRanges[i]; i++)
			{
				fprintf(atfParamFile, " %d ", l_atfRangesInt[j][i]);
			}
		}
	}
	fprintf(atfParamFile, "\n");
	fclose(atfParamFile);
}

void Reshaper::setParamsATF(ReshapingInfo *reshape)
{
  float scaleValueInv;
  if (reshape->m_isChromaHistScale)
  {
    // Cb
    scaleValueInv = evaluateFloatRepn(evaluateIntRepn(1.0 / (reshape->m_chromaHistScaleCb), QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH), QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
    m_chromaHistScaleCbInv = scaleValueInv;
    m_chromaHistScaleCb = 1.0 / scaleValueInv;
    m_chromaHistOffsetCb = reshape->m_chromaHistOffsetCb;

	//Cr
    scaleValueInv = evaluateFloatRepn(evaluateIntRepn(1.0 / (reshape->m_chromaHistScaleCr), QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH), QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
    m_chromaHistScaleCrInv = scaleValueInv;
    m_chromaHistScaleCr = 1.0 / scaleValueInv;
    m_chromaHistOffsetCr = reshape->m_chromaHistOffsetCr;
    m_flagScaleChromaHist = true;
  }
}

#endif

//----------------------------------------------------------------------------------------
#if MODE0_UPDATE
void Reshaper::InitSceneData()
{
  m_sceneNum = 0;
  m_sceneData.sceneFirstFno = 0;
  m_sceneData.srcPrimary = 0;
  m_sceneData.sceneIsFades = 0;

  m_sceneData.prev_frameNo = 0;
  m_sceneData.prev_org_I_CH = 0.0;
  m_sceneData.curr_org_I_CH = 0.0;
#if RESHAPER_JOINT_OPT
  memset(m_lumaDQP, 0, 15*sizeof(int));
#endif
}
#else
void Reshaper::readSceneMetadata(char *metaFile)
{
  FILE *reshapingMetaFile = fopen (metaFile, "r");
  char line[65536];
  int readCharCount;

  //find out number of scenes
  m_numOfScenes = 0;
  if ( reshapingMetaFile != NULL)
  {
    while (!feof(reshapingMetaFile))
    {
      if (fgets(line, 65536,  reshapingMetaFile) != NULL)
        ++m_numOfScenes;
    }
    fclose (reshapingMetaFile);

    if (m_numOfScenes < 1)
    {
      fprintf(stderr, "Error reading metadata file [%s].\n", metaFile);
      exit(EXIT_FAILURE);
    }
  }
  else
  {
    m_numOfScenes = 1;
  }

  m_sceneData = new SceneData[m_numOfScenes];  //remember to delete
  reshapingMetaFile = fopen (metaFile, "r");
  if ( reshapingMetaFile != NULL)
  {
    for  ( int sceneNum = 0; sceneNum < m_numOfScenes; sceneNum++)
    {
      if (fgets(line, 65536,  reshapingMetaFile) != NULL)
      {
        sscanf(line, "%d %d %d %n", &m_sceneData[sceneNum].sceneFirstFno, &m_sceneData[sceneNum].srcPrimary, &m_sceneData[sceneNum].sceneIsFades, &readCharCount);
      }
    }
  }
  else
  {
    int sceneNum = 0;
    m_sceneData[sceneNum].sceneFirstFno = 0;
    m_sceneData[sceneNum].srcPrimary = 0;
    m_sceneData[sceneNum].sceneIsFades = 0;
  }
  m_sceneNum = 0;
}
#endif
#if MODE0_UPDATE
void Reshaper::frameAnalysis(const Frame *inp)
{
  double meta_max = (double)(MAX_LUMA_RESHAPING_LUT_SIZE-1);
  int i, idx;
  double cur_I_CH=0.0, cur_I_CL=0.0;
  int max_value, min_value, tmp_CL, tmp_CH;

  max_value = inp->m_maxPelValue[0];
  min_value = 0;

  tmp_CL = max_value;
  tmp_CH = min_value;
  for (i = 0; i < inp->m_compSize[0]; i++) 
  {
    idx = inp->m_ui16Comp[0][i];
    if (idx < tmp_CL) 
      tmp_CL = idx;
    if (idx > tmp_CH )
      tmp_CH = idx;
  }

  assert (inp->m_sampleRange <= SR_FULL);
  if (inp->m_sampleRange == SR_STANDARD)
  {
	  cur_I_CL = ((double)tmp_CL/4.0-16.0)/219.0; //hardcode to deal with 10bit video for now
	  cur_I_CH = ((double)tmp_CH/4.0-16.0)/219.0;
  }
  else if (inp->m_sampleRange == SR_FULL)
  {
	  cur_I_CL = (double)tmp_CL/meta_max; 
	  cur_I_CH = (double)tmp_CH/meta_max;
  }
  
  m_sceneData.prev_org_I_CH = cur_I_CH;
  m_sceneData.curr_org_I_CH = cur_I_CH;
}
#endif

#if NO_INTRA_PERIOD_UPDATE
void Reshaper::genFrameStats(const Frame *inp, FrameStats *framestats)
{
  double meta_max = (double)(MAX_LUMA_RESHAPING_LUT_SIZE-1);
  int i, idx;
  double org_I_CH=0.0, org_I_CL=0.0;
  int max_value, min_value, tmp_CL, tmp_CH;

  double *Idata = new double[inp->m_compSize[0]];
  int blk_size   = 16;
  int blk_width  = (int)dRound((double)inp->m_width[0] / blk_size);
  int blk_height = (int)dRound((double)inp->m_height[0] / blk_size);
  SceneBlkData sceneBlock;
  sceneBlock.blockSize  = blk_size;
  sceneBlock.blockWidth  = blk_width;
  sceneBlock.blockHeight = blk_height;
  sceneBlock.blockDiff = new double[blk_width*blk_height];
  sceneBlock.blockStd  = new double[blk_width*blk_height];
  sceneBlock.blockMean = new double[blk_width*blk_height];
  double thres_mean_min = 50.0, thres_mean_max = 400.0;
  double thres_diff = 256.0, thres_std = 16.0;
  double num_bright_flat = 0.0, num_bright_sharp = 0.0, num_dark_flat = 0.0, num_dark_sharp = 0.0, num_total = 0.0;

  int i_pos, j_pos;

  max_value = inp->m_maxPelValue[0];
  min_value = 0;

  tmp_CL = max_value;
  tmp_CH = min_value;
  for (i = 0; i < inp->m_compSize[0]; i++) 
  {
    idx = inp->m_ui16Comp[0][i];
    if (idx < tmp_CL) 
      tmp_CL = idx;
    if (idx > tmp_CH )
      tmp_CH = idx;
  }

  assert (inp->m_sampleRange <= SR_FULL);
  if (inp->m_sampleRange == SR_STANDARD)
  {
    org_I_CL = ((double)tmp_CL/4.0-16.0)/219.0; //hardcode to deal with 10bit video for now
    org_I_CH = ((double)tmp_CH/4.0-16.0)/219.0;
    thres_mean_min = 106.8152;
    thres_mean_max = 406.522;  
    thres_diff = 219.214; 
    thres_std = 13.7; 
  }
  else if (inp->m_sampleRange == SR_FULL)
  {
    org_I_CL = (double)tmp_CL/meta_max; 
    org_I_CH = (double)tmp_CH/meta_max;
    thres_mean_min = 50.0, 
      thres_mean_max = 400.0;
    thres_diff = 256.0; 
    thres_std = 16.0;
  }


  // reading luma component 
  for (i = 0; i < inp->m_compSize[0]; i++) 
  {
    Idata[i] = (double)inp->m_ui16Comp[0][i];
  }

  blockproc(Idata, inp->m_width[0], inp->m_height[0], sceneBlock);

  // count class numbers
  num_bright_flat	= 0.0;
  num_bright_sharp	= 0.0;
  num_dark_flat		= 0.0;
  num_dark_sharp	= 0.0;
  num_total			= 0.0;
  for (i_pos=0; i_pos<blk_height; i_pos++)
  {
    for (j_pos=0; j_pos<blk_width; j_pos++)
    {
      if (sceneBlock.blockMean[i_pos*blk_width+j_pos] > thres_mean_max 
        && sceneBlock.blockDiff[i_pos*blk_width+j_pos] < thres_diff)
        num_bright_flat++;
      if (sceneBlock.blockMean[i_pos*blk_width+j_pos] > thres_mean_max 
        && sceneBlock.blockDiff[i_pos*blk_width+j_pos] >= thres_diff)
        num_bright_sharp++;
      if (sceneBlock.blockMean[i_pos*blk_width+j_pos] < thres_mean_min 
        && sceneBlock.blockStd[i_pos*blk_width+j_pos] < thres_std
        && sceneBlock.blockDiff[i_pos*blk_width+j_pos] > 0.0 )
        num_dark_flat++;
      if (sceneBlock.blockMean[i_pos*blk_width+j_pos] < thres_mean_min
        && sceneBlock.blockStd[i_pos*blk_width+j_pos] >= thres_std
        && sceneBlock.blockDiff[i_pos*blk_width+j_pos] > 0.0 )
        num_dark_sharp++;
      num_total++;
    }
  }
  num_bright_flat	/= (double)blk_width*blk_height;
  num_bright_sharp	/= (double)blk_width*blk_height;
  num_dark_flat		/= (double)blk_width*blk_height;
  num_dark_sharp	/= (double)blk_width*blk_height;
  
  framestats->num_bright_flat = num_bright_flat;
  framestats->num_bright_sharp = num_bright_sharp;
  framestats->num_dark_flat = num_dark_flat;
  framestats->num_dark_sharp = num_dark_sharp;
}
  
#endif

/**
- Automatic algorithm (m37267) to determine luma reshaping model based on scene analysis 
\param   inp            pointer to image frame to be analyzed
\param   metaData       struct containing basic scene info
\retval  org_CL         min sample value 
\retval  org_CH         max sample value
\retval  scale_factor   derived luma scaling factor
\retval  alpha          derived power alpha
*/
void Reshaper::sceneAnalysis(const Frame *inp, double *org_CL, double *org_CH, double *scale_factor, double *alpha, SceneData metaData)
{
  double YfactorTab[5] = {0.80, 0.84, 0.90, 0.94, 0.90};
  double CfactorTab[2][2] = { {1.2344, 1.7913}, {1.3991, 2.1735} };		
  double P_factor, T_factor, I_factor;
  double power_alpha;
  double meta_max = (double)(MAX_LUMA_RESHAPING_LUT_SIZE-1);
  int i, idx;
  double org_I_CH=0.0, org_I_CL=0.0, org_P_CH, org_P_CL, org_T_CH, org_T_CL;
  int max_value, min_value, tmp_CL, tmp_CH;
#if MODE0_UPDATE
  double ratio = 1.03;
#else
  double ratio = 1.05;
#endif

  double *Idata = new double[inp->m_compSize[0]];
  int blk_size   = 16;
  int blk_width  = (int)dRound((double)inp->m_width[0] / blk_size);
  int blk_height = (int)dRound((double)inp->m_height[0] / blk_size);
  SceneBlkData sceneBlock;
  sceneBlock.blockSize  = blk_size;
  sceneBlock.blockWidth  = blk_width;
  sceneBlock.blockHeight = blk_height;
  sceneBlock.blockDiff = new double[blk_width*blk_height];
  sceneBlock.blockStd  = new double[blk_width*blk_height];
  sceneBlock.blockMean = new double[blk_width*blk_height];
  double thres_mean_min = 50.0, thres_mean_max = 400.0;
  double thres_diff = 256.0, thres_std = 16.0;
  double num_bright_flat = 0.0, num_bright_sharp = 0.0, num_dark_flat = 0.0, num_dark_sharp = 0.0, num_total = 0.0;
#if MODE0_UPDATE
  double sum_bright = 0.0, sum_dark = 0.0;
  double thres_bright_l1 = 0.3, thres_bright_l2 = 0.05, thres_dark_l1 = 0.3, thres_dark_l2 = 0.02;
  double thres_flat_l1 = 0.55, thres_flat_l2 = 0.3, thres_sharp_l1 = 0.1, thres_sharp_l2 = 0.2;
#else
  double thres_sharp_l1 = 0.15, thres_sharp_l2 = 0.03, thres_flat_l1 = 0.2, thres_flat_l2 = 0.5;
#endif
  int i_pos, j_pos;

  max_value = inp->m_maxPelValue[0];
  min_value = 0;

  tmp_CL = max_value;
  tmp_CH = min_value;
  for (i = 0; i < inp->m_compSize[0]; i++) 
  {
    idx = inp->m_ui16Comp[0][i];
    if (idx < tmp_CL) 
      tmp_CL = idx;
    if (idx > tmp_CH )
      tmp_CH = idx;
  }

  assert (inp->m_sampleRange <= SR_FULL);
  if (inp->m_sampleRange == SR_STANDARD)
  {
	  org_I_CL = ((double)tmp_CL/4.0-16.0)/219.0; //hardcode to deal with 10bit video for now
	  org_I_CH = ((double)tmp_CH/4.0-16.0)/219.0;
	  thres_mean_min = 106.8152;
	  thres_mean_max = 406.522;  
	  thres_diff = 219.214; 
	  thres_std = 13.7; 
  }
  else if (inp->m_sampleRange == SR_FULL)
  {
	  org_I_CL = (double)tmp_CL/meta_max; 
	  org_I_CH = (double)tmp_CH/meta_max;
	  thres_mean_min = 50.0, 
	  thres_mean_max = 400.0;
	  thres_diff = 256.0; 
	  thres_std = 16.0;
  }

  tmp_CL = max_value;
  tmp_CH = min_value;
  for (i = 0; i < inp->m_compSize[1]; i++) 
  {
    idx = inp->m_ui16Comp[1][i];
    if (idx < tmp_CL) 
      tmp_CL = idx;
    if (idx > tmp_CH )
      tmp_CH = idx;
  }
  org_P_CL = (double)tmp_CL/meta_max;
  org_P_CH = (double)tmp_CH/meta_max;

  tmp_CL = max_value;
  tmp_CH = min_value;
  for (i = 0; i < inp->m_compSize[2]; i++) 
  {
    idx = inp->m_ui16Comp[2][i];
    if (idx < tmp_CL) 
      tmp_CL = idx;
    if (idx > tmp_CH )
      tmp_CH = idx;
  }
  org_T_CL = (double)tmp_CL/meta_max;
  org_T_CH = (double)tmp_CH/meta_max;

  //increase the range for full scene
  org_I_CL = dClip(org_I_CL/ratio, 0.0f, 1.0f);
  org_I_CH = dClip(org_I_CH*ratio, 0.0f, 1.0f);
  org_P_CL = dClip(org_P_CL/ratio, 0.0f, 1.0f);
  org_P_CH = dClip(org_P_CH*ratio, 0.0f, 1.0f);
  org_T_CL = dClip(org_T_CL/ratio, 0.0f, 1.0f);
  org_T_CH = dClip(org_T_CH*ratio, 0.0f, 1.0f);

  // determine the parameter for forward reshaping
  if (metaData.sceneIsFades == 0)
  {
    // reading luma component 
    for (i = 0; i < inp->m_compSize[0]; i++) 
    {
      Idata[i] = (double)inp->m_ui16Comp[0][i];
    }

    blockproc(Idata, inp->m_width[0], inp->m_height[0], sceneBlock);

    // count class numbers
    num_bright_flat	= 0.0;
    num_bright_sharp	= 0.0;
    num_dark_flat		= 0.0;
    num_dark_sharp	= 0.0;
    num_total			= 0.0;
    for (i_pos=0; i_pos<blk_height; i_pos++)
    {
      for (j_pos=0; j_pos<blk_width; j_pos++)
      {
        if (sceneBlock.blockMean[i_pos*blk_width+j_pos] > thres_mean_max 
          && sceneBlock.blockDiff[i_pos*blk_width+j_pos] < thres_diff)
          num_bright_flat++;
        if (sceneBlock.blockMean[i_pos*blk_width+j_pos] > thres_mean_max 
          && sceneBlock.blockDiff[i_pos*blk_width+j_pos] >= thres_diff)
          num_bright_sharp++;
        if (sceneBlock.blockMean[i_pos*blk_width+j_pos] < thres_mean_min 
          && sceneBlock.blockStd[i_pos*blk_width+j_pos] < thres_std
          && sceneBlock.blockDiff[i_pos*blk_width+j_pos] > 0.0 )
          num_dark_flat++;
        if (sceneBlock.blockMean[i_pos*blk_width+j_pos] < thres_mean_min
          && sceneBlock.blockStd[i_pos*blk_width+j_pos] >= thres_std
          && sceneBlock.blockDiff[i_pos*blk_width+j_pos] > 0.0 )
          num_dark_sharp++;
        num_total++;
      }
    }
    num_bright_flat	/= (double)blk_width*blk_height;
    num_bright_sharp	/= (double)blk_width*blk_height;
    num_dark_flat		/= (double)blk_width*blk_height;
    num_dark_sharp	/= (double)blk_width*blk_height;
  }

  // replace original parameters
  I_factor = (double)1.0;
  P_factor = (double)1.0;
  T_factor = (double)1.0;
  power_alpha = 1.0;

#if MODE0_UPDATE
  sum_bright = num_bright_flat	+ num_bright_sharp;
  sum_dark   = num_dark_flat	+ num_dark_sharp;
  if(sum_bright > thres_bright_l1)				
  {
	  if (num_bright_flat > thres_flat_l1 && num_bright_sharp > thres_sharp_l1)
		  power_alpha = 2.2;
	  else if (num_bright_flat > thres_flat_l2 && num_bright_sharp > thres_sharp_l2)
		  power_alpha = 2.0;
	  else if (sum_dark > thres_dark_l2)
		  power_alpha = 1.7;
	  else
		  power_alpha = 1.2;
  }
  else if (sum_dark > thres_dark_l1)			
  {
	  power_alpha = 1.0;
  }
  else
  {
	  if (sum_dark > thres_dark_l2 && sum_bright > thres_bright_l2 && num_bright_sharp > 0.1*thres_sharp_l2)
		  power_alpha = 1.2;
	  else
		  power_alpha = 1.0;
  }
#else
  if (metaData.sceneIsFades == 1)														
    power_alpha = 1.0;
  else if (num_bright_sharp > thres_sharp_l1 && num_bright_flat > thres_flat_l2)		
    power_alpha = 2.2;
  else if (num_bright_sharp > thres_sharp_l1 && num_bright_flat > thres_flat_l1)	 
    power_alpha = 2.0;
  else if (num_dark_sharp > 0.1 * thres_sharp_l1 && num_dark_flat > 0.1 * thres_flat_l2)
    power_alpha = 1.7;
  else if (num_dark_sharp > 0.1 * thres_sharp_l1 && num_dark_flat > 0.1 * thres_flat_l1)
    power_alpha = 1.2;
  else if (num_bright_sharp > thres_sharp_l2 && num_bright_flat > thres_flat_l1)	 
    power_alpha = 1.2;
  else																					
    power_alpha = 1.0;
#endif

  if (metaData.srcPrimary == 0)		// 709 content
  {
    if (power_alpha > 2.0)			
    {
      I_factor = YfactorTab[4];		
      P_factor = CfactorTab[1][0];
      T_factor = CfactorTab[1][1];
    }
    else if (power_alpha > 1.7)		
    {
      I_factor = YfactorTab[4];		
      P_factor = CfactorTab[1][0];
      T_factor = CfactorTab[1][1];
    }
    else if (power_alpha > 1.0)		
    {
      I_factor = YfactorTab[4];		
      P_factor = CfactorTab[1][0];
      T_factor = CfactorTab[1][1];
    }
    else								
    {
      I_factor = YfactorTab[4];		
      P_factor = CfactorTab[1][0];
      T_factor = CfactorTab[1][1];
    }
  }
  else if (metaData.srcPrimary == 3)	// P3 content
  {
    if (power_alpha >= 1.7)			
    {
      I_factor = YfactorTab[4];		 
      P_factor = CfactorTab[0][0];
      T_factor = CfactorTab[0][1];
    }
    else
    {
      I_factor = YfactorTab[4];		 
      P_factor = CfactorTab[0][0];
      T_factor = CfactorTab[0][1];
    }
  }

  //asign value
  org_CL[0] = org_I_CL;
  org_CL[1] = org_P_CL;
  org_CL[2] = org_T_CL;

  org_CH[0] = org_I_CH;
  org_CH[1] = org_P_CH;
  org_CH[2] = org_T_CH;

  scale_factor[0] = I_factor;
  scale_factor[1] = P_factor;
  scale_factor[2] = T_factor;
  *alpha = power_alpha;

}

#if LUMA_RESHAPE_METHOD
double Reshaper::PQ2L(double PQ)
{
  double L;

  // Parameters used for ST-2084
  double ni = 1/0.15930176;
  double mi = 1/78.84375;
  double c1 = 0.8359375;
  double c2 = 18.8515625;
  double c3 = 18.6875;

  //ST-2084 space is not defined for PQ<0 or PQ>1. 
  PQ = dMax(0.0f,dMin(1.0f,PQ));
  L = pow(PQ, mi);
  L = 10000 * pow(dMax(L-c1,0)/(c2-c3*L), ni);
  L = dMax(0,dMin(10000,L));
  return(L);
}

double Reshaper::L2PQ(double L)
{
  // Parameters used for ST-2084
  double n = 0.15930176;
  double m = 78.84375;
  double c1 = 0.8359375;
  double c2 = 18.8515625;
  double c3 = 18.6875;

  // ST-2084 is not defined for L<0 or L>10000
  L = dMax(0.0f,dMin(10000.0,L));

  //Calculate PQ code values (no rounding)
  double Yn = pow((L/10000.0), n);
  double PQ = pow(((c1+c2*Yn) / (1.0+c3*Yn)),m);
  PQ = dMax(0.0f,dMin(1.0f,PQ));

  return PQ;
}


/**
- Function to generate unified forward reshaping parameters  
\param   inp            pointer to image frame to be processed
\param   iPos           POC
\param   reshaping      struct containing unified reshaping model 
*/
void Reshaper::initReshaping(double *power_alpha, double *scalefactor, int *content_profile, double *I_CH, double *I_CL, FrameStats frame_stats, int poc_dist, int *opt)
{
  double old_alpha = *power_alpha;
  double alpha, scale_factor, old_org_I_CH, old_org_I_CL; 
  double org_I_CH, org_I_CL;
  double sum_dark, sum_bright;

  double ratio=1.03;
  
  *opt = 1;

  old_org_I_CL = PQ2L(*I_CL*ratio);
  old_org_I_CH = PQ2L(*I_CH/ratio);

  org_I_CL = old_org_I_CL;
  org_I_CH = old_org_I_CH;
  
  alpha = *power_alpha;
  scale_factor = *scalefactor;
    
  //intialize reshaper parameter with default profiles based on content charateristics
  sum_bright = frame_stats.num_bright_flat+frame_stats.num_bright_sharp;
  sum_dark = frame_stats.num_dark_flat + frame_stats.num_dark_sharp;

  if (poc_dist<=5 && poc_dist>0)
  {
    scale_factor=0.8;
    alpha=1.0;
#if RESHAPER_JOINT_OPT
    int temp[15]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
  }
  else
  {
    if (frame_stats.num_bright_flat >=0.25) 
    {
      if (old_alpha==1.2)
      {
        if (frame_stats.num_bright_flat<0.3)
        {
          int profile[15] = {1, 2, 2, 4, 5, 5, 5, 4, 5, 5, 4, 4, 3, 2, 0};
          memcpy(content_profile, profile, 15*sizeof(int));
          scale_factor=1.0;
          alpha=1.4;
#if RESHAPER_JOINT_OPT
          int temp[15]= {-3, -1, -2, -1, 2, 3, 2, 1, 2, 2, 1, 1, 0, -1, -3};
          memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
        }
        else if (frame_stats.num_bright_sharp < 0.1)
        {
          int profile[15] = {0, 1, 2, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 0};
          memcpy(content_profile, profile, 15*sizeof(int));
          scale_factor=1.0;
          alpha=2.0; 
#if RESHAPER_JOINT_OPT
          int temp[15]= {-3, -2, 0, 2, 2, 3, 3, 1, 1, 2, 1, 2, 3, 0, -3};
          memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
        }
		else 
		{
            int profile[15] = {0, 1, 1, 2, 5, 5, 5, 4, 5, 5, 5, 5, 2, 2, 0};
            memcpy(content_profile, profile, 15*sizeof(int));
            scale_factor=0.9;
            alpha=2.4;
            org_I_CL *= 3.0;   
#if RESHAPER_JOINT_OPT
            int temp[15]= {-3, -3, -3, -2, -1, 0, 0, -1, 2, 3, 3, 2, -2, -2, -3};
            memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
        }
      }
      else if (old_alpha==1.7)
      {
        int profile[15] = {1, 1, 3, 3, 2, 5, 5, 4, 4, 5, 5, 5, 5, 5, 0};
        memcpy(content_profile, profile, 15*sizeof(int));
        scale_factor=0.92;
        alpha=2.6;
#if RESHAPER_JOINT_OPT
        int temp[15]= {-3, -3, -1, -1, -2, 1, 1, 0, 0, 1, 2, 3, 3, 3, -3};
        memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
      }
      else if(old_alpha==2.0)
      {
        int profile[15] = {0, 3, 2, 2, 2, 2, 2, 2, 4, 4, 4, 4, 5, 5, 0};
        memcpy(content_profile, profile, 15*sizeof(int));
        scale_factor=1.0;
        alpha=2.0;
#if RESHAPER_JOINT_OPT
        int temp[15]= {-3, -1, -2, -2, -1, -2, -1, -2, 3, 2, 2, 3, 3, 3, -3};
        memcpy(m_lumaDQP, temp, 15*sizeof(int));
#endif
      }
      else 
      {
        if (frame_stats.num_bright_flat < 0.6)
        {
			int profile[15] = {0, 1, 2, 2, 2, 2, 2, 2, 2, 5, 5, 4, 5, 5, 0};
			memcpy(content_profile, profile, 15*sizeof(int));
			scale_factor=1.0;
			alpha=2.4;
#if RESHAPER_JOINT_OPT
            int temp[15]= {-3, -2, -2, -2, -1, -1, -1, -2, -1, 2, 2, 1, 3, 3, -3};
            memcpy(m_lumaDQP, temp, 15*sizeof(int));
#endif
        }
        else if (frame_stats.num_bright_flat < 0.75)
        {
          int profile[15] = {0, 1, 3, 3, 2, 2, 2, 2, 5, 5, 5, 5, 5, 3, 0};
          memcpy(content_profile, profile, 15*sizeof(int));
          scale_factor=1.0;
          alpha=2.2;
#if RESHAPER_JOINT_OPT
          int temp[15]= {-3, -3, -1, -1, -2, -2, -2, -2, 3, 2, 3, 3, 3, 1, -3};
          memcpy(m_lumaDQP, temp, 15*sizeof(int));
#endif
        }
        else
        {
          int profile[15] = {0, 1, 3, 3, 4, 5, 5, 5, 5, 4, 5, 5, 4, 5, 0};
          memcpy(content_profile, profile, 15*sizeof(int));
          scale_factor=0.9;
          alpha=2.2;
          org_I_CL*=1.45;
#if RESHAPER_JOINT_OPT
          int temp[15]= {-3, -3, -2, -2, -1, 0, 0, 0, 1, 1, 2, 3, 1, 2, -3};
          memcpy(m_lumaDQP, temp, 15*sizeof(int));
#endif
        }
      }
    }
    else 
    {
      if (frame_stats.num_dark_flat > 0.4)
      {
        if (frame_stats.num_dark_flat > 0.7)
        {
          int profile[15] = {2, 5, 5, 5, 4, 5, 4, 4, 5, 3, 3, 5, 5, 3, 0};
          memcpy(content_profile, profile, 15*sizeof(int));
          scale_factor=1.0;
          alpha=1.2;
#if RESHAPER_JOINT_OPT
          int temp[15]= {-1, 1, 2, 2, 1, 2, 1, 1, 2, 0, 0, 3, 3, 0, -3};
          memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
        }
        else
        {
          scale_factor=0.8;
          alpha=1.0;
#if RESHAPER_JOINT_OPT
          int temp[15]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
          memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
        }
      }
      else
      {
        if (old_alpha == 1.2)
        {
          if (frame_stats.num_dark_flat<0.05)
          {
            int profile[15] = {2, 2, 5, 4, 5, 5, 5, 4, 5, 4, 4, 5, 2, 2, 2};
            memcpy(content_profile, profile, 15*sizeof(int));
            scale_factor=0.8;
            alpha=1.0;
#if RESHAPER_JOINT_OPT
            int temp[15]= {-2, -2, 1, 0, 2, 3, 2, 1, 2, 1, 1, 3, 1, 0, 0};
            memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
          }
          else
          {
            int profile[15] = {1, 4, 5, 5, 5, 5, 4, 4, 5, 3, 3, 3, 3, 3, 0};
            memcpy(content_profile, profile, 15*sizeof(int));
            scale_factor=1.0;
            alpha=1.3;
#if RESHAPER_JOINT_OPT
            int temp[15]= {-3, 1, 1, 3, 3, 3, 2, 2, 3, 1, 0, 0, 0, 0, -3};
            memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
           }
        }
        else  
        {
          if (old_org_I_CH < 3000.0)
          {
            if (sum_dark<0.01)
            {
              int profile[15] = {2, 5, 5, 5, 5, 5, 4, 4, 5, 4, 4, 4, 4, 0, 0};
              memcpy(content_profile, profile, 15*sizeof(int));
              scale_factor=1.0;
              alpha=1.0;
              *opt = 0;
             }
            else if (sum_dark<0.05)
            {
              int profile[15] = {2, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 2, 2, 2, 0};
              memcpy(content_profile, profile, 15*sizeof(int));
              scale_factor=1.0;
              alpha=1.0;
              org_I_CH =2500.0;  
              *opt = 0;
            }
            else if (sum_dark<0.10)
            {
              int profile[15] = {1, 2, 5, 5, 5, 5, 4, 4, 4, 3, 2, 3, 2, 0, 0};
              memcpy(content_profile, profile, 15*sizeof(int));
              scale_factor=1.0;
              alpha=1.2;
#if RESHAPER_JOINT_OPT
              int temp[15]= {-2, -1, 2, 3, 2, 3, 2, 1, 2, 0, -1, 0, -1, -3, -3};
              memcpy(m_lumaDQP, temp, 15*sizeof(int));
#endif
            }
            else
            {
              scale_factor=0.8;
              alpha=1.0;
#if RESHAPER_JOINT_OPT
              int temp[15]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
              memcpy(m_lumaDQP, temp, 15*sizeof(int)); 
#endif
            }
          }
          else 
          {
            if (sum_dark>0.01)
            {
              int profile[15] = {2, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 2, 2, 2, 0};
              memcpy(content_profile, profile, 15*sizeof(int));
              scale_factor=1.0;
              alpha=1.0;
              *opt = 0;
            }
            else
            {
              if (sum_bright<0.15)
              {
                int profile[15] = {1, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 5, 5, 2, 0};
                memcpy(content_profile, profile, 15*sizeof(int));
                scale_factor=1.0;
                alpha=1.0;
                *opt = 0;
              }
              else
              {
                int profile[15] = {1, 2, 5, 5, 5, 5, 5, 4, 4, 4, 4, 5, 5, 2, 0};
                memcpy(content_profile, profile, 15*sizeof(int));
                scale_factor=1.0;
                alpha=1.0;
                *opt = 0;
              }
            }
          }
        }

      }
    }
  }
  *power_alpha = alpha;
  *scalefactor = scale_factor;
  *I_CH = org_I_CH;
  *I_CL = org_I_CL;
}

void Reshaper::lumaReshaping(const Frame *inp, double *norm_LUT, double old_alpha, double old_org_I_CL, double old_org_I_CH, FrameStats frame_stats, int poc_dist, int *opt)
{
  int tot_cw = MAX_LUMA_RESHAPING_LUT_SIZE;
  int hist_bins = MAX_LUMA_HIST_BINS;
  int hist_lens = tot_cw/hist_bins;
  int num_bands = 15;
  double controls [16] = {0,0.0625610948191591,0.156402737047899,0.250244379276637,0.312805474095797,0.344086021505374,0.437927663734117,0.469208211143690,0.500488758553276,0.594330400782013,0.625610948191590,0.688172043010759,0.750733137829903,0.813294232649075,0.907135874877799,1};
  int i, j;
  int X_bins[MAX_LUMA_HIST_BINS+1];
  double x_bins[MAX_LUMA_HIST_BINS+1];
  int num_bins[15];
  int sum_bins;
  int CW_bins[MAX_LUMA_HIST_BINS], CW_bins_pow[MAX_LUMA_HIST_BINS], CW_bins_preassign[MAX_LUMA_HIST_BINS];
  double precision=0.000001;
  
  int last_idx_numbins[15];
  int min_start_bin_idx, max_end_bin_idx;
  int bin_profile[MAX_LUMA_HIST_BINS];
  int c_profile;
  int used_codewords, budget_cw, residue_cw;
  double CW_bins_LUT_all[MAX_LUMA_RESHAPING_LUT_SIZE], Y_LUT_all[MAX_LUMA_RESHAPING_LUT_SIZE];
  int Forward_LUT[MAX_LUMA_RESHAPING_LUT_SIZE], Backward_LUT[MAX_LUMA_RESHAPING_LUT_SIZE];
  double alpha, org_I_CL, org_I_CH, scale_factor=1.0;
  int content_profile[15] = {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};
  double ratio=1.03;
  
  *opt=1;

  alpha = old_alpha;
  org_I_CH = old_org_I_CH;
  org_I_CL = old_org_I_CL;
  initReshaping(&alpha, &scale_factor, content_profile, &org_I_CH, & org_I_CL, frame_stats, poc_dist, opt);

  for (i=0; i<=MAX_LUMA_HIST_BINS; i++)
  {
    X_bins[i] = i*hist_lens;
    x_bins[i] = (double)X_bins[i]/(double)tot_cw;
  }

  memset(num_bins, 0, 15*sizeof(int));
   sum_bins = 0;
  for (i=0; i<num_bands; i++)
  {
    for (j=0; j<MAX_LUMA_HIST_BINS; j++)
    {
      if (x_bins[j+1]>=controls[i] && x_bins[j+1] < controls[i+1])
        num_bins[i]++;
    }
    sum_bins += num_bins[i];
  }
  if (sum_bins<MAX_LUMA_HIST_BINS)
    num_bins[14] += MAX_LUMA_HIST_BINS-sum_bins;
  
  for (i=0; i<MAX_LUMA_HIST_BINS; i++)
  {
    CW_bins[i] = hist_lens;
  }
 
  for (i=0; i<MAX_LUMA_HIST_BINS; i++)
  {
    double y_bins_pow = pow(x_bins[i],alpha);
    double y_bins_pow_plus1 = pow(x_bins[i+1],alpha);
    CW_bins_pow[i] = double2IntCeil((y_bins_pow_plus1-y_bins_pow)*tot_cw-precision);
  }

  last_idx_numbins[0] = num_bins[0];
  for (i=1; i<num_bands; i++)
  {
    last_idx_numbins[i] =  last_idx_numbins[i-1] + num_bins[i];
  }
 
  min_start_bin_idx = int(floor(L2PQ(org_I_CL)*hist_bins));
  max_end_bin_idx = int(floor(L2PQ(org_I_CH)*hist_bins));



  for (i=0; i<hist_bins; i++)
  {
    for (j=0; j<num_bands; j++)
    {
        if (i < last_idx_numbins[j])
        {
            bin_profile[i] = content_profile[j];
            break;
        }
        else
            continue;
    }
    if (i < min_start_bin_idx || i > max_end_bin_idx)
        bin_profile[i] = 0;
  }

  c_profile = 0;
  for (j=0; j<num_bands; j++)
  {
    if (c_profile < content_profile[j])
      c_profile = content_profile[j];
  }
  
  used_codewords = 0;
  for (i=0; i<hist_bins; i++)
  {
    int profile = bin_profile[i];
    if (profile == 0)
      CW_bins_preassign[i] = 0;
    else if (profile == 1)
      CW_bins_preassign[i] = (CW_bins[i]+1)/2;
    else if (profile == 2)
      CW_bins_preassign[i]  = CW_bins[i];
    else if (profile == 3)
      CW_bins_preassign[i]  = (CW_bins_pow[i]+CW_bins[i]+1)/2;
    else 
      CW_bins_preassign[i]  = iMax(CW_bins_pow[i],CW_bins[i]); 

    used_codewords +=  CW_bins_preassign[i];
  }

  if (used_codewords > tot_cw)  
  {
    for (i=0; i<hist_bins; i++)
    {
      double sum_ratio = used_codewords/tot_cw;
      CW_bins_preassign[i] = (int)((double)CW_bins_preassign[i]/sum_ratio);
    }
  }
  else
  {
     budget_cw = tot_cw - used_codewords;
     residue_cw = budget_cw;
     while (residue_cw!=0)
     {
       for (i=hist_bins-1; i>=0; i--)
       {
         if (bin_profile[i] == c_profile)
         {
           CW_bins_preassign[i] += 1;
           residue_cw--;
         }
         if (residue_cw==0)
           break;
       }
     }
  }

  sum_bins = 0;
  for (i=0; i<hist_bins; i++)
  {
    sum_bins += CW_bins_preassign[i];
  }

  assert(sum_bins == tot_cw);

  for (i=0; i<hist_bins; i++)
  {
    double temp = (double)CW_bins_preassign[i]/hist_lens;
    for (j=0; j<hist_lens; j++)
    {
      CW_bins_LUT_all[i*hist_lens+j] = temp;
    }
  }
  
  Y_LUT_all[0] = CW_bins_LUT_all[0];
  for (i=1; i<tot_cw; i++)
  {
    Y_LUT_all[i] = Y_LUT_all[i-1] + CW_bins_LUT_all[i];
  }

  for (i=0; i<tot_cw; i++)
  {
    Forward_LUT[i] = iClip( (int)(Y_LUT_all[i]*scale_factor+0.5), 0, tot_cw-1);
  }

  BuildForwardReshapingLUT(Forward_LUT, Backward_LUT,tot_cw);

  for (i=0; i<tot_cw; i++)
  {
    norm_LUT[i] = (double)Backward_LUT[i]/(double)(tot_cw-1);
  }

}
#endif

#if NO_INTRA_PERIOD_UPDATE
void Reshaper::storePrevLumaMetaData(UnifiedLumaReshapeInfo *PrevLumaReshape,UnifiedLumaReshapeInfo *pUnifiedLumaReshape)
{
	  PrevLumaReshape->luma_bit_depth = pUnifiedLumaReshape->luma_bit_depth;
	  PrevLumaReshape->reshape_num_ranges_minus1 = pUnifiedLumaReshape->reshape_num_ranges_minus1;
	  PrevLumaReshape->reshape_equal_ranges = pUnifiedLumaReshape->reshape_equal_ranges;
	  PrevLumaReshape->reshape_global_offset_val = pUnifiedLumaReshape->reshape_global_offset_val;
	  for (int s = 0; s < pUnifiedLumaReshape->reshape_num_ranges_minus1 + 1; s++)
		  PrevLumaReshape->reshape_range_val[s] = pUnifiedLumaReshape->reshape_range_val[s];
	  PrevLumaReshape->coeff_log2_offset_minus2 = pUnifiedLumaReshape->coeff_log2_offset_minus2;
	  PrevLumaReshape->reshape_continuity_flag = pUnifiedLumaReshape->reshape_continuity_flag;
	  for (int s = 0; s < pUnifiedLumaReshape->reshape_num_ranges_minus1 + 2; s++)
		PrevLumaReshape->constantCoeffSignalled[s] = pUnifiedLumaReshape->constantCoeffSignalled[s];
		if(pUnifiedLumaReshape->reshape_continuity_flag)
	  PrevLumaReshape->linearCoeffSignalled = pUnifiedLumaReshape->linearCoeffSignalled;
}

void Reshaper::compareLumaMetaData(UnifiedLumaReshapeInfo *PrevLumaReshape,UnifiedLumaReshapeInfo *CurrLumaReshape, int * isUpdate)
{
	if (PrevLumaReshape->luma_bit_depth != CurrLumaReshape->luma_bit_depth)
	{
		*isUpdate = 1;
		return;
	}
	if (PrevLumaReshape->reshape_num_ranges_minus1 != CurrLumaReshape->reshape_num_ranges_minus1)
	{
		*isUpdate = 1;
		return;
	}
	if (PrevLumaReshape->reshape_equal_ranges != CurrLumaReshape->reshape_equal_ranges)
	{
		*isUpdate = 1;
		return;
	}
	if (PrevLumaReshape->reshape_global_offset_val != CurrLumaReshape->reshape_global_offset_val)
	{
		*isUpdate = 1;
		return;
	}			

	for (int s = 0; s < CurrLumaReshape->reshape_num_ranges_minus1 + 1; s++)
		if (PrevLumaReshape->reshape_range_val[s]!= CurrLumaReshape->reshape_range_val[s])
		{
			*isUpdate = 1;
			return;
		}

	if (PrevLumaReshape->coeff_log2_offset_minus2 != CurrLumaReshape->coeff_log2_offset_minus2)
	{
		*isUpdate = 1;
		return;
	}
	if (PrevLumaReshape->reshape_continuity_flag != CurrLumaReshape->reshape_continuity_flag)
	{
		*isUpdate = 1;
		return;
	}

	for (int s = 0; s < CurrLumaReshape->reshape_num_ranges_minus1 + 2; s++)
		if (PrevLumaReshape->constantCoeffSignalled[s]!= CurrLumaReshape->constantCoeffSignalled[s])
		{
			*isUpdate = 1;
			return;
		}

	if (PrevLumaReshape->reshape_continuity_flag)
	{
		if (PrevLumaReshape->linearCoeffSignalled != CurrLumaReshape->linearCoeffSignalled)
		{
			*isUpdate = 1;
			return;
		}
	}
	*isUpdate = 0;
	return;
}
#endif


/**
- Function to generate unified forward reshaping parameters  
\param   inp            pointer to image frame to be processed
\param   iPos           POC
\param   reshaping      struct containing unified reshaping model 
*/
void Reshaper::forwardReshaping ( const Frame *inp, int iPos, ReshapingInfo reshaping) {
  double P_factor, T_factor, I_factor;
  double power_alpha;
  int idx, out_idx;
  double I_range, I_CH=0.0, I_CL=0.0, P_range, P_CH, P_CL, T_range, T_CH, T_CL;
  double org_I_CH=0.0, org_I_CL=0.0, org_P_CH, org_P_CL, org_T_CH, org_T_CL;
  double a1_I, a0_I, a1_P, a0_P, a1_T, a0_T;
  double meta_max = (double)(MAX_LUMA_RESHAPING_LUT_SIZE-1);
  double dTemp, norm_val=0.0;
#if MODE0_UPDATE
  int isParameterUpdate=0;
  double prev_I_CH, diff_I_CH, thres_diff_I_CH=0.03;
#endif
#if LUMA_RESHAPE_METHOD
 Int opt=1;
#endif

#if TCH_CROSSPLANE_CRCB
  double  logY_lut[MAX_LUMA_RESHAPING_LUT_SIZE];
#endif
  double  target_lut[MAX_LUMA_RESHAPING_LUT_SIZE];
  int     index[MAX_LUMA_RESHAPING_LUT_SIZE];

  double org_CL[3], org_CH[3], scale_factor[3];
#if NO_INTRA_PERIOD_UPDATE
  FrameStats curr_frame_stats;
#endif

#if MODE0_UPDATE
  //analyze frame data to decide whether to update
  m_sceneData.srcPrimary = reshaping.m_srcPrimary;
  prev_I_CH = m_sceneData.prev_org_I_CH;
  frameAnalysis(inp);

#if NO_INTRA_PERIOD_UPDATE
  if (inp->m_frameNo == 0)
  {
	  isParameterUpdate = 1;
      genFrameStats(inp, &curr_frame_stats);
      m_sceneDataOld.prev_frameNo = 0;
  }
  else if (inp->m_frameNo % reshaping.m_intraPeriod == 0)
  {
    genFrameStats(inp, &curr_frame_stats);
    if (fabs(curr_frame_stats.num_bright_flat-m_sceneDataOld.frameStats.num_bright_flat)>=0.25)
    {
      isParameterUpdate = 1;
    }
  }
#else
  if (inp->m_frameNo % reshaping.m_intraPeriod == 0)
  {
	  isParameterUpdate = 1;	  
  }
#endif
  else 
  {
	diff_I_CH = abs(m_sceneData.curr_org_I_CH - prev_I_CH)/prev_I_CH;

	if(diff_I_CH > thres_diff_I_CH)
	{
		isParameterUpdate = 1;
#if NO_INTRA_PERIOD_UPDATE
        genFrameStats(inp, &curr_frame_stats);
#endif
	}
  }
#endif

#if MODE0_UPDATE
if (((isParameterUpdate) &&!reshaping.m_reshapeMode)||(m_recomputeReshapeParams && reshaping.m_reshapeMode))
  {
#else
  if (reshaping.m_reshapeMode  == 0)
  {
    m_recomputeReshapeParams = 0;
    if (m_sceneNum < m_numOfScenes)
      if (iPos >= m_sceneData[m_sceneNum].sceneFirstFno)
        m_recomputeReshapeParams = 1;
  }

#if TCH_CROSSPLANE_CRCB
#if TCH_AUTO_WRITE_LUT
  if (m_recomputeReshapeParams)
#else
    m_recomputeReshapeParams = 0;
    if (  (iPos >= m_sceneData[m_sceneNum].sceneFirstFno)  && (m_sceneNum < m_numOfScenes)  )
#endif
#else
  if (m_sceneNum < m_numOfScenes)
    if (iPos >= m_sceneData[m_sceneNum].sceneFirstFno)
#endif
  {
#if TCH_CROSSPLANE_CRCB
#if !TCH_AUTO_WRITE_LUT
      m_recomputeReshapeParams = 1;
#endif
#endif
    while ( (iPos > m_sceneData[m_sceneNum].sceneFirstFno) && (m_sceneNum < m_numOfScenes-1))
    {
      m_sceneNum++;
    }

    if (iPos < m_sceneData[m_sceneNum].sceneFirstFno)
    {
      m_sceneNum--;
    }
#endif

    if (reshaping.m_reshapeMode  == 0) {
      //step1: analyze scene to derive automatic luma reshaping parameters 
#if MODE0_UPDATE
	sceneAnalysis(inp, org_CL, org_CH, scale_factor, &power_alpha, m_sceneData);
#else
    sceneAnalysis(inp, org_CL, org_CH, scale_factor, &power_alpha, m_sceneData[m_sceneNum]);
#endif

#if NO_INTRA_PERIOD_UPDATE
    m_sceneDataOld.curr_org_I_CH = org_CH[0];
    m_sceneDataOld.frameStats = curr_frame_stats;
#endif

    I_factor = scale_factor[0];
    P_factor = scale_factor[1];
    T_factor = scale_factor[2];

      org_I_CL = org_CL[0];
      org_P_CL = org_CL[1];
      org_T_CL = org_CL[2];

      org_I_CH = org_CH[0];
      org_P_CH = org_CH[1];
      org_T_CH = org_CH[2];

      I_range = I_factor * 1.0f;
      I_CH = dClip ( (0.5 + I_range/2.0), 0.0f, 1.0f);
      I_CL = dClip ( (0.5 - I_range/2.0), 0.0f, 1.0f);

      P_range = P_factor *  ((org_P_CH - org_P_CL)/(org_I_CH - org_I_CL));
      P_CH = dClip ( (0.5 + P_range/2.0), 0.0f, 1.0f);
      P_CL = dClip ( (0.5 - P_range/2.0), 0.0f, 1.0f);

      T_range = T_factor *   ((org_T_CH - org_T_CL)/(org_I_CH - org_I_CL));
      T_CH = dClip ((0.5 + T_range/2.0), 0.0f, 1.0f);
      T_CL = dClip ((0.5 - T_range/2.0), 0.0f, 1.0f);

      a1_I = (org_I_CH - org_I_CL) / (I_CH - I_CL);
      a0_I = org_I_CL - a1_I * I_CL; 

      power_alpha = 1.0/power_alpha;
    }

    ////create forward reshaping LUT for chroma
#if QC_CBCR_m37064
    if (reshaping.m_reshapeMode == 0) {
      setParamsATF(&reshaping);
      for(idx=0; idx < MAX_LUMA_RESHAPING_LUT_SIZE; idx++)
      {
#if MODE0_UPDATE
        m_chromaHistOffsetCb = (1 - 1 / m_chromaHistScaleCb) / 2;
        m_chromaHistOffsetCr = (1 - 1 / m_chromaHistScaleCr) / 2;
#endif
		  switch (inp->m_sampleRange){
		  case SR_FULL:
			  dTemp = dClip( ((double) (idx/meta_max-0.5)*m_chromaHistScaleCb  + m_chromaHistOffsetCb + 0.5), 0.0f, 1.0f);
			  out_idx =  (uint16) fClip(fRound((float)( dTemp * meta_max)), 0.0f, (float)meta_max);
			  m_forwardChromaReshapingLUT[0][idx] = out_idx;
			  dTemp = dClip( ((double) (idx/meta_max-0.5)*m_chromaHistScaleCr  + m_chromaHistOffsetCr + 0.5), 0.0f, 1.0f);
			  out_idx =  (uint16) fClip(fRound((float)( dTemp * meta_max)), 0.0f, (float)meta_max);
			  m_forwardChromaReshapingLUT[1][idx] = out_idx; 
			  break;
		  case SR_STANDARD:
			  dTemp = dClip(((double)((idx / 4.0) - 128) / 224.0 * m_chromaHistScaleCb + m_chromaHistOffsetCb), -0.5f, 0.5f);
			  out_idx = (uint16)fClip(fRound((float)(896.0f *dTemp + 512.0f)), 0.0f, (float)meta_max);
			  m_forwardChromaReshapingLUT[0][idx] = out_idx;
			  dTemp = dClip(((double)((idx / 4.0) - 128) / 224.0 * m_chromaHistScaleCr + m_chromaHistOffsetCr), -0.5f, 0.5f);
			  out_idx = (uint16)fClip(fRound((float)(896.0f *dTemp + 512.0f)), 0.0f, (float)meta_max);
			  m_forwardChromaReshapingLUT[1][idx] = out_idx;
			  break;
		  default:
			fprintf(stderr, "Error setting of YUV output sample range, currently only full range and std range are supported. Process exiting.\n");
    		exit(EXIT_FAILURE);
		  }
      }
    }

    a0_P = m_chromaHistScaleCbInv;
    a1_P = -m_chromaHistOffsetCb;
    a0_T = m_chromaHistScaleCrInv;
    a1_T = -m_chromaHistOffsetCr;
#endif
#if MODE0_UPDATE
    // Put the reverse shaping
  {
    //printf("Check decoder-side LUT\n");
    int reshaperRangeId;
    if(inp->m_sampleRange == 0)
      reshaperRangeId = 1;  // Standard range
    else if(inp->m_sampleRange == 1)
      reshaperRangeId = 0;  // Full range
    else
      reshaperRangeId = 2;  // Unsupported range
    int bitDepth = 10;  // AKR
    int defaultPrecShift = INVERSE_FRAC_PRECISION;

    int reshapeNumRanges[3] = {0,1,1};
    bool reshapeEqualRangesFlag[3] = {0,1,1};
    int reshapeOffsetFracBitDepth= 10;
    int reshapeScaleFracBitDepth= 8;
    int reshapeRangeVal[3][1];
    int reshapeScaleVal[3][1];
    int reshapeGlobalOffsetVal[3];
    reshapeScaleVal[1][0] = evaluateIntRepn(m_chromaHistScaleCbInv, QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
    reshapeScaleVal[2][0] = evaluateIntRepn(m_chromaHistScaleCrInv, QC_ATF_SCALE_BIT_DEPTH, QC_ATF_SCALE_FRAC_BIT_DEPTH);
    reshapeGlobalOffsetVal[1] = evaluateIntRepn(m_chromaHistOffsetCb, QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH);
    reshapeGlobalOffsetVal[2] = evaluateIntRepn(m_chromaHistOffsetCr, QC_ATF_OFFSET_BIT_DEPTH, QC_ATF_OFFSET_FRAC_BIT_DEPTH);
    // -----------------------------
    // Calculation of sample values
    // -----------------------------
    int minSampleVal[3]={0};
    int deltaSampleVal[3] = {0};
    int maxSampleVal[3]={0};
    getMinMaxSampleVal(reshaperRangeId, 10, minSampleVal, maxSampleVal, deltaSampleVal);

    // -----------------------------
    //DI_notes: for LUMA component, a LUT is pre-computed. The following process is only for chroma.

    // --------------------------------------
    // Calculation of scales, input and output ranges
    // --------------------------------------
    std::vector<fixedPrec> rangeValues[3], scaleValues[3];
    // Calculate Range Values
    for(int c = 1; c < 3; c++) {
      rangeValues[c].resize(reshapeNumRanges[c]);
      if(reshapeEqualRangesFlag[c]) {
        for(int i = 0; i < reshapeNumRanges[c]; i++) {
          int val = ((1 << reshapeOffsetFracBitDepth) -1 + ((reshapeNumRanges[c]+1) >> 1)) / reshapeNumRanges[c];
          rangeValues[c][i] = fixedPrec(val, reshapeOffsetFracBitDepth);
        }
      } else {
        printf("Need more code to test this branch.\n"); exit(1);
      }
    }
    // Scale and inverse scale values
    for(int c = 1; c < 3; c++) {
      scaleValues[c].resize(reshapeNumRanges[c]);
      for(int i = 0; i < reshapeNumRanges[c]; i++) {
        int val = reshapeScaleVal[c][i];
        scaleValues[c][i] = fixedPrec(val, reshapeScaleFracBitDepth);  
      }
    }

    // Calculate input and output ranges
    std::vector<fixedPrec> inputRanges[3], outputRanges[3];
    for(int c = 1; c < 3; c++) {
      Int shiftOffset, smallestVal;
      Int shiftVal = bitDepth - reshapeOffsetFracBitDepth;
      fixedPrec smallestValueInp = getSmallestVals( shiftVal, bitDepth, minSampleVal, c );
      shiftVal = bitDepth - (reshapeOffsetFracBitDepth + defaultPrecShift);
      fixedPrec smallestValueOut = getSmallestVals( shiftVal, bitDepth, minSampleVal, c );

      inputRanges[c]. resize(reshapeNumRanges[c]+1);
      outputRanges[c].resize(reshapeNumRanges[c]+1);  

      {
		    for(Int i = 0; i <= reshapeNumRanges[c]; i++) {
		      if(i == 0) {
			    Int shiftInvScale          = (1 << reshapeScaleFracBitDepth);
			    Int tempValNr = ((1 << (reshapeScaleFracBitDepth+ defaultPrecShift)) + shiftInvScale);
			    Int tempValDr = (scaleValues[c][0].val);
			    Int tempVal   = tempValNr / tempValDr;
			    fixedPrec invScaleValues = fixedPrec(tempVal, defaultPrecShift);

			    fixedPrec multValue = invScaleValues;

			    Int val = reshapeGlobalOffsetVal[c];
			    Int rangeVal = (val * deltaSampleVal[c] + (1 << (bitDepth - 1))) >> bitDepth;
          // fixedPrec tempMultValue = fixedPrec((multValue.val * deltaSampleVal[c] + (1 << (bitDepth - 1))) >> bitDepth,multValue.fracBitDepth);
          fixedPrec tempMultValue = multValue;
			    fixedPrec tempValue  = (fixedPrec(-1,0) * fixedPrec(rangeVal,reshapeOffsetFracBitDepth)) * tempMultValue + smallestValueOut;
			    outputRanges[c][0] =  tempValue;

			    inputRanges[c][0] = smallestValueInp;
		      } else {
			    Int shiftInvScale          = (1 << reshapeScaleFracBitDepth);
			    Int tempValNr = ((1 << (reshapeScaleFracBitDepth+ defaultPrecShift)) + shiftInvScale);
			    Int tempValDr = (scaleValues[c][i-1].val);
			    Int tempVal   = tempValNr / tempValDr;
			    fixedPrec invScaleValues = fixedPrec(tempVal, defaultPrecShift);

			    fixedPrec multValue = invScaleValues;

			    Int val = (rangeValues[c][i-1].val * deltaSampleVal[c] + (1 << (bitDepth - 1))) >> bitDepth;
			    fixedPrec rangeValue = fixedPrec(val, reshapeOffsetFracBitDepth);

			    inputRanges[c][i]  = inputRanges[c][i-1]  + rangeValue;
          //fixedPrec tempMultValue = fixedPrec((multValue.val * deltaSampleVal[c] + (1 << (bitDepth - 1))) >> bitDepth,multValue.fracBitDepth);
          fixedPrec tempMultValue = multValue;
			    outputRanges[c][i] = outputRanges[c][i-1] + (rangeValue * tempMultValue);
		      }
		    }
      } //    if (pInfo->reshapeChromaCrosschannelFlag == 1)
	/*    for(Int i = 0 ; i <= reshapeNumRanges[c]; i++)
	    {
        printf("%d, %d\n", inputRanges[c][i].val >> (inputRanges[c][i].fracBitDepth - 10) , outputRanges[c][i].val>> (outputRanges[c][i].fracBitDepth - 10) );
      }
      printf("\n");
	  */
    }

    std::vector<fixedPrec> offsetValue[3];
    // Calculate offset values

    for(Int c = 1; c < 3; c++) {
      Int precOffsetDeltaBits = INVERSE_FRAC_PRECISION + reshapeScaleFracBitDepth;
      offsetValue[c].resize(reshapeNumRanges[c]);
      for(Int i = 0; i < reshapeNumRanges[c]; i++) {
        offsetValue[c][i] = inputRanges[c][i] * fixedPrec(1 << precOffsetDeltaBits, precOffsetDeltaBits) - (outputRanges[c][i] * scaleValues[c][i]);
      }
    }

    // Perform the mapping
      int bitDepthDelta = defaultPrecShift + reshapeOffsetFracBitDepth - bitDepth;
      // Identify the index range and fracBitDepth
      Int idxS, fracBitDepthShiftVal;

      for(Int c = 1; c < 3; c++) {
        // if(g_val == 0 && c != 0)
        {
      Int smallestTempPelOut = 1024;
      Int largestTempPelOut  = 0;
	        for(Int i = 0 ; i < 1024; i++)
	        {
		        Int tempPel;
		        tempPel = i;
		        identifyComponentRangeIdx( tempPel, bitDepthDelta, outputRanges[c], idxS);
		        if(idxS == 0) {
		          tempPel = inputRanges[c][idxS].val;
		          fracBitDepthShiftVal = reshapeOffsetFracBitDepth;
		        } else if(idxS == reshapeNumRanges[c]+1) {
		          tempPel = inputRanges[c][idxS-1].val;
		          fracBitDepthShiftVal = reshapeOffsetFracBitDepth;
		        } else {  
		          tempPel =  (( tempPel * scaleValues[c][idxS-1].val ) << (defaultPrecShift + reshapeOffsetFracBitDepth - bitDepth)) 
			          + offsetValue[c][idxS-1].val;

		          fracBitDepthShiftVal = defaultPrecShift + reshapeScaleFracBitDepth + reshapeOffsetFracBitDepth - bitDepth;
		        }
		        tempPel = (tempPel + (1 << (fracBitDepthShiftVal - 1))) >> fracBitDepthShiftVal; 
		        Int tempPelOut;
            tempPelOut = tempPel;
		        tempPelOut = ( tempPelOut < minSampleVal[c] ) ? minSampleVal[c] : ((tempPelOut > maxSampleVal[c]) ? maxSampleVal[c] : tempPelOut);
            if(tempPelOut > largestTempPelOut) { largestTempPelOut = tempPelOut; }
            if(tempPelOut < smallestTempPelOut) { smallestTempPelOut = tempPelOut; }
            m_forwardChromaReshapingLUT[c-1][tempPelOut] = i; 
            //printf("%d, %d\n", i, tempPelOut);
	        }
         // printf("\n");
          for(int i = 0; i < smallestTempPelOut; i++)
            m_forwardChromaReshapingLUT[c-1][i] = minSampleVal[c]; 
          for(int i = largestTempPelOut; i < 1024; i++)
            m_forwardChromaReshapingLUT[c-1][i] = maxSampleVal[c]; 
        }
      } //    if (pInfo->reshapeChromaCrosschannelFlag == 1)
    }
#endif



#if TCH_CROSSPLANE_CRCB
    if (reshaping.m_reshapeMode == 1) {
      float ba = (float) m_temporalBa;
       
      //Compute LOG curve
      LogGParam prm_default={ba,2.8f,1.0f};  //automatic  settings Ba,gamma,b0
      //compute Y LUT

      const std::vector<float> loggLutY=Tch::logG_PQi(prm_default, (float) m_peakLum); // now we can encode forward (and decode inverse) pq_logGinv();

      //step 2: generate target inverse luma reshaping LUT
      int     bitDepthL = (int)(log((double)MAX_LUMA_RESHAPING_LUT_SIZE)/log(2.)+.5);
      float   offSR = 16.0 * (1<<(bitDepthL-8));
      float   scaleSR = 219. * (1<<(bitDepthL-8));

      for(idx=0; idx < MAX_LUMA_RESHAPING_LUT_SIZE; idx++)
      {
        int   idxFR = std::max(0,std::min((int)meta_max,(int)( meta_max * ((float)idx-offSR) / scaleSR )));

        if (inp->m_sampleRange == SR_STANDARD)
          out_idx = (uint16)fClip(fRound((float)(offSR + loggLutY[idxFR]*scaleSR)), 0.0f, (float)meta_max);
        else
          out_idx =  (uint16)fClip(fRound((float)(  loggLutY[idx]*meta_max)), 0.0f, (float)meta_max);

        logY_lut[idx] = (double)out_idx;
        index[idx] = idx;
      }
      BuildInverseYlogLUT(logY_lut, target_lut,(1 << inp->m_bitDepthComp[0]));
    }
    else {
#endif
#if LUMA_RESHAPE_METHOD
      lumaReshaping(inp, target_lut, 1.0/power_alpha, org_I_CL, org_I_CH, curr_frame_stats, inp->m_frameNo-m_sceneDataOld.prev_frameNo, &opt);
      m_sceneDataOld.prev_frameNo = inp->m_frameNo;
#endif
      //step 2: generate target inverse luma reshaping LUT
      for(idx=0; idx < MAX_LUMA_RESHAPING_LUT_SIZE; idx++)
      {
#if LUMA_RESHAPE_METHOD
        norm_val = target_lut[idx];
        if (inp->m_sampleRange == SR_STANDARD)
        {
          //out_idx =  (uint16) fClip(fRound((876.0f * norm_val + 64.0f)), 0.0f, (float) meta_max);  // float(0-1) to fixed (std range)
          printf("not supported!\n");
          exit(-1);
        }
        else
        {
          out_idx =  (uint16) fClip(fRound((meta_max * norm_val)), 0.0f, (float) meta_max);  // float(0-1) to (full range)
        }
        target_lut[idx] = (double)out_idx;
        index[idx] = idx;
#else
		  if (inp->m_sampleRange == SR_STANDARD)
		  {
				dTemp = fClip((float)((((double)idx/4.0) - 16.0)/219.0), 0.0f, 1.0f);  // fixed (std range) to float (0-1) full range normalized value
				norm_val = fClip((float)(pow((dTemp - I_CL)/(I_CH - I_CL), power_alpha) * (org_I_CH - org_I_CL) + org_I_CL),0.0f, 1.0f);
				out_idx =  (uint16) fClip(fRound((876.0f * norm_val + 64.0f)), 0.0f, (float) meta_max);  // float(0-1) to fixed (std range)
				target_lut[idx] = (double)out_idx;
				index[idx] = idx;
		  }
		  else if (inp->m_sampleRange == SR_FULL)
		  {
				dTemp = (double)idx/(double)meta_max;
				out_idx =  (uint16)fClip(fRound((float)((pow((dTemp - I_CL)/(I_CH - I_CL), power_alpha) * (org_I_CH - org_I_CL) + org_I_CL)*meta_max)), 0.0f, (float)meta_max);
				target_lut[idx] = (double)out_idx;
				index[idx] = idx;
		  }
#endif
      }

	  
#if TCH_CROSSPLANE_CRCB
    }
#endif

    //step 3: polynomial approximation
    {
      int input_bit_depth = inp->m_bitDepthComp[0];

      int			lut_len				= 1 << input_bit_depth;
      bool		smoothness_flag;

      smoothness_flag		= true;
      int nFractionBits;
      nFractionBits = 2;
      UnifiedLumaReshapeInfo UnifiedLumaReshape;

      AutomaticMinimalMSE_Fitting(lut_len, target_lut, input_bit_depth, nFractionBits, smoothness_flag, &UnifiedLumaReshape);

#if NO_INTRA_PERIOD_UPDATE
	  if (inp->m_frameNo == 0) // store metadata for 0th frame
	  {
		  storePrevLumaMetaData(&(m_sceneData.PrevLumaReshape), &UnifiedLumaReshape);
	  }
	  else //compare newly derived luma metadata with previous one
	  {
		   compareLumaMetaData(&(m_sceneData.PrevLumaReshape), &UnifiedLumaReshape, &isParameterUpdate);
	  }
	  //if the luma metadata is exactly same as previous, then no need to signal it
	  if (isParameterUpdate)
	  {
		  if (inp->m_frameNo != 0)  //first frame already stored
			storePrevLumaMetaData(&(m_sceneData.PrevLumaReshape), &UnifiedLumaReshape);
#endif	 
#if RESHAPER_JOINT_OPT
          UnifiedLumaReshape.opt = opt;
          for (int k=0; k<15; k++)
          {
            UnifiedLumaReshape.lumaDQP[k] = m_lumaDQP[k];
          }
#endif
	  WriteMetaData(reshaping.atfParamsLumaFilename,iPos,&UnifiedLumaReshape);
      if (reshaping.m_reshapeMode == 0) 
	  {
		writeCbCrATF(reshaping.atfParamCbCrFilename, iPos);
      }

      BuildInverseReshapingLUT(UnifiedLumaReshape.nAllSegmentLengths,UnifiedLumaReshape.allSegmentLengths,UnifiedLumaReshape.fractionBits,UnifiedLumaReshape.polyCoeff,lut_len,m_inverseLumaReshapingLUT);

      //step 4: use inverse LUT to approximate forward LUT
      BuildForwardReshapingLUT(m_inverseLumaReshapingLUT, m_forwardLumaReshapingLUT,lut_len);
  
#if NO_INTRA_PERIOD_UPDATE
	 } // end of if (isParameterUpdate)
#endif
    }
    m_sceneNum++;
#if MODE0_UPDATE
	m_sceneData.prev_frameNo = inp->m_frameNo;
	m_sceneData.sceneIsFades = 0;
#endif
  }
}

/**
- Function to perform reshaping in fix point integer domain based on table lookup 
\param   inp            pointer to image frame buffer to be processed
\param   reshaping      struct containing unified reshaping model 
\retval  out            pointer to image frame buffer processed
*/
void Reshaper::process ( Frame* out, const Frame *inp, ReshapingInfo reshaping)
{
  int temp_idx;
  int *LumaLUT, *ChromaLUT[2];
  
    LumaLUT = m_forwardLumaReshapingLUT;
    ChromaLUT[0] = m_forwardChromaReshapingLUT[0];
    ChromaLUT[1] = m_forwardChromaReshapingLUT[1];

  // forward luma reshaping: forward LUT table look up
  for (int i = 0; i < inp->m_compSize[0]; i++) {
    temp_idx = inp->m_ui16Comp[0][i]; 
    out->m_ui16Comp[0][i] =  LumaLUT[temp_idx];
  }


  if (reshaping.m_reshapeMode == 0) {      
    // forward chroma reshaping 
    for (int i = 0; i < inp->m_compSize[1]; i++) 
    {
      temp_idx = inp->m_ui16Comp[1][i]; 
      out->m_ui16Comp[1][i] =  ChromaLUT[0][temp_idx];

      temp_idx = inp->m_ui16Comp[2][i]; 
      out->m_ui16Comp[2][i] =  ChromaLUT[1][temp_idx];
    }
  }
}

#if TCH_CROSSPLANE_CRCB
double Reshaper::TchComputeMeanY(const uint16 * yin, int width, int height, int size, int frameNumber, SampleRange range) 
{
  double            mean = 0.;
  double            meanWithoutBlack = 0.;
  int               cptwithoutBlack = 0;
  const double      blackLevel = 0.1;
  const int         histogram_length = 1000;
  std::vector<int>  histogram(histogram_length,0);
  double            Ymax = 0.;

  int     bitDepthL = (int)(log((double)MAX_LUMA_RESHAPING_LUT_SIZE)/log(2.)+.5);
  float   offSR = 16.0 * (1<<(bitDepthL-8));
  float   scaleSR = 219. * (1<<(bitDepthL-8));

  if (m_peakLum == 0.)
  {
    for (int i = 0; i < size; i++)
    {
      double  YfullRange = (range == SR_FULL) ? (yin[i]) : ( 1023. * (yin[i]-offSR) / scaleSR );
      double Y = Tch::PQi( (float) (YfullRange) /1023.)*10000;
      if (Y > Ymax)
        Ymax = Y;
    }
    if (Ymax > 4000.)
      m_peakLum = 5000.;
    else
      m_peakLum = 4000.;
  }


  for (int i = 0; i < size; i++)
  {
    double  YfullRange = (range == SR_FULL) ? (yin[i]) : ( 1023. * (yin[i]-offSR) / scaleSR );
    double  Y = Tch::PQi( (float) (YfullRange) /1023.)*10000;
    mean += Y;
    histogram[ (int) ( std::min<double>( 1., Y/m_peakLum ) * (histogram_length-1) ) ]++;

    if (Y > Ymax) 
      Ymax = Y;
    if (Y > blackLevel)
    {
      meanWithoutBlack += Y;
      cptwithoutBlack++;
    }
  }
  mean /= size;
  if (cptwithoutBlack)
    meanWithoutBlack /= cptwithoutBlack;

  const double      ratio = .75;
  int               nbPix = 0;
  int               hist_idx = histogram_length-1;
  double            meanHistBright = 0.;
  while ( nbPix < ratio * size )
  {
    meanHistBright += (hist_idx+.5) * histogram[hist_idx];
    nbPix += histogram[hist_idx];
    hist_idx--;
  }
  meanHistBright = (meanHistBright / nbPix) * m_peakLum/histogram_length;
  return meanHistBright;
}

float  Reshaper::computeGlogParameters(int frameNumber, const Frame &inp) {

  int width=inp.m_width[0];
  int height=inp.m_height[0];
  int size=inp.m_compSize[0];
  const uint16 * const Ycomp=inp.m_ui16Comp[0];  

  double meanY = TchComputeMeanY(Ycomp, width, height, size, frameNumber, inp.m_sampleRange);
  double Ba = -.0003*meanY*meanY + .558*meanY + 4.6767;

  if(frameNumber == 0)
  {
    m_temporalBa = Ba;
    m_recomputeReshapeParams = 1;
    m_last_written_Ba = m_temporalBa;
  }
  else
  {
    double  ratio = Ba/m_temporalBa;
    if (ratio > 0.6666 && ratio < 1.5) // stabilize Ba
    {
      m_temporalBa = 0.025*Ba + 0.975*m_temporalBa;
    }
    else if (ratio > 3 || ratio < 0.3333) // scene change
    {
      m_temporalBa = Ba;
    }
    else
    {
      m_temporalBa = 0.1*Ba + 0.9*m_temporalBa;
    }

    if ((m_temporalBa < (1-TCH_BA_WRITE_THRESHOLD)*m_last_written_Ba) || (m_temporalBa > (1+TCH_BA_WRITE_THRESHOLD)*m_last_written_Ba))
    {
      m_recomputeReshapeParams = 1;
      m_last_written_Ba = m_temporalBa;
    }
    else
      m_recomputeReshapeParams = 0;
  }

  return   (float) m_temporalBa;
}
#endif

#if ARRS_MODEL_OPTIMIZATION

void  Reshaper::process ( Frame* out, Frame *inp, ReshapingInfo reshaping, int POC)
{
  int i,j, lumaOffset, lumaPosition,chromaPosition;
  float *floatData[4];
  float InputMinMax[6], OutputMinMax[6];
#if ARRS_MATCH_INVERSE
  int inLuma, outLuma;
#else
  double inLuma, outLuma, scaleFactor;
#endif
  // First: Setup
  m_reshaperParams.newScene = false;
  floatData[0]= new float[inp->m_compSize[0]];
  floatData[1]= new float[inp->m_compSize[1]];
  floatData[2]= new float[inp->m_compSize[2]];
  floatData[3]= new float[inp->m_compSize[0]];

  InputMinMax[0]= 0.0f;   InputMinMax[1]= 1.0f;
  InputMinMax[2]=-0.5f;  InputMinMax[3]= 0.5f;  
  InputMinMax[4]=-0.5f;  InputMinMax[5]= 0.5f;

  OutputMinMax[0]= 0.0f;  OutputMinMax[1]= 1.0f;
  OutputMinMax[2]=-0.5f;  OutputMinMax[3]= 0.5f;
  OutputMinMax[4]=-0.5f;  OutputMinMax[5]= 0.5f;
  OutputMinMax[2]= ((-0.5f+reshaping.m_chromaHistOffsetCb)/reshaping.m_chromaHistScaleCb);  OutputMinMax[3]= (1.0f/reshaping.m_chromaHistScaleCb) + OutputMinMax[2];
  OutputMinMax[4]= ((-0.5f+reshaping.m_chromaHistOffsetCr)/reshaping.m_chromaHistScaleCr);  OutputMinMax[5]= (1.0f/reshaping.m_chromaHistScaleCr) + OutputMinMax[4];

  // Step1: convert the input to normalized float:
  for(int comp=0; comp<3; ++comp)
  {
      fixedToFloat( (void *)&inp->m_ui16Comp[comp][0], floatData[comp], inp->m_bitDepthComp, inp->m_sampleRange, InputMinMax, inp->m_compSize, comp );
  }

  // Step2: find optimal forward reshaping parameters
  getReshaperParameters(POC, floatData[0], inp->m_width[0], inp->m_height[0], &reshaping);
  

  // Step3: Calcuate optimal inv-reshaper Model-parameters to be transmitted
  // create inverse LUT
  if(m_reshaperParams.newScene)
    generateLUTs(POC, InputMinMax, OutputMinMax, inp->m_bitDepthComp, out->m_bitDepthComp, inp->m_sampleRange, out->m_sampleRange, reshaping );
  
  // Step4: apply forward reshaper
  
#if ARRS_MATCH_INVERSE
  // Luma reshaping
  for(i=0;i<inp->m_height[0]*inp->m_width[0];i++)
    out->m_ui16Comp[0][i]=m_forwardLumaReshapingLUT[inp->m_ui16Comp[0][i]];
#else
  // Luma reshaping
  for(i=0;i<inp->m_height[0]*inp->m_width[0];i++)
    floatData[3][i] = (float)pow((double)floatData[0][i], m_reshaperParams.reshaperAlpha);  
#endif
  //Chroma Reshaping
  // picture is assumed to be 420

  chromaPosition=0;
  for (i = 0; i < inp->m_height[1]; i++) {
    lumaOffset=i*inp->m_width[0]*2;
    for(j = 0; j < inp->m_width[1]; j++) {
      lumaPosition=lumaOffset + j*2;
      //Collocated Luma position
#if COLOCATED_LUMA_MODE==0
#if ARRS_MATCH_INVERSE
      outLuma = out->m_ui16Comp[0][lumaPosition];
      inLuma = m_inverseLumaReshapingLUT[outLuma];
#else
      inLuma = floatData[0][lumaPosition];
      outLuma = floatData[3][lumaPosition];
#endif
#else
#if ARRS_MATCH_INVERSE
      outLuma=(out->m_ui16Comp[0][lumaPosition] + out->m_ui16Comp[0][lumaPosition+1] + out->m_ui16Comp[0][lumaPosition+inp->m_width[0]] + out->m_ui16Comp[0][lumaPosition+inp->m_width[0]+1] + 2) >> 2;
      inLuma = m_inverseLumaReshapingLUT[outLuma];
#else
      outLuma=(floatData[3][lumaPosition] + floatData[3][lumaPosition+1] + floatData[3][lumaPosition+inp->m_width[0]] + floatData[3][lumaPosition+inp->m_width[0]+1])/4.0;
      inLuma = (float)pow((double)outLuma, 1.0/m_reshaperParams.reshaperAlpha);  
#endif
#endif
      // chroma scale factor
#if ARRS_MATCH_INVERSE
      out->m_ui16Comp[1][chromaPosition] = (int)floor((float)((m_forwardChromaReshapingLUT[0][inLuma]*(inp->m_ui16Comp[1][chromaPosition]-m_offsetArray[1]))/CHROMA_INV_LUt_SCALE) + 0.5) + m_offsetArray[0];
      out->m_ui16Comp[2][chromaPosition] = (int)floor((float)((m_forwardChromaReshapingLUT[1][inLuma]*(inp->m_ui16Comp[2][chromaPosition]-m_offsetArray[3]))/CHROMA_INV_LUt_SCALE) + 0.5) + m_offsetArray[2];
#else
      scaleFactor = (double)(outLuma + reshaping.blackLevelLiftTarget)/(double)(inLuma + reshaping.blackLevelLiftInput);
      scaleFactor = dClip(scaleFactor, (double)MIN_CROSS_LAYER_RESHAPE_SCALE, (double)MAX_CROSS_LAYER_RESHAPE_SCALE);
      // reshape Cb and Cr
      floatData[1][chromaPosition]=(float)(floatData[1][chromaPosition]*scaleFactor);
      floatData[2][chromaPosition]=(float)(floatData[2][chromaPosition]*scaleFactor);
#endif
      ++chromaPosition;
    }
  }

  
#if !ARRS_MATCH_INVERSE
  // Step4: convert the reshaped signal to fixed point
  floatToFixed( floatData[3], (void *)&out->m_ui16Comp[0][0], out->m_bitDepthComp, out->m_sampleRange, OutputMinMax, out->m_compSize, 0 );
  for(int comp=1; comp<3; ++comp)
  {
      floatToFixed( floatData[comp], (void *)&out->m_ui16Comp[comp][0], out->m_bitDepthComp, out->m_sampleRange, OutputMinMax, out->m_compSize, comp );
  }
#endif
  // Last Cleanup:
  for(i=0;i<4;i++)
  {
    delete[] floatData[i];
    floatData[i]=NULL;
  }
}

void Reshaper::getReshaperParameters(int frameNumber, float *Ycomp, int width, int height, ReshapingInfo *reshaping )
{
  double min, max;  
  double picLinearAverageLuma=0;
  double masking, temp;
  int blk_size   = STAT_BLOCK_SIZE;
  int blk_width  = (int)floor((double)width / blk_size);  // ignore left sub-block pixels
  int blk_height = (int)floor((double)height / blk_size); // ignore bottom sub-block pixels
  int pos;

  SceneBlkData sceneBlock;
  sceneBlock.blockSize  = blk_size;
  sceneBlock.blockWidth  = blk_width;
  sceneBlock.blockHeight = blk_height;
  sceneBlock.blockMinIntensity = new double[blk_width*blk_height];
  sceneBlock.blockMaxIntensity = new double[blk_width*blk_height];
  sceneBlock.blockAvgIntensity = new double[blk_width*blk_height];
    

  // determine the parameter for forward reshaping
  blockproc(Ycomp, width, height, sceneBlock, (double)(1.0/reshaping->alphaInput)); // KM: should use temporal gamma?
  min=1e30;
  max=0;
  for (pos=0; pos<blk_width*blk_height; pos++)
  {
    picLinearAverageLuma+=sceneBlock.blockAvgIntensity[pos];
    masking=MASKING_FACTOR*sceneBlock.blockAvgIntensity[pos];
    if ((sceneBlock.blockMinIntensity[pos]+masking) < min)
      min=sceneBlock.blockMinIntensity[pos]+masking;
    if (sceneBlock.blockMaxIntensity[pos] > max)
      max=sceneBlock.blockMaxIntensity[pos];
  }

  picLinearAverageLuma/=(double)(blk_width*blk_height);

  if(frameNumber==m_startFrame)
  {
    m_reshaperParams.temporalBlackLevel = (float)min;
    m_reshaperParams.temporalPeakBrightness = (float)max;
    m_reshaperParams.temporalAvgBrightness = (float)picLinearAverageLuma;
  }
  else
  {
  m_reshaperParams.temporalBlackLevel=(1-ADAPTATION_RATE)*m_reshaperParams.temporalBlackLevel+(float)(ADAPTATION_RATE*min);
  m_reshaperParams.temporalPeakBrightness=(1-ADAPTATION_RATE)*m_reshaperParams.temporalPeakBrightness+(float)(ADAPTATION_RATE*max);
  m_reshaperParams.temporalAvgBrightness=(1-ADAPTATION_RATE)*m_reshaperParams.temporalAvgBrightness+(float)(ADAPTATION_RATE*picLinearAverageLuma);
  }
  
  if(m_reshaperParams.temporalBlackLevel < (float)BLACK_ADAPTATION_FACTOR*m_reshaperParams.temporalAvgBrightness)
    temp = (double)GAMMA_LIFT + 0.2*(log10((double)m_reshaperParams.temporalPeakBrightness/((double)BLACK_ADAPTATION_FACTOR*m_reshaperParams.temporalAvgBrightness)));
  else
    temp = (double)GAMMA_LIFT + 0.2*(log10((double)m_reshaperParams.temporalPeakBrightness/(double)m_reshaperParams.temporalBlackLevel));
  temp=dClip(temp, GAMMA_CORRECTION_MIN, GAMMA_CORRECTION_MAX);
  if (frameNumber!=m_startFrame)
  temp=(double)(1-ADAPTATION_RATE)*m_reshaperParams.temporalGamma+(double)ADAPTATION_RATE*temp;

  if((abs(1-temp/m_reshaperParams.temporalGamma) > SCENE_CHANGE_THR) || (frameNumber==m_startFrame) )
  {
    m_reshaperParams.newScene=true;
    m_reshaperParams.temporalGamma = (float)temp;
    m_reshaperParams.reshaperAlpha=(double)1.0/(double)(reshaping->alphaInput*m_reshaperParams.temporalGamma*reshaping->gammaTarget);
  }
}

void Reshaper::blockproc(float *input, int inp_wid, int inp_hei, SceneBlkData sceneBLK, double TFGamma) 
{
  int size = sceneBLK.blockSize;
  int wid = sceneBLK.blockWidth;
  int hei = sceneBLK.blockHeight;
  int block_size=size*size;
  double *block = new double[block_size];
  int i, j, i_pos, j_pos;
  double mean, min, max;


  ///////////// anaylze scene block data //////////////
  for (i_pos=0; i_pos<hei; i_pos++)
  {
    for (j_pos=0; j_pos<wid; j_pos++)
    {
      // read one nxn block
      for (i=0; i<size; i++)
      {
        for (j=0; j<size; j++)
        {
          block[i*size+j] = (double)input[(i_pos*size+i) * inp_wid + (j_pos*size+j)];
        }
      }

      //

      // calculating block statitics
      min = (block[0]<=0) ? 0.0:pow(block[0], TFGamma);
      max = (block[0]<=0) ? 0.0:pow(block[0], TFGamma);
      mean= (block[0]<=0) ? 0.0:pow(block[0], TFGamma);

      for (i=1; i<block_size; i++)
      {

        block[i]= (block[i]<=0) ? 0.0:pow(block[i], TFGamma);
        if (block[i]<min)
          min = block[i];
        if (block[i]>max)
          max = block[i];
        mean += block[i];
      }
           
      mean /= (double)(block_size);

      sceneBLK.blockAvgIntensity[i_pos*wid+j_pos] = mean;
      sceneBLK.blockMinIntensity[i_pos*wid+j_pos] = min;
      sceneBLK.blockMaxIntensity[i_pos*wid+j_pos] = max;
    }
  }
  delete[] block;
}


void Reshaper::generateLUTs (int frameNumber, float *preScalingMinMax, float *postScalingMinMax, int *prebitDepth, int *postbitDepth, SampleRange preSampleRange, SampleRange postSampleRange, ReshapingInfo reshaping )
{
  float floatVal[4][MAX_LUMA_RESHAPING_LUT_SIZE];
  double tempdLut[MAX_LUMA_RESHAPING_LUT_SIZE];
  double gamma, inBLL, outBLL;
  uint16 temp16u[2];
  int numParams;
  float chromaScale[2];
  int sizeTemp[3]={1, 1, 1};
  int lutSize[3]={MAX_LUMA_RESHAPING_LUT_SIZE, MAX_LUMA_RESHAPING_LUT_SIZE, MAX_LUMA_RESHAPING_LUT_SIZE};
  bool smoothness = true;
  int nFractionBits = 2;
  UnifiedLumaReshapeInfo UnifiedLumaReshape;
#if TCH_CROSSPLANE_CRCB
  Tch::CrossPlaneReshaper *chromaReshaper = new Tch::CrossPlaneReshaper();
#endif
  
  if(postbitDepth[0]>8)
  {
    uint16 fixed16Val[MAX_LUMA_RESHAPING_LUT_SIZE];
  for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      fixed16Val[i]=(uint16)i;
    fixedToFloat( &fixed16Val[0], &floatVal[0][0], &postbitDepth[0], postSampleRange, postScalingMinMax, &lutSize[0], 0 );
  }
  else
  {
    uint16 fixed8Val[MAX_LUMA_RESHAPING_LUT_SIZE];
    for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      fixed8Val[i]=(uint8)i;
    fixedToFloat( &fixed8Val[0], &floatVal[0][0], &postbitDepth[0], postSampleRange, postScalingMinMax, &lutSize[0], 0 );
  }

  // populate forward Luma reshaping LUT
  gamma = (double)(1.0/m_reshaperParams.reshaperAlpha);
  inBLL = (double)reshaping.blackLevelLiftInput;
  outBLL = (double)reshaping.blackLevelLiftTarget;

  for (int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
    floatVal[1][i] = (float)pow((double)floatVal[0][i], gamma);  

  if(postbitDepth[0]>8)
  {
    uint16 fixed16Val[MAX_LUMA_RESHAPING_LUT_SIZE];
    floatToFixed(&floatVal[1][0], (void *)&fixed16Val[0], &postbitDepth[0], postSampleRange, preScalingMinMax, &lutSize[0], 0 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
    for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      tempdLut[i]=(double)fixed16Val[i];
  }
  else
  {
    uint8 fixed8Val[MAX_LUMA_RESHAPING_LUT_SIZE];
    floatToFixed(&floatVal[1][0], (void *)&fixed8Val[0], &postbitDepth[0], postSampleRange, preScalingMinMax, &lutSize[0], 0 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
    for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      tempdLut[i]=(double)fixed8Val[i];
  }
  

  AutomaticMinimalMSE_Fitting(lutSize[0], tempdLut, postbitDepth[0], nFractionBits, smoothness, &UnifiedLumaReshape);
  WriteMetaData(reshaping.atfParamsLumaFilename,frameNumber,&UnifiedLumaReshape);


#if ARRS_MATCH_INVERSE
  BuildInverseReshapingLUT(UnifiedLumaReshape.nAllSegmentLengths,UnifiedLumaReshape.allSegmentLengths,UnifiedLumaReshape.fractionBits,UnifiedLumaReshape.polyCoeff,lutSize[0],m_inverseLumaReshapingLUT);
  assert(MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE==MAX_LUMA_RESHAPING_LUT_SIZE);// KM: Fix the BuildForwardReshapingLUT function to operate on different LUT_size for inverse and forward
  BuildForwardReshapingLUT(m_inverseLumaReshapingLUT, m_forwardLumaReshapingLUT,MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE); 


  if(postbitDepth[0]>8)
  {
    uint16 fixed16Val[MAX_LUMA_RESHAPING_LUT_SIZE];
    for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      fixed16Val[i]=(uint16)m_inverseLumaReshapingLUT[i];
    fixedToFloat( &fixed16Val[0], &floatVal[1][0], &postbitDepth[0], postSampleRange, preScalingMinMax, &lutSize[0], 0 );// KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
  }
  else
  {
    uint8 fixed8Val[MAX_LUMA_RESHAPING_LUT_SIZE];
    for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
      fixed8Val[i]=(uint8)m_inverseLumaReshapingLUT[i];
    fixedToFloat( &fixed8Val[0], &floatVal[1][0], &postbitDepth[0], postSampleRange, preScalingMinMax, &lutSize[0], 0 );// KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 

  }
#endif

  // populate forward Chroma reshaping LUT
  for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
  {
    floatVal[2][i] = (float)((floatVal[1][i]+inBLL)/(floatVal[0][i]+outBLL));
    floatVal[2][i] = fClip(floatVal[2][i], 1.0f/MAX_CROSS_LAYER_RESHAPE_SCALE, 1.0f/MIN_CROSS_LAYER_RESHAPE_SCALE);
  }

  // Cb

  floatVal[1][0]=postScalingMinMax[2];
  floatVal[1][1]=postScalingMinMax[3];
  sizeTemp[1]=2;
  temp16u[0]=0;
  temp16u[1]=0;
  floatToFixed(&floatVal[1][0], (void *)&temp16u, &postbitDepth[0], postSampleRange, postScalingMinMax, sizeTemp, 1 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
  chromaScale[0]=(float)(temp16u[1]-temp16u[0])/(floatVal[1][1]-floatVal[1][0]);
  m_offsetArray[0]=temp16u[1]-(int)floor(chromaScale[0]*floatVal[1][1]+0.5);


  floatVal[1][0]=preScalingMinMax[2];
  floatVal[1][1]=preScalingMinMax[3];
  sizeTemp[1]=2;
  temp16u[0]=0;
  temp16u[1]=0;
  floatToFixed(&floatVal[1][0], (void *)&temp16u, &postbitDepth[0], postSampleRange, preScalingMinMax, sizeTemp, 1 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
  chromaScale[1]=(float)(temp16u[1]-temp16u[0])/(floatVal[1][1]-floatVal[1][0]);
  m_offsetArray[1]=temp16u[1]-(int)floor(chromaScale[1]*floatVal[1][1]+0.5);

  chromaScale[0]=chromaScale[1]/chromaScale[0];
  for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
    floatVal[3][i]=floatVal[2][i]*chromaScale[0];
  //encode color correction LUTs with PWL model
  chromaReshaper->setCrossCompLUT(floatVal[3], MAX_LUMA_RESHAPING_LUT_SIZE, 0);
  

  // Cr:

  floatVal[1][0]=postScalingMinMax[4];
  floatVal[1][1]=postScalingMinMax[5];
  sizeTemp[2]=2;
  temp16u[0]=0;
  temp16u[1]=0;
  floatToFixed(&floatVal[1][0], (void *)&temp16u, &postbitDepth[0], postSampleRange, postScalingMinMax, sizeTemp, 2 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
  chromaScale[0]=(float)(temp16u[1]-temp16u[0])/(floatVal[1][1]-floatVal[1][0]);
  m_offsetArray[2]=temp16u[1]-(int)floor(chromaScale[0]*floatVal[1][1]+0.5);


  floatVal[1][0]=preScalingMinMax[4];
  floatVal[1][1]=preScalingMinMax[5];
  sizeTemp[2]=2;
  temp16u[0]=0;
  temp16u[1]=0;
  floatToFixed(&floatVal[1][0], (void *)&temp16u, &postbitDepth[0], postSampleRange, preScalingMinMax, sizeTemp, 2 ); // KM: The bit-depth and sample range are assumed to be the same for inverse reshaper. This can change. 
  chromaScale[1]=(float)(temp16u[1]-temp16u[0])/(floatVal[1][1]-floatVal[1][0]);
  m_offsetArray[3]=temp16u[1]-(int)floor(chromaScale[1]*floatVal[1][1]+0.5);
  
  chromaScale[0]=chromaScale[1]/chromaScale[0];
  for(int i=0; i<MAX_LUMA_RESHAPING_LUT_SIZE; i++)
    floatVal[3][i]=floatVal[2][i]*chromaScale[0];
  //encode color correction LUTs with PWL model
  chromaReshaper->setCrossCompLUT(floatVal[3], MAX_LUMA_RESHAPING_LUT_SIZE, 1);
  


  // compute parameters
  chromaReshaper->encode_beta0_lut();
  numParams = 2;
  if( (postScalingMinMax[2]==postScalingMinMax[4]) && (postScalingMinMax[3]==postScalingMinMax[5]) && (preScalingMinMax[2]==preScalingMinMax[4]) && (preScalingMinMax[3]==preScalingMinMax[5]))
    numParams = 1;
  // Write the Cr param here (if needed)
  //write ATF data for chroma
#if TCH_CROSSPLANE_CRCB
  chromaReshaper->writeCbCrATFCrossPlane(reshaping.atfParamCbCrFilename, frameNumber-m_startFrame, getRangeId(), 2, m_offsetArray, numParams);
#if ARRS_MATCH_INVERSE
  chromaReshaper->getForwardCrossScaleLUT(m_forwardChromaReshapingLUT[0], MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE, 0);
  chromaReshaper->getForwardCrossScaleLUT(m_forwardChromaReshapingLUT[1], MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE, 1);
#endif
  // free objects
  if (chromaReshaper != NULL) {
    delete chromaReshaper;
    chromaReshaper = NULL;
  }
#endif
  return;
}

void Reshaper::floatToFixed( float *input, void *output, int *bitDepth, SampleRange sampleRange, float *minMax, int *size, int comp )
{
  double cmp_range, cmp_offset, cmp_scale;
  float min=minMax[2*comp];
  float max=minMax[2*comp+1];
  float temp;
  switch (sampleRange) {
  case SR_FULL:
    cmp_range = (double)(max-min);
    cmp_scale = (double)((1 << (bitDepth[comp]))-1) / cmp_range;
    cmp_offset = -(double)min*(double)(1 << (bitDepth[comp])) / cmp_range;
    break;
  case SR_STANDARD:
    cmp_range = (double)(max-min);
    if (comp==Y_COMP)
    {
      cmp_scale = (double)((1 << (bitDepth[comp]-8)) * 219.0) / cmp_range;
    }
    else
    {
      cmp_scale = (double)((1 << (bitDepth[comp]-8)) * 224.0) / cmp_range;
    }
    cmp_offset = ((double)(1 << (bitDepth[comp]-8)) * 16.0) - ((double)min * cmp_scale);

    break;
  default:
    fprintf(stderr, "Unsupported sampel range %d\n", sampleRange);
    exit(EXIT_FAILURE);
    break;
  }

  if(bitDepth[comp] > 8)
  {
    uint16 *outBuff = (uint16 *)output;
    for(int i=0; i<size[comp]; i++)
    {
      temp = fClip(input[i], min, max);
      outBuff[i] = (uint16)floor(cmp_scale * (double)temp  + cmp_offset + 0.5);
    }
  }
  else
  {
    uint8 *outBuff = (uint8 *)output;
    for(int i=0; i<size[comp]; i++)
    {
      temp = fClip(input[i], min, max);
      outBuff[i] = (uint8)floor(cmp_scale * (double)temp  + cmp_offset + 0.5);
    }
  }
  return;
}



void Reshaper::fixedToFloat( void *input, float *output, int *bitDepth, SampleRange sampleRange, float *minMax, int *size, int comp )
{
  double cmp_range, cmp_offset, cmp_scale;
  float min=minMax[2*comp];
  float max=minMax[2*comp+1];
  float  temp;

  switch (sampleRange) {
  case SR_FULL:       
    cmp_range = (double)(max-min);
    cmp_scale = cmp_range/(double)((1 << bitDepth[comp])-1);
    cmp_offset = (-min)*(double)(1 << (bitDepth[comp]))/(double)(((1 << bitDepth[comp])-1));
    break;
  case SR_STANDARD:
    cmp_range = (double)(max-min);
    
    if(comp == Y_COMP)
      cmp_scale = cmp_range / (double)((1 << (bitDepth[comp]-8)) * 219.0);
    else
      cmp_scale = cmp_range / (double)((1 << (bitDepth[comp]-8)) * 224.0);
    cmp_offset = ((double)(1 << (bitDepth[comp]-8)) * 16.0 * cmp_scale) - ((double)min);  
    break;
  default:
    fprintf(stderr, "Unsupported sampel range %d\n", sampleRange);
    exit(EXIT_FAILURE);
    break;
  }

  if(bitDepth[comp] > 8)
  {
    uint16 *inBuff = (uint16 *)input;
    for (int i=0; i<size[comp]; i++)
    {
      temp = (float)((cmp_scale * (double)inBuff[i]) - cmp_offset);
      output[i] = fClip(temp, min, max);
    }
  }
  else
  {
    uint8 *inBuff = (uint8 *)input;
    for (int i=0; i<size[comp]; i++)
    {
      temp = (float)((cmp_scale * (double)inBuff[i]) - cmp_offset);
      output[i] = fClip(temp, min, max);
    }
  }
  return;
}
#endif
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
#endif