/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComReshape.cpp
    \brief    common reshaper class
*/
#include "TComReshape.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#if HDR_ETM_RESHAPER
//! \ingroup TLibCommon
//! \{
/**
 - Determine number and length of all segments from signalled information, presence of last flat segment determined if sum of lengths less than full model domain i.e. 1024
 \param   nSegmentSignaled      number of segments explicitly signaled
 \param   SegmentLengths        signalled segment lengths
 \param   global_offset_val     length of initial flat segment
 \param   total_length          total length of model i.e. 1<< bit_depth
 \retval  pNumberAllSegments    number of all segments
 \retval  pAllSegmentLengths    length of all segments not just those signalled
 \retval  pFirstSegmentFlat     flag indicating first segment is flat
 \retval  pLastSegmentFlat      flag indicating last segment is flat
 */
Void TComReshape::DeriveAllSegments(Int nSegmentSignaled,Int SegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS],Int global_offset_val,Int total_length,Int *pNumberAllSegments,Int pAllSegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS],Bool *pFirstSegmentFlat,Bool *pLastSegmentFlat)
{
  Int nAllSegments;
  Int sum;
  Int RemainingLength;
  Int first;
  Int i;

  nAllSegments = nSegmentSignaled;

  *pFirstSegmentFlat = (global_offset_val>0);
  if(*pFirstSegmentFlat)
  {
    nAllSegments++;
    first = 1;
    pAllSegmentLengths[0] = global_offset_val;
  }  
  else
  {
    first = 0;
  }
  for(i=0;i<nSegmentSignaled;i++)
    pAllSegmentLengths[i+first] = SegmentLengths[i];

  sum = 0;
  for(i=0;i<nAllSegments;i++)
    sum = sum + pAllSegmentLengths[i];
  RemainingLength = total_length - sum;
  assert(RemainingLength>=0);
  // TBD fix for limit of current syntax can't signal pivot point of 1024 so sum of lengths is 1023 at most
  if(RemainingLength==1)
  {
    pAllSegmentLengths[nAllSegments-1] = pAllSegmentLengths[nAllSegments-1] + 1;
    RemainingLength = 0;
  }
  if(RemainingLength>0)
  {
    *pLastSegmentFlat = true;
    nAllSegments++;
    pAllSegmentLengths[nAllSegments-1] = RemainingLength;
  }
  else
  {
    *pLastSegmentFlat = false;
  }

  *pNumberAllSegments = nAllSegments;
  return;
}
/**
 - Determine piecewise model polynomial coefficients from reduced signalled set.  Note coefficients on flat intervals are also set , return full set of model coefficients
 
 \param   nSegmentSignaled    number of segments explicitly signaled
 \param   SegmentLengths      signalled segment lengths
 \param   constantCoeff       values of constant coefficients signalled
 \param   linearCoeff         value of signalled linear coefficient
 \param   FirstSegmentFlatFlag flag indicating last segment is flat
 \param   LastSegmentFlatFlag flag indicating last segment is flat
 \param   nFractionBits       number of fraction bits for model constant coefficients
 \retval  polyCoeff           piecewise polynomial model coefficients
 */
Void TComReshape::DeriveSmoothPolynomialCoefficients(Int nSegmentSignaled, Int SegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS], Int constantCoeff[1+MAX_LUMA_RESHAPE_SEGMENTS], Int linearCoeff, Bool FirstSegmentFlatFlag, Bool LastSegmentFlatFlag,Int nFractionBits[MAX_LUMA_PRED_ORDER+1],Int64 polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_LUMA_PRED_ORDER+1])
{
  Int64 Atemp[1+MAX_LUMA_RESHAPE_SEGMENTS];
  Int64 Btemp[1+MAX_LUMA_RESHAPE_SEGMENTS];
  Int64 s;
  Int FirstSegment; // first signalled segment 0 or 1
  Int nSegments;

  nSegments = nSegmentSignaled;
  if(FirstSegmentFlatFlag)
  {
    FirstSegment = 1;
    nSegments = nSegments +1;
  }
  else
  {
    FirstSegment = 0;
  }
  if(LastSegmentFlatFlag)
  {
    nSegments = nSegments +1;
  }
  // set temporary A values for derivation
  for(s=0;s<nSegmentSignaled+1;s++)
  {
    Atemp[s] = constantCoeff[s];
  }
  Btemp[0] = (Int64) linearCoeff;
  // solve for remaining B values for derivation
  for(s=1;s<nSegmentSignaled+1;s++)
  {
    Int64 Numerator;
    Numerator = (Atemp[s] - Atemp[s-1])<<(nFractionBits[1]-nFractionBits[0] + 1);
    Numerator = Numerator + (SegmentLengths[s-1]>>1);
 
    Btemp[s] = (Numerator/SegmentLengths[s-1]) - Btemp[s-1];
  }
  // solve for C values intermediate calculations may exceed 32 bits and set output A an B values in signalled range
  for(s=0;s<nSegmentSignaled;s++)
  {
    Int64 Numerator;
    polyCoeff[s+FirstSegment][0] = Atemp[s];
    polyCoeff[s+FirstSegment][1] = Btemp[s];
    Numerator = ((Int64)(Btemp[s+1] - Btemp[s])) << (nFractionBits[2]-nFractionBits[1] - 1);
    Numerator = Numerator + (SegmentLengths[s]>>1);
    polyCoeff[s+FirstSegment][2] = (Int64) Numerator/SegmentLengths[s];
  }  
  if(FirstSegmentFlatFlag)
  {
    polyCoeff[0][0] = constantCoeff[0];
    polyCoeff[0][1] = 0;
    polyCoeff[0][2] = 0;
  }  
  if(LastSegmentFlatFlag)
  {

    polyCoeff[nSegments-1][0] = constantCoeff[nSegmentSignaled];
    polyCoeff[nSegments-1][1] = 0;
    polyCoeff[nSegments-1][2] = 0;

  }
}
/**
 - Determine piecewise model polynomial coefficients from reduced signalled set.  Note coefficients on flat Intervals are also set , return full set of model coefficients
 
 \param   nSegmentSignaled    number of segments explicitly signaled
 \param   SegmentLengths      signalled segment lengths
 \param   constantCoeff       values of constant coefficients signalled
 \param   FirstSegmentFlatFlag flag indicating last segment is flat
 \param   LastSegmentFlatFlag flag indicating last segment is flat
 \param   nFractionBits       number of fraction bits for model constant coefficients
 \retval  polyCoeff           piecewise polynomial model coefficients
 */
Void TComReshape::DerivePiecewiseLinearPolynomialCoefficients(Int nSegmentSignaled, Int SegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS], Int constantCoeff[1+MAX_LUMA_RESHAPE_SEGMENTS],Bool FirstSegmentFlatFlag,Bool LastSegmentFlatFlag,Int nFractionBits[MAX_LUMA_PRED_ORDER+1],Int64 polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_LUMA_PRED_ORDER+1])
{
  Int64 s;
  Int FirstSegment; // first signalled segment 0 or 1
  Int nSegments;

  nSegments = nSegmentSignaled;
  if(FirstSegmentFlatFlag)
  {
    FirstSegment = 1;
    nSegments = nSegments + 1;
  }
  else
  {
    FirstSegment = 0;
  }
  if(LastSegmentFlatFlag)
  {
    nSegments = nSegments + 1;
  }
  // set A, B, C values within segments 
  for(s=0;s<nSegmentSignaled;s++)
  {
    polyCoeff[s+FirstSegment][0] = constantCoeff[s];
    polyCoeff[s+FirstSegment][1] = (((constantCoeff[s+1]-constantCoeff[s])<<(nFractionBits[1]-nFractionBits[0])) + (SegmentLengths[s]>>1))/SegmentLengths[s];
    polyCoeff[s+FirstSegment][2] = 0;
  }
  if(FirstSegmentFlatFlag)
  {
    polyCoeff[0][0] = constantCoeff[0];
    polyCoeff[0][1] = 0;
    polyCoeff[0][2] = 0;
  }  
  if(LastSegmentFlatFlag)
  {

    polyCoeff[nSegments-1][0] = constantCoeff[nSegmentSignaled];
    polyCoeff[nSegments-1][1] = 0;
    polyCoeff[nSegments-1][2] = 0;

  }
}
/**
 - Build inverse reshaping LUT from full set of model segments and coefficients
 
 \param   nSegmentAll         number of segments explicitly signaled and inferred
 \param   allSegmentLengths   lengths of all segments thoes explicitly signalled and inferred
 \param   nFractionBits       number of fraction bits for model constant coefficients
 \param   polyCoeff           piecewise polynomial model coefficients
 \param   lut_size            size of LUT in number of samples
 \retval  InverseReshapingLUT LUT describing inverse reshaping
 */
Void TComReshape::BuildInverseReshapingLUT(Int nSegmentAll, Int allSegmentLengths[MAX_LUMA_RESHAPE_SEGMENTS],Int nFractionBits[MAX_LUMA_PRED_ORDER+1], Int64 polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_LUMA_PRED_ORDER+1],Int lut_size, Int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE])
{
  Int s;
  Int dx;
  Int idx;
  Int64 result;
  Int xi[MAX_LUMA_RESHAPE_SEGMENTS];
  Int order;
  Int64 ModPolyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_LUMA_PRED_ORDER + 1];

  // modifiy coefficients
  // determine left endpoInts of each segment
  xi[0]=0;
  for(s=1;s<nSegmentAll;s++)
    xi[s] = xi[s-1] + allSegmentLengths[s-1];
  // constant coefficients 
  order = 0;
  for(s=0;s<nSegmentAll;s++)
  {
    ModPolyCoeff[s][order] = ( 

      (((Int64)polyCoeff[s][0]) << (nFractionBits[2]-nFractionBits[0])) - 
      (((Int64)polyCoeff[s][1]) << (nFractionBits[2]-nFractionBits[1]))*xi[s] + 
      (((Int64)polyCoeff[s][2])*xi[s]*xi[s]) + 
      (((Int64)1)<<(nFractionBits[2]-nFractionBits[0]-1)) ) >> 

      (nFractionBits[2]-nFractionBits[0]);
  }
  // linear coefficients 
  order = 1;
  for(s=0;s<nSegmentAll;s++)
  {
    ModPolyCoeff[s][order] = ( 
      (((Int64)polyCoeff[s][1]) <<(nFractionBits[2]-nFractionBits[1])) - 
      (((Int64)polyCoeff[s][2])*xi[s]*2) + 
      (((Int64)1)<<(nFractionBits[2]-nFractionBits[1]-1)) ) >> 

      (nFractionBits[2]-nFractionBits[1]);
  }
  // quadratic coefficients unmodified
  order = 2;
  for(s=0;s<nSegmentAll;s++)
    ModPolyCoeff[s][order] = polyCoeff[s][order];
  // calculate using modified coefficients and description relative to absolute zero A' + B'*x + C'*x*x
  idx = 0;
  for(s=0;s<nSegmentAll;s++)
  {
    for(dx=0;dx<allSegmentLengths[s];dx++)
    {
      result = ModPolyCoeff[s][0] << (nFractionBits[2]-nFractionBits[0]);
      result = result + (ModPolyCoeff[s][1] << (nFractionBits[2]-nFractionBits[1]))*((Int64)idx);
      result = result + ModPolyCoeff[s][2]*((Int64)idx)*idx;                             // TBD note on rounding offset in the spec text right now
#if GUARD_INV_LUT
      if(idx<MAX_LUMA_RESHAPING_LUT_SIZE)
      {
        InverseReshapingLUT[idx] = Clip3( 0, MAX_LUMA_RESHAPING_LUT_SIZE-1, (Int)(result >> nFractionBits[2])); // TBD determine upper clipping limit from bitdepth
        idx = idx+1;
	  }	
#else
      InverseReshapingLUT[idx] = Clip3( 0, MAX_LUMA_RESHAPING_LUT_SIZE-1, (Int)(result >> nFractionBits[2])); // TBD determine upper clipping limit from bitdepth
      idx = idx+1;
#endif
    }
  }
}


/**
- Build forward reshaping LUT from full set of model segments and coefficients
\param   InverseReshapingLUT LUT describing inverse reshaping
\param   lut_size            size of LUT in number of samples
\retval  ForwardReshapingLUT LUT describing forward reshaping
*/
void TComReshape::BuildForwardReshapingLUT(int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int ForwardReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size)
{
  int low = InverseReshapingLUT[0];
  int high = InverseReshapingLUT[lut_size-1];
  int j;
  int first = 0;;
  for(int i=0;i<lut_size;i++)
    if (InverseReshapingLUT[0] < InverseReshapingLUT[i])
    {
      first = i-1;
      break;
    }
    int last=lut_size-1;
    for(int i=lut_size-1;i>=0;i--)
      if (InverseReshapingLUT[lut_size-1] > InverseReshapingLUT[i])
      {
        last = i+1;
        break;
      }
      for(int i=0;i<lut_size;i++)
      {
        if(i<=low)
        {
          ForwardReshapingLUT[i] = first;
        }
        else if(i>=high)
        {
          ForwardReshapingLUT[i] = last;
        }
        else
        {
          for(j=0;j<lut_size-1;j++)
            if(InverseReshapingLUT[j]>=i)
            {
              ForwardReshapingLUT[i] = j;
              break;
            }
        }
      }
}


Void TComReshape::BuildIdentityLUT(Int LUT[MAX_LUMA_RESHAPING_LUT_SIZE],Int lut_size)
{
  Int i;
  for(i=0;i<lut_size;i++)
    LUT[i] = i;
}
/**
 - set fraction bits for all coefficients of differnt order from order 0 coefficients
 
 \param   coeff_log2_offset_minus2  syntax element defining number of fraction bits
 \param   bitDepth                  bitdepth of source
 \retval  fractionBits              array of number of coefficient fraction bits per  order
 */
Void TComReshape::DetermineAllFractionBits(Int coeff_log2_offset_minus2,Int bitDepth,Int fractionBits[MAX_LUMA_PRED_ORDER+1])
{
  Int i;
  for(i=0;i<MAX_LUMA_PRED_ORDER+1;i++)
    fractionBits[i] = coeff_log2_offset_minus2 + 2 + i*bitDepth;
  return;
}
Void TComReshape::WriteCoefficientData(FILE *fp,ReshapeInfo *pInfo)
{
  Int i;
  fprintf(fp,"fraction bits per order ");
  for(i=0;i<MAX_LUMA_PRED_ORDER+1;i++)
    fprintf(fp,"%d ",pInfo->fractionBits[i]);
  fprintf(fp,"\n");
  fprintf(fp,"%d Signalled Lengths\n",pInfo->reshape_num_ranges_minus1 + 1);
  for(i=0;i<pInfo->reshape_num_ranges_minus1 + 1;i++)
  {
    fprintf(fp,"%d ",pInfo->reshape_range_val[i]);
  }
  fprintf(fp,"\n");  
  fprintf(fp,"\n");
  fprintf(fp,"%1s %4s %4s %6s %6s\n","i","Li","Ai","Bi","Ci");
  for(i=0;i<pInfo->nAllSegmentLengths;i++)
  {
    fprintf(fp,"%1d ",i);
    fprintf(fp,"%4d ",pInfo->allSegmentLengths[i]);    
    fprintf(fp,"%lld ",pInfo->polyCoeff[i][0]);
    fprintf(fp,"%lld ",pInfo->polyCoeff[i][1]);
    fprintf(fp,"%lld ",pInfo->polyCoeff[i][2]);
    fprintf(fp,"\n");
  }
  return;
}
Void TComReshape::WriteMetaData(FILE *fp,ReshapeInfo *pInfo)
{
  fprintf(fp,"%d ",pInfo->POC);
  fprintf(fp,"%d ",pInfo->luma_bit_depth);
  fprintf(fp,"%d ",pInfo->reshape_num_ranges_minus1 + 1);
  fprintf(fp,"%d ",pInfo->reshape_equal_ranges);
  fprintf(fp,"%d ",pInfo->reshape_global_offset_val);
  for (Int s = 0; s < pInfo->reshape_num_ranges_minus1 + 1; s++)
    fprintf(fp,"%d ",pInfo->reshape_range_val[s]);
  fprintf(fp,"%d ",pInfo->coeff_log2_offset_minus2 + 2);
  fprintf(fp,"%d ",pInfo->reshape_continuity_flag);
  for (Int s = 0; s < pInfo->reshape_num_ranges_minus1 + 2; s++)
    fprintf(fp,"%d ",pInfo->constantCoeffSignalled[s]);
  if(pInfo->reshape_continuity_flag)
    fprintf(fp,"%d ",pInfo->linearCoeffSignalled);
  return;
}
Void TComReshape::ReadMetaData(FILE *fp,ReshapeInfo *pInfo)
{
  Int TempInt;
  Int s;
  fscanf(fp,"%d ",&pInfo->POC);
  fscanf(fp,"%d ",&pInfo->luma_bit_depth);
  fscanf(fp,"%d ",&TempInt);
  pInfo->reshape_num_ranges_minus1 = TempInt-1;
  fscanf(fp,"%d ",&pInfo->reshape_equal_ranges);
  fscanf(fp,"%d ",&pInfo->reshape_global_offset_val);
  for (s = 0; s < pInfo->reshape_num_ranges_minus1 + 1; s++)
    fscanf(fp,"%d ",&pInfo->reshape_range_val[s]);
  fscanf(fp,"%d ",&TempInt);
  pInfo->coeff_log2_offset_minus2 = TempInt-2;
  fscanf(fp,"%d ",&pInfo->reshape_continuity_flag);
  for (s = 0; s < pInfo->reshape_num_ranges_minus1 + 2; s++)
    fscanf(fp,"%d ",&pInfo->constantCoeffSignalled[s]);
  if(pInfo->reshape_continuity_flag)
    fscanf(fp,"%d ",&pInfo->linearCoeffSignalled);
#if RESHAPER_JOINT_OPT 
  fscanf(fp,"%d ",&pInfo->opt);
  if (pInfo->opt)
  {
    for (Int i=0; i<15; i++)
    {
      fscanf(fp, "%d ", &pInfo->dQP[i]);
      //printf("%d ", pInfo->dQP[i]);
    }
    //printf("\n");
  }
#endif
  // derived values
  pInfo->reshapePresentFlag = true;
  DetermineAllFractionBits(pInfo->coeff_log2_offset_minus2,pInfo->luma_bit_depth,pInfo->fractionBits);
  for (s = 0; s < pInfo->reshape_num_ranges_minus1 + 2; s++)
  {
    pInfo->constantCoeffSignalledInt[s] = pInfo->constantCoeffSignalled[s] >> pInfo->fractionBits[0];
    pInfo->constantCoeffSignalledFrac[s] = pInfo->constantCoeffSignalled[s] & ((1<< pInfo->fractionBits[0])-1);
  }
  pInfo->linearCoeffSignalledInt = pInfo->linearCoeffSignalled >> pInfo->fractionBits[1];
  pInfo->linearCoeffSignalledFrac = pInfo->linearCoeffSignalled & ((1<< pInfo->fractionBits[1])-1);
  return;
}


Float TComReshape::evaluateFloatRepn(Int x, Int const numBitsVal, Int const numBitsFrac)
{
  assert(x < (1 << numBitsVal) && x >= 0);

  Int signPart = x >> numBitsFrac;
  Int numSignBits = numBitsVal - numBitsFrac;
  assert(signPart >= 0 && signPart < (1 << numSignBits));

  Int fracPart = x - (signPart << numBitsFrac);
  assert(fracPart >= 0 && fracPart < (1 << numBitsFrac));

  Float retVal = signPart + (float)fracPart / (1 << numBitsFrac);

  assert(retVal >= 0 && retVal < (1 << numBitsVal));
  return retVal;
}

Int TComReshape::evaluateIntRepn(Float const x, Int const numBitsVal, Int const numBitsFrac)
{
  assert(x >= 0);   // May need to check if more code needed to support this
  Float fracVal = x - (Int)x;
  assert(fracVal >= 0 && fracVal < 1);

  Int numBitsSignif = numBitsVal - numBitsFrac;
  if (numBitsSignif == 0)
  {
    assert(x < 1 && (x - fracVal) == 0); // If not, repn not sufficient to losslessly signal mantissa
  }
  else
  {
    assert(x - fracVal < (1 << numBitsSignif)); // If not, repn not sufficient to losslessly signal mantissa
  }

  // Find the integer repn. of the expo
  Int intRepnOfFrac = (Int)floor((fracVal * (1 << numBitsFrac)) + 0.5); // / (1 << numBitsFrac);

  // Find the new integer representation
  Int retVal = (Int)x * (1 << numBitsFrac) + intRepnOfFrac;

  assert(retVal >= 0 && retVal < (1 << numBitsVal));
  return retVal;
}


Int TComReshape::evaluateIntRepnTCH(Float const x, Int const numBitsVal, Int const numBitsFrac)
{
    int     nbInt=numBitsVal-numBitsFrac;
    int     maxi = (1<<nbInt) / 2 - 1; // /2 to manage negative values
    float   px = x;

    if (nbInt==0) nbInt=1;
    --nbInt; // sign
    if (maxi<=0) maxi=1;

    assert(px<=maxi && px>=-maxi); // need to use more bits to code integer value of px

    float fracVal;
    int signMask=0;
    if (px<0.f) { fracVal=-(px-(float)((int)px)); signMask=(1<<(numBitsVal-1)); px = -px; }   // always code the sign
    else         fracVal=px - (float)((int)px);
    int intRepnOfFrac = (int)((fracVal * (float)(1 << numBitsFrac)) + 0.5f);

    int intPart=(int)px;
    Int retVal = (intPart<< numBitsFrac) + intRepnOfFrac;

    retVal|=signMask;
    assert(retVal < (1u << numBitsVal));
    return retVal;
}


Float TComReshape::evaluateFloatRepnTCH(Int x, Int const numBitsVal, Int const numBitsFrac)
{
  assert(x < (1 << numBitsVal) );
  const int signMask=(1<<(numBitsVal-1));
  bool neg=false;
  if (x&signMask) { neg=true; x=(x&(signMask-1)); }

  int intPart = x >> numBitsFrac;
  int fracPart = x - (intPart << numBitsFrac);
  float retVal = (float)intPart + (float)fracPart / (float)(1 << numBitsFrac);
  if (neg) retVal=-retVal;
  return retVal;
}

Void TComReshape::ReadChromaMetaData(FILE *fpt1, ReshapeInfo *pInfo)
{
	Int chromId, numComponents = 2;
	Int tmp;
	Float tempFloat;
	Int iPOC;

	//POC
	fscanf(fpt1, "%d ", &iPOC);
	assert(iPOC == pInfo->POC);

	pInfo->reshaperATFId = 16;		// Set default PQ value
	fscanf(fpt1, "%d ", &(pInfo->reshaperRangeId));           
	fscanf(fpt1, "%d ", &chromId);            // Domain of applicability (not used currently)
	fscanf(fpt1, "%d ", &tmp);
	pInfo->reshapeChromaCrosschannelFlag = tmp;  // chroma plane mode 0-auto, 1-cross 
	if (pInfo->reshapeChromaCrosschannelFlag == 1)
	{
		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeScaleIntBitDepth = tmp;
		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeOffsetIntBitDepth = tmp;
		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeScaleFracBitDepth = tmp;
		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeOffsetFracBitDepth = tmp;

		pInfo->reshapeOffsetBitDepth = pInfo->reshapeOffsetIntBitDepth + pInfo->reshapeOffsetFracBitDepth;
		pInfo->reshapeScaleBitDepth = pInfo->reshapeScaleIntBitDepth + pInfo->reshapeScaleFracBitDepth;

		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeNegativeScalesPresentFlag = tmp;
	}
  fscanf(fpt1, "%d ", &numComponents);    // Number of chroma model-parameters for reshaping
  pInfo->reshapeNumComps=numComponents+1; // Total number of model-parameters for reshaping (add "1" for Luma)

	for (Int c = 1; c < numComponents + 1; c++) // chromaATF parameters
	{
		fscanf(fpt1, "%d ", &pInfo->reshapeNumRanges[c]);            // Number of ranges
		fscanf(fpt1, "%d ", &tmp);
		pInfo->reshapeEqualRangesFlag[c] = tmp; // Equal width ranges

		fscanf(fpt1, "%d ", &tmp);
		tempFloat = evaluateFloatRepn(tmp, pInfo->reshapeOffsetIntBitDepth + pInfo->reshapeOffsetFracBitDepth, pInfo->reshapeOffsetFracBitDepth);

		pInfo->reshapeGlobalOffsetVal[c] = evaluateIntRepn(tempFloat, pInfo->reshapeOffsetIntBitDepth + pInfo->reshapeOffsetFracBitDepth, pInfo->reshapeOffsetFracBitDepth);            // Zero offset value

		for (Int i = 0; i < pInfo->reshapeNumRanges[c]; i++)
		{
			fscanf(fpt1, "%d ", &tmp);

			if (pInfo->reshapeNegativeScalesPresentFlag == 1)
				tempFloat = evaluateFloatRepnTCH(tmp, pInfo->reshapeScaleIntBitDepth + pInfo->reshapeScaleFracBitDepth, pInfo->reshapeScaleFracBitDepth);
			else
				tempFloat = evaluateFloatRepn(tmp, pInfo->reshapeScaleIntBitDepth + pInfo->reshapeScaleFracBitDepth, pInfo->reshapeScaleFracBitDepth);
			if (pInfo->reshapeNegativeScalesPresentFlag == 1)
				pInfo->reshapeScaleVal[c][i] = evaluateIntRepnTCH(tempFloat, pInfo->reshapeScaleIntBitDepth + pInfo->reshapeScaleFracBitDepth, pInfo->reshapeScaleFracBitDepth); // Output bit depth
			else
				pInfo->reshapeScaleVal[c][i] = evaluateIntRepn(tempFloat, pInfo->reshapeScaleIntBitDepth + pInfo->reshapeScaleFracBitDepth, pInfo->reshapeScaleFracBitDepth);; // Output bit depth
		}
		if (!pInfo->reshapeEqualRangesFlag[c])
		{
			for (Int i = 0; i < pInfo->reshapeNumRanges[c]; i++)
			{
				fscanf(fpt1, "%d ", &tmp);
				tempFloat = evaluateFloatRepn(tmp, pInfo->reshapeOffsetIntBitDepth + pInfo->reshapeOffsetFracBitDepth, pInfo->reshapeOffsetFracBitDepth);
				pInfo->reshapeRangeVal[c][i] = evaluateIntRepn(tempFloat, pInfo->reshapeOffsetIntBitDepth + pInfo->reshapeOffsetFracBitDepth, pInfo->reshapeOffsetFracBitDepth); // Output bit depth
			}
		}
		if (pInfo->reshapeChromaCrosschannelFlag == 1)
		{
		  fscanf(fpt1, "%d ", &tmp);    // offset1
		  pInfo->reshapeOffset1[c-1]=tmp; 
		  fscanf(fpt1, "%d ", &tmp);    // offset2
		  pInfo->reshapeOffset2[c-1]=tmp; 
		}
	}
  if(numComponents<2)
  {
		pInfo->reshapeNumRanges[2]=pInfo->reshapeNumRanges[1];
		pInfo->reshapeEqualRangesFlag[2] = pInfo->reshapeEqualRangesFlag[1];
		pInfo->reshapeGlobalOffsetVal[2] = pInfo->reshapeGlobalOffsetVal[1];
    memcpy((void *)pInfo->reshapeScaleVal[2], (void *)pInfo->reshapeScaleVal[1], pInfo->reshapeNumRanges[2]*sizeof(pInfo->reshapeScaleVal[1][0]));
		if (!pInfo->reshapeEqualRangesFlag[2])
		{
      memcpy((void *)pInfo->reshapeRangeVal[2], (void *)pInfo->reshapeRangeVal[1], pInfo->reshapeNumRanges[2]*sizeof(pInfo->reshapeRangeVal[1][0]));
		}
		if (pInfo->reshapeChromaCrosschannelFlag == 1)
    {
      pInfo->reshapeOffset1[1] = pInfo->reshapeOffset1[0];
      pInfo->reshapeOffset2[1] = pInfo->reshapeOffset2[0];
    }
  }

	return;
}
#if CHORMA_DQP_IMP
Void TComReshape::identifyComponentRangeIdx(Pel sampleVal, Int shiftVal, std::vector<fixedPrec> &rangeValues,  Int &idxS)
{
  Int shiftedSampleVal = sampleVal << shiftVal;
  if( shiftedSampleVal < rangeValues[0].val ) {
    idxS = 0;
  } else if( shiftedSampleVal > rangeValues.back().val ) {
    idxS = rangeValues.size();
  } else {
    for(idxS = 1; idxS < rangeValues.size(); idxS++) {
      if( (shiftedSampleVal > rangeValues[idxS-1].val) && (shiftedSampleVal <= rangeValues[idxS].val) )
        break;
    }
  }
}

Void TComReshape::buildInverseReshapingCrossPlane(ReshapeInfo const *pInfo, std::vector<fixedPrec> *inputRanges, std::vector<fixedPrec> *scaleValues, std::vector<fixedPrec> *offsetValue, Int *chromaLutU, Int *chromaLutV)
{
  Int     bitDepthL = pInfo->luma_bit_depth;
  Int     deltaBD1 = TCH_DEFAULT_PRECISION + pInfo->reshapeOffsetFracBitDepth - bitDepthL;

  for (int c=1; c<3; c++)
  {
    for(Int i = 0; i < MAX_LUMA_RESHAPING_LUT_SIZE; i++ )
    {
      Int   idx = 0;
      Int   ii = i << deltaBD1;

      if ( ii <= inputRanges[c][0].val )
        idx = 0;
      else if ( ii > inputRanges[c][pInfo->reshapeNumRanges[c]].val )
        idx = pInfo->reshapeNumRanges[c] - 1;
      else
      {
        for (idx=0; idx<pInfo->reshapeNumRanges[c]; idx++)
        {
          if (ii > inputRanges[c][idx].val && ii <= inputRanges[c][idx+1].val)
            break;
        }
      }

      Int   val1 = ( i * scaleValues[c][idx].val ) << deltaBD1;
      Int   val2 = offsetValue[c][idx].val;
      Int   val = val1 + val2;
      if (c==1)
        chromaLutU[i] = val;
      else
        chromaLutV[i] = val;
    }
  }
}

#endif
#endif //HDR_ETM_RESHAPER
//! \}
